// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

#ifndef YAMICORE_CHANNEL_H_INCLUDED
#define YAMICORE_CHANNEL_H_INCLUDED

#include <yami4-core/core.h>
#include "details-fwd.h"
#include "options.h"
#include <cstddef>

#ifdef YAMI4_WITH_OPEN_SSL
#ifdef _WIN32
// this is necessary to avoid conflicts between
// WinSock2.h (used by YAMI4) and winsock.h (used by OpenSSL):
#include <WinSock2.h>
#endif // _WIN32
#include <openssl/ssl.h>
#endif // YAMI4_WITH_OPEN_SSL

// selected per platform
#include <details-types.h>

namespace yami
{

namespace details
{

enum incoming_fsm { read_frame_header, read_frame_payload };

const std::size_t word_size = 4;
const std::size_t frame_head_size = 4 * word_size;

class selector;

class channel
{
public:

    // for initializing newly created channels
    // this function attempts to create a physical connection
    // according to the given target
    // should be called outside of critical section
    core::result init(allocator & alloc, mutex & mtx,
        const options & configuration_options,
        const core::parameters * overriding_options,
        const char * target,
        core::incoming_message_dispatch_function incoming_message_callback,
        void * incoming_message_hint,
        core::event_notification_function event_notification_callback,
        void * event_notification_hint,
        core::io_error_function io_error_callback,
        void * io_error_callback_hint);

    // for initializing channels that are already physically created
    // by listeners
    // target is transfered from listener to this channel
    core::result init(allocator & alloc, mutex & mtx,
        const options & configuration_options,
        char * target, io_descriptor_type & fd, protocol prot,
        std::size_t preferred_frame_size,
        core::incoming_message_dispatch_function incoming_message_callback,
        void * incoming_message_hint,
        core::event_notification_function event_notification_callback,
        void * event_notification_hint,
        core::io_error_function io_error_callback,
        void * io_error_callback_hint);

#ifdef YAMI4_WITH_NETX
    void set_netx_context(NX_IP * ip, NX_PACKET_POOL * packet_pool, TX_EVENT_FLAGS_GROUP * events);
    void register_input_event();
    void close_tcp_connection();
    core::result post_init_accepted_channel();
#endif // YAMI4_WITH_NETX
#ifdef YAMI4_WITH_LWIP
    void set_lwip_context(EventGroupHandle_t events);
    void register_udp_input_event(struct pbuf * p, const ip_addr_t * addr, u16_t port);
    void register_tcp_connected();
    void register_tcp_input_event(struct pbuf * p);
    void post_init_accepted_channel();
#endif // YAMI4_WITH_LWIP
    
#ifdef YAMI4_WITH_OPEN_SSL
    void set_client_ssl(SSL * ssl);
    void set_server_ssl(SSL * ssl);
#endif // YAMI4_WITH_OPEN_SSL

    void close_connection();
    void clean();

    std::size_t get_frame_size() const;

    // buffers are transferred into the internally managed outgoing queue
    core::result post(std::size_t priority,
        char * * buffers, const std::size_t * buffer_sizes,
        std::size_t number_of_frames,
        bool & first_frame,
        core::message_progress_function progress_callback,
        void * progress_hint);

    // intended for cases when the incoming frame is physically read
    // outside of the channel, but also used for datagrams (UDP)
    core::result process_complete_incoming_frame(
        const char * buffer, const std::size_t buffer_size);

    core::result do_some_work(io_direction direction, bool & close_me);

#ifdef YAMI4_WITH_QNX
    // used for QNX channels to send, in a blocking way,
    // a single outgoing frame (if there is any)
    core::result send_blocking_datagram(bool & close_me);
    
    // used to establish the channel id for receiving responses
    // from remote peers, this id is sent together with each outgoing message
    void set_default_qnx_listening_channel_id(int chid);
    
    // used by channel to interrupt the selector
    // after successful send of the QNX frame,
    // so that the worker thread can wake up if it happened to be waiting
    void set_selector(selector & slc);
#endif // YAMI4_WITH_QNX
    
    core::result post_close(std::size_t priority, bool & close_me);

    const char * get_target() const { return target_; }
    const char * move_target();

    const io_descriptor_type & get_io_descriptor() const { return fd_; }

    io_direction get_io_direction() const;

    bool has_buffered_data() const { return buffer_available_ > buffer_consumed_; }
    
#ifdef YAMI4_WITH_OPEN_SSL
    SSL * get_ssl() const { return ssl_; }

    void set_pending_read(bool pending) { pending_read_ = pending; }
    bool get_pending_read() const { return pending_read_; }
#endif // YAMI4_WITH_OPEN_SSL

    protocol get_protocol() const { return protocol_; }

    void set_selector_index(std::size_t index) { selector_index_ = index; }
    std::size_t get_selector_index() const { return selector_index_; }

    void inc_ref() { ++ref_count_; }
    void dec_ref() { --ref_count_; }
    bool can_be_removed() const { return ref_count_ == 0; }

    // for unit tests
    const outgoing_frame * get_first_outgoing_frame() const;
    const outgoing_frame * get_last_outgoing_frame() const;
    const incoming_message_frame_list * get_first_incoming_frame_list() const;
    const incoming_message_frame_list * get_last_incoming_frame_list() const;
    incoming_fsm get_incoming_state() const;
    
    std::size_t get_pending_outgoing_bytes() const;

    // for unit tests:
    channel() {}
    
private:
    
    channel(const channel &);
    void operator=(const channel &);

    void common_init(allocator & alloc, mutex & mtx,
        const options & configuration_options,
        const core::parameters * overriding_options,
        core::incoming_message_dispatch_function incoming_message_callback,
        void * incoming_message_hint,
        core::event_notification_function event_notification_callback,
        void * event_notification_hint,
        core::io_error_function io_error_callback,
        void * io_error_callback_hint);

    outgoing_frame * * find_outgoing_insertion_point(std::size_t priority);

    // cleans all frames starting with the given one
    void clean_outgoing_frames(outgoing_frame * frame);

    void notify_cancellation(
        core::message_progress_function progress_callback,
        void * progress_hint);

    void parse_incoming_frame_header();

    core::result store_incoming_frame();

    core::result dispatch_short_incoming_message();
    core::result dispatch_incoming_message(
        incoming_message_frame_list * list);

    // cleans all incoming frame lists and their frames
    void clean_incoming_messages();

    // cleans the given list
    void clean_incoming_frames(incoming_message_frame_list * list);

    void notify_progress_and_consume_frame_list();

    core::result do_some_input();
    core::result do_some_output(bool & close_me);

    core::result connect();
    core::result connect_to_tcp(const char * tcp_address);
    core::result connect_to_udp(const char * udp_address);
#if !defined(_WIN32) && !defined(YAMI4_WITH_NETX) && !defined(YAMI4_WITH_LWIP) && !defined(YAMI4_WITH_TIRTOS)
    core::result connect_to_file(const char * file_name);
    core::result connect_to_unix(const char * path);
#endif // !_WIN32 && !YAMI4_WITH_NETX && !YAMI4_WITH_LWIP && !YAMI4_WITH_TIRTOS
#ifdef YAMI4_WITH_QNX
    core::result connect_to_qnx(const char * qnx_endpoint);
#endif // YAMI4_WITH_QNX
#ifdef YAMI4_WITH_OPEN_SSL
    core::result connect_to_tcps(const char * tcp_address);
#endif // YAMI4_WITH_OPEN_SSL

    core::result read_bytes(char * buf, std::size_t size,
        std::size_t & readn);
    core::result write_bytes(const char * buf, std::size_t size,
        std::size_t & written);

    options configuration_options_;
    allocator * alloc_;

#ifdef YAMI4_WITH_OPEN_SSL
    SSL * ssl_;
    bool pending_read_;
#endif // YAMI4_WITH_OPEN_SSL

    mutex * mtx_; // provided by channel group

#ifdef YAMI4_WITH_NETX
    NX_IP * ip_;
    NX_PACKET_POOL * packet_pool_;
    TX_EVENT_FLAGS_GROUP * events_;
    
    // effective_id_ points to fd_ or to the object prepared by the listener
    io_descriptor_type * effective_fd_;
    bool tcp_channel_disconnected_;
    bool channel_created_by_listener_;
#endif // YAMI4_WITH_NETX
#ifdef YAMI4_WITH_LWIP
    struct pbuf * received_pbuf_;
    EventGroupHandle_t events_;
#endif // YAMI4_WITH_LWIP
    
    char * target_;
    protocol protocol_;
    io_direction direction_;
    io_descriptor_type fd_;
    std::size_t selector_index_;

    // used only for reading with TCP and allocated lazily
    char * frame_buffer_;
    std::size_t buffer_capacity_;
    std::size_t buffer_available_;
    std::size_t buffer_consumed_;
    
    std::size_t preferred_frame_size_;

    // used only with UDP, as it is needed with every send operation
#if defined(YAMI4_WITH_NETX) || defined(YAMI4_WITH_LWIP)
    host_type target_address_;
    port_type target_port_;
#else
    void * target_address_;
    int target_address_size_;
#endif // YAMI4_WITH_NETX || YAMI4_WITH_LWIP
    
    char * datagram_buffer_;
    
#ifdef YAMI4_WITH_QNX
    // used only with QNX channels
    int connection_id_;
    selector * selector_;
    // buffer for serialized chid prefix
    char qnx_listener_channel_id_serialized_[4u];
#endif // YAMI4_WITH_QNX
    
    std::size_t ref_count_;

    // soft close allows output of remaining existing frames,
    // hard close means dropping everything
    enum mode { operational, soft_close, hard_close };
    mode mode_;

    // state of outgoing queue

    std::size_t output_frame_offset_;
    outgoing_frame * first_outgoing_frame_;
    outgoing_frame * last_outgoing_frame_;

    // state of incoming queue

    incoming_fsm incoming_state_;
    std::size_t read_offset_; // within a given state

    char frame_head_buffer_[frame_head_size];
    int current_message_id_;
    std::size_t current_frame_number_;
    std::size_t current_message_header_size_;
    std::size_t current_frame_payload_size_;
    char * current_frame_payload_;
    bool last_frame_;

    incoming_message_frame_list * first_incoming_frame_list_;
    incoming_message_frame_list * last_incoming_frame_list_;
    
    std::size_t pending_outgoing_bytes_;

    core::incoming_message_dispatch_function incoming_message_callback_;
    void * incoming_message_hint_;

    core::event_notification_function event_notification_callback_;
    void * event_notification_hint_;

    core::io_error_function io_error_callback_;
    void * io_error_callback_hint_;
};

} // namespace details

} // namespace yami

#endif // YAMICORE_CHANNEL_H_INCLUDED
