// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

//
// README:
//
// This file is NOT a stand-alone program,
// its main function is at the end of the file
// and is supposed to be a task function, started
// after all hardware and interrupts are initialized.
//
// The following addresses and mask are used during tests:
//
#define MY_TARGET "udp://192.168.11.112:12345"

#include <yami4-core/agent.h>
#include <yami4-core/parameters.h>

#include <cassert>
#include <cstdio>
#include <cstring>

using namespace yami::core;

void dispatch_function(
    void * hint,
    const char * source,
    const char * /* header_buffers */ [],
    std::size_t /* header_buffer_sizes */ [],
    std::size_t /* num_of_header_buffers */,
    const char * body_buffers[],
    std::size_t body_buffer_sizes[],
    std::size_t num_of_body_buffers)
{
    agent * server_agent = static_cast<agent *>(hint);

    // ignore the header, but deserialize the message body

    parameters body;

    result res = body.deserialize(
        body_buffers, body_buffer_sizes, num_of_body_buffers);
    assert(res == ok);

    // extract the content

    const char * value;
    std::size_t value_length;

    res = body.get_string("question", value, value_length);
    assert(res == ok);

    assert(std::strncmp(value, "ping", value_length) == 0);

    // prepare and send back the response

    parameters empty_header;
    parameters reply_body;

    res = reply_body.set_string("answer", "pong");
    assert(res == ok);

    res = server_agent->post(source, empty_header, reply_body);
    assert(res == ok);
}

void new_connection_function(
    void * /* hint */,
    const char * source,
    std::size_t /* index */, std::size_t /* sequence_number */)
{
    std::printf("accepted incoming connection from %s\n", source);
}

void closed_connection(
    void * /* hint */,
    const char * name, result /* reason */)
{
    std::printf("closed channel %s\n", name);
}

extern "C"
void MX_LWIP_Init(void);

extern "C"
void run_server(void * argument)
{
    MX_LWIP_Init();
    
    agent server_agent;

    // create the server agent and activate a listener on the given address

    result res = server_agent.init(
        &dispatch_function, &server_agent,
        &closed_connection, NULL);
    assert(res == ok);

    const char * resolved_address;
    res = server_agent.add_listener(
        MY_TARGET,
        &new_connection_function, NULL,
        &resolved_address);
    assert(res == ok);

    std::printf("The server is listening on %s\n", resolved_address);

    // run the event loop
    // normally this loop would be executed by a separate thread,
    // but this server has nothing else to do, so it is possible
    // to process the I/O events here:

    std::printf("waiting for messages...\n");

    const std::size_t time_out = 100; // 1 second

    while (true)
    {
        res = server_agent.do_some_work(time_out);
        assert(res == ok || res == timed_out);
    }

    // never reaches here
}
