// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

//
// README:
//
// This file is NOT a stand-alone program,
// its main function is at the end of the file
// and is supposed to be a task function, started
// after all hardware and interrupts are initialized.
//
// The following addresses and mask are used during tests:
//
#define SERVER_TARGET "udp://192.168.11.111:12345"

#include <yami4-core/agent.h>
#include <yami4-core/parameters.h>

#include <cassert>
#include <cstdio>

#include <FreeRTOS.h>
#include <task.h>

using namespace yami::core;

agent client_agent;

// worker thread function that is used to process I/O events
// for the client agent
void worker_thread_function(void * argument)
{
    const std::size_t time_out = 1000; // 1 second

    while (true)
    {
        result res = client_agent.do_some_work(time_out);
        assert(res == ok || res == timed_out);
    }
}

extern "C"
void MX_LWIP_Init(void);

extern "C"
void run_client(void * argument)
{
    MX_LWIP_Init();
    
    // create the client agent
    // and open a channel to the given server

    result res = client_agent.init(NULL, NULL, NULL, NULL);
    assert(res == ok);
    
    res = client_agent.open(SERVER_TARGET);
    assert(res == ok);

    // start the worker thread
    
    xTaskCreate(&worker_thread_function, "Worker thread", 1024, NULL, 3, NULL);

    // prepare lines of text for sending to the server
    // and post each one for transmission
    
    const std::size_t line_max_len = 20;
    char line[line_max_len];
    int counter = 0;
    
    while (true)
    {
        std::sprintf(line, "Hello %d", ++counter);
        
        parameters empty_header;
        parameters body;

        // the "content" field name is arbitrary,
        // but needs to be recognized at the server side

        res = body.set_string("content", line);
        assert(res == ok);

        res = client_agent.post(
            SERVER_TARGET, empty_header, body);
        assert(res == ok);
        
        std::printf("posted for sending to server: %s\n", line);
        
        // this not only makes a steady pace,
        // but also gives a chance to the worker thread to do its part
        
        vTaskDelay(1000);
    }
}
