// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

#include "../channel.h"
#include <yami4-core/allocator.h>
#include "../io_error_handler.h"
#include "../network_utils.h"
#include "../options.h"
#include "mutex.h"
#include <cstring>

#include <FreeRTOS.h>
#include <event_groups.h>
#include <lwip/tcp.h>
#include <lwip/udp.h>

using namespace yami;
using namespace details;

namespace // unnamed
{

void udp_channel_input_notify(void * arg, struct udp_pcb * pcb, struct pbuf * p, const ip_addr_t * addr, u16_t port)
{
    channel * this_channel = reinterpret_cast<channel *>(arg);
    
    this_channel->register_udp_input_event(p, addr, port);
}

err_t tcp_connected_notify(void * arg, struct tcp_pcb * tpcb, err_t err)
{
    // even in the case of connection error, the channel can be set up
    // as if ready for communication, the actual error will be reported
    // at the first attempt to write any data to the channel
    
    channel * this_channel = reinterpret_cast<channel *>(arg);
    
    this_channel->register_tcp_connected();
    
    return ERR_OK;
}

err_t tcp_channel_input_notify(void * arg, struct tcp_pcb * pcb, struct pbuf * p, err_t err)
{
    channel * this_channel = reinterpret_cast<channel *>(arg);
    
    this_channel->register_tcp_input_event(p);
    
    return ERR_OK;
}

} // unnamed namespace

void channel::set_lwip_context(EventGroupHandle_t events)
{
    events_ = events;
}

void channel::register_udp_input_event(struct pbuf * p, const ip_addr_t * addr, u16_t port)
{
    // remember the arriving packet and
    // notify selector about this event, so it can wake up:
    
    if (received_pbuf_ != NULL)
    {
        // silently drop older packet
        
        (void)pbuf_free(received_pbuf_);
    }
    
    received_pbuf_ = p;
    
    (void)xEventGroupSetBits(events_, channel_input_event);
}

void channel::register_tcp_input_event(struct pbuf * p)
{
    // remember the arriving packet and
    // notify selector about this event, so it can wake up:
    
    if (received_pbuf_ != NULL)
    {
        // silently drop older packet
        
        (void)pbuf_free(received_pbuf_);
    }
    
    received_pbuf_ = p;
    
    (void)xEventGroupSetBits(events_, channel_input_event);
}

void channel::register_tcp_connected()
{
    // this effectively enables communication:
    
    direction_ = inout;
    
    (void)xEventGroupSetBits(events_, channel_output_event);
}

void channel::post_init_accepted_channel()
{
    tcp_arg(fd_.tcp, this);
    
    received_pbuf_ = NULL;
    
    tcp_recv(fd_.tcp, &tcp_channel_input_notify);
}

void channel::close_connection()
{
    if (received_pbuf_ != NULL)
    {
        (void)pbuf_free(received_pbuf_);
    }
    
    if (protocol_ == proto_udp)
    {
        udp_remove(fd_.udp);
    }
    else if (protocol_ == proto_tcp)
    {
        (void)tcp_close(fd_.tcp);
    }
}

core::result channel::connect_to_tcp(const char * tcp_address)
{
    bool socket_created = false;
    ip_address ipa;
    core::result res = parse_address(*alloc_, tcp_address, ipa,
        io_error_callback_, io_error_callback_hint_);
    if (res == core::ok)
    {
        struct tcp_pcb * pcb = tcp_new();
        if (pcb != NULL)
        {
            socket_created = true;
            
            fd_.tcp = pcb;
            
            tcp_arg(pcb, this);
            
            received_pbuf_ = NULL;
            
            err_t cc = tcp_bind(fd_.tcp, &ip_addr_any, 0);
            if (cc == ERR_OK)
            {
                cc = tcp_connect(fd_.tcp, &ipa.host, ipa.port, &tcp_connected_notify);
                if (cc == ERR_OK)
                {
                    tcp_recv(fd_.tcp, &tcp_channel_input_notify);
                    
                    res = core::ok;
                }
                else
                {
                    handle_io_error_with_code("connect tcp socket", cc,
                        io_error_callback_, io_error_callback_hint_);
                    
                    (void)tcp_close(fd_.tcp);
                    
                    res = core::io_error;
                }
            }
            else
            {
                handle_io_error_with_code("bind tcp socket", cc,
                    io_error_callback_, io_error_callback_hint_);
                
                (void)tcp_close(fd_.tcp);
                
                res = core::io_error;
            }
        }
        else
        {
            handle_io_error("create tcp socket",
                io_error_callback_, io_error_callback_hint_);
            
            res = core::io_error;
        }
    }
    
    if (res == core::ok)
    {
        // stream-oriented transport, allocate a read buffer
        
        buffer_capacity_ = configuration_options_.tcp_frame_size;
        
        frame_buffer_ = static_cast<char *>(
            alloc_->allocate(buffer_capacity_));
        if (frame_buffer_ == NULL)
        {
            res = core::no_memory;
        }
    }
    
    if (res == core::ok)
    {
        preferred_frame_size_ =
            configuration_options_.tcp_frame_size;
        
        protocol_ = proto_tcp;
        direction_ = none; // until connected
    }
    else
    {
        if (socket_created)
        { 
            (void)tcp_close(fd_.tcp);
        }
    }

    return res;
}

core::result channel::connect_to_udp(const char * udp_address)
{
    ip_address ipa;
    core::result res = parse_address(*alloc_, udp_address, ipa,
        io_error_callback_, io_error_callback_hint_);
    if (res == core::ok)
    {
        struct udp_pcb * pcb = udp_new();
        if (pcb != NULL)
        {
            fd_.udp = pcb;
            
            received_pbuf_ = NULL;
            
            udp_recv(fd_.udp, &udp_channel_input_notify, this);
            
            target_address_ = ipa.host;
            target_port_ = ipa.port;
            
            protocol_ = proto_udp;
            direction_ = inout;
            
            preferred_frame_size_ =
                configuration_options_.udp_frame_size;
            
            datagram_buffer_ = static_cast<char *>(
                alloc_->allocate(preferred_frame_size_));
            if (datagram_buffer_ == NULL)
            {
                res = core::no_memory;
                
                udp_remove(fd_.udp);
            }
        }
        else
        {
            handle_io_error("create udp socket",
                io_error_callback_, io_error_callback_hint_);
            
            res = core::io_error;
        }
    }
    
    return res;
}

core::result channel::read_bytes(char * buf, std::size_t size,
    std::size_t & readn)
{
    core::result res;

    switch (protocol_)
    {
    case proto_udp:

        {
            // the UDP frame has to be read as a single operation
            
            if (received_pbuf_ != NULL)
            {
                u16_t bytes_copied = pbuf_copy_partial(received_pbuf_, buf, size, 0ul);
                if (bytes_copied != 0ul)
                {
                    readn = static_cast<std::size_t>(bytes_copied);
                    
                    res = core::ok;
                }
                else
                {
                    handle_io_error("udp pbuf copy",
                        io_error_callback_, io_error_callback_hint_);
                    
                    res = core::io_error;
                }
            }
            else
            {
                // defensive against possible hazards with selector,
                // it is OK to attempt reading a packet that does not exist
                
                readn = 0ul;
                
                res = core::bad_state;
            }
        }
        
        break;

    case proto_tcp:
        
        {
            // the frame buffer was allocated at the connection time
            // or when the channel was added to the group,
            // so it can be readily used here
            
            res = core::ok;
            
            if (buffer_consumed_ == buffer_available_)
            {
                // read new portion of data to the buffer
                
                if (received_pbuf_ != NULL)
                {
                    // only at most buffer_capacity_ is read from the incoming pbuf,
                    // the remaining data (if any) is silently dropped
                    
                    u16_t bytes_copied = pbuf_copy_partial(received_pbuf_, frame_buffer_, buffer_capacity_, 0ul);
                    if (bytes_copied != 0ul)
                    {
                        tcp_recved(fd_.tcp, bytes_copied);
                        
                        buffer_available_ = static_cast<std::size_t>(bytes_copied);
                        buffer_consumed_ = 0;
                        
                        res = core::ok;
                    }
                    else
                    {
                        handle_io_error("tcp copy from pbuf",
                            io_error_callback_, io_error_callback_hint_);
                        
                        res = core::io_error;
                    }
                }
                else
                {
                    res = core::channel_closed;
                }
            }
            
            if (res == core::ok)
            {
                // consume data from the buffer
                
                const std::size_t left_in_buffer =
                    buffer_available_ - buffer_consumed_;
                std::size_t to_read;
                if (size <= left_in_buffer)
                {
                    to_read = size;
                }
                else
                {
                    // defensive, it is possible to ask for
                    // more data that fits in tcp_frame_size,
                    // because the remote side might have a different
                    // preferred frame size and that bigger size
                    // was already accepted in the header
                    // -> in this case the data is read piece-wise
                    // from the smaller buffer
                    
                    to_read = left_in_buffer;
                }
                
                std::memcpy(buf, frame_buffer_ + buffer_consumed_, to_read);
                
                buffer_consumed_ += to_read;
                
                readn = to_read;
            }
        }
        
        break;
    }
    
    return res;
}

core::result channel::write_bytes(const char * buf, std::size_t size,
    std::size_t & written)
{
    core::result res;

    switch (protocol_)
    {
    case proto_udp:
    
        {
            // the UDP frame is either successfully pushed as a whole
            // or the operation is considered as failing
            
            struct pbuf * pb = pbuf_alloc(PBUF_TRANSPORT, size, PBUF_RAM);
            if (pb != NULL)
            {
                err_t cc = pbuf_take(pb, buf, size);
                if (cc == ERR_OK)
                {
                    cc = udp_sendto(fd_.udp, pb, &target_address_, target_port_);
                    if (cc == ERR_OK)
                    {
                        written = size;
                        
                        res = core::ok;
                    }
                    else
                    {
                        handle_io_error_with_code("udp send", cc,
                            io_error_callback_, io_error_callback_hint_);
                        
                        res = core::io_error;
                    }
                }
                else
                {
                    handle_io_error_with_code("udp buffer operation", cc,
                        io_error_callback_, io_error_callback_hint_);
                    
                    res = core::io_error;
                }
                
                (void)pbuf_free(pb);
            }
            else
            {
                handle_io_error("udp buffer allocation",
                    io_error_callback_, io_error_callback_hint_);
                
                res = core::no_memory;
            }
        }
        
        break;
        
    case proto_tcp:
        
        {
            std::size_t send_buf_size = static_cast<std::size_t>(tcp_sndbuf(fd_.tcp));
            
            std::size_t bytes_to_send;
            if (send_buf_size < size)
            {
                bytes_to_send = send_buf_size;
            }
            else
            {
                bytes_to_send = size;
            }
            
            err_t cc = tcp_write(fd_.tcp, buf, bytes_to_send, TCP_WRITE_FLAG_COPY);
            if (cc == ERR_OK)
            {
                cc = tcp_output(fd_.tcp);
                if (cc == ERR_OK)
                {
                    written = bytes_to_send;
                    
                    res = core::ok;
                }
                else
                {
                    handle_io_error_with_code("tcp output", cc,
                        io_error_callback_, io_error_callback_hint_);
                    
                    res = core::io_error;
                }
            }
            else
            {
                handle_io_error_with_code("tcp write", cc,
                    io_error_callback_, io_error_callback_hint_);
                
                res = core::io_error;
            }
        }
        
        break;
    }

    return res;
}
