// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

#include "../listener.h"
#include <yami4-core/allocator.h>
#include "../channel_group.h"
#include <yami4-core/fatal_errors.h>
#include "../io_error_handler.h"
#include "../network_utils.h"
#include "../options.h"
#include <stdio.h>
#include <cstdlib>
#include <cstring>

#include <FreeRTOS.h>
#include <event_groups.h>
#include <lwip/tcp.h>
#include <lwip/udp.h>

using namespace yami;
using namespace details;

namespace // unnamed
{

void udp_listener_receive_notify(void * arg, struct udp_pcb * pcb, struct pbuf * p, const ip_addr_t * addr, u16_t port)
{
    listener * this_listener = static_cast<listener *>(arg);
    
    this_listener->register_udp_event(p, addr, port);
}

err_t tcp_listener_connect_notify(void * arg, struct tcp_pcb * newpcb, err_t err)
{
    if (err == ERR_OK)
    {
        listener * this_listener = static_cast<listener *>(arg);
        
        this_listener->register_tcp_event(newpcb);
    }
    
    return ERR_OK;
}

} // unnamed namespace

void listener::set_lwip_context(EventGroupHandle_t events)
{
    events_ = events;
}

void listener::register_udp_event(struct pbuf * p, const ip_addr_t * addr, u16_t port)
{
    // remember the arriving packet and
    // notify selector about this event, so it can wake up:
    
    if (accepted_pbuf_ != NULL)
    {
        // silently drop older packet
        
        (void)pbuf_free(accepted_pbuf_);
    }
    
    accepted_pbuf_ = p;
    accepted_addr_ = *addr;
    accepted_port_ = port;
    
    (void)xEventGroupSetBits(events_, listener_event);
}

void listener::register_tcp_event(struct tcp_pcb * new_pcb)
{
    // remember the incoming connection and notify selector about this event:
    
    accepted_pcb_ = new_pcb;
    
    (void)xEventGroupSetBits(events_, listener_event);
}

core::result listener::prepare_tcp(const char * address)
{
    protocol_ = proto_tcp;

    ip_address ipa;
    core::result res = parse_address(*alloc_, address, ipa,
        io_error_callback_, io_error_callback_hint_);

    if (res == core::ok)
    {
        bool socket_created = false;
        
        struct tcp_pcb * pcb = tcp_new();
        if (pcb != NULL)
        {
            fd_.tcp = pcb;
            
            socket_created = true;
            
            err_t cc = tcp_bind(fd_.tcp, IP_ANY_TYPE, ipa.port);
            if (cc == ERR_OK)
            {
                const u8_t backlog = static_cast<u8_t>(
                   group_->get_options().tcp_listen_backlog);
                
                struct tcp_pcb * replaced_pcb =
                    tcp_listen_with_backlog_and_err(fd_.tcp, backlog, &cc);
                if (replaced_pcb != NULL)
                {
                    // the original fd_.tcp was deallocated by the listen function
                    // and the new (smaller) control block was allocated to replace it
                    
                    fd_.tcp = replaced_pcb;
                    
                    tcp_arg(fd_.tcp, this);
                    
                    accepted_pcb_ = NULL;
                    
                    tcp_accept(fd_.tcp, &tcp_listener_connect_notify);
                    
                    res = core::ok;
                }
                else
                {
                    handle_io_error_with_code("tcp listen", cc,
                        io_error_callback_, io_error_callback_hint_);
                    
                    res = core::io_error;
                }
            }
            else
            {
                handle_io_error_with_code("tcp listener bind", cc,
                    io_error_callback_, io_error_callback_hint_);
                
                res = core::io_error;
            }
            
            if (res == core::ok)
            {
                // recreate the target based on (possibly)
                // system-assigned values

                // max length of the string representing the target
                // (arbitrary, but should allow for prefix,
                // address and port)
                const std::size_t max_target_length = 30;

                char * buf = static_cast<char *>(
                    alloc_->allocate(max_target_length));

                if (buf != NULL)
                {
                    port_type assigned_port = fd_.tcp->local_port;
                    
                    // the address was provided by user
                    // and resolved to number in the parse phase
                    
                    // note: address is in host byte order
                    u32_t host = ntohl(ipa.host.addr);
                    
                    int a, b, c, d;
                    a = (host >> 24) & 0xff;
                    b = (host >> 16) & 0xff;
                    c = (host >> 8) & 0xff;
                    d = host & 0xff;
                    
                    ::snprintf(buf, max_target_length,
                        "tcp://%d.%d.%d.%d:%d", a, b, c, d, assigned_port);
                    
                    target_ = buf;
                }
                else
                {
                    res = core::no_memory;
                }
            }
        }

        if (res != core::ok)
        {
            if (socket_created)
            {
                (void)tcp_close(fd_.tcp);
            }
        }
    }
    
    return res;
}

core::result listener::prepare_udp(const char * address)
{
    protocol_ = proto_udp;

    accepted_pbuf_ = NULL;
    
    ip_address ipa;
    core::result res = parse_address(*alloc_, address, ipa,
        io_error_callback_, io_error_callback_hint_);

    if (res == core::ok)
    {
        bool socket_created = false;
        
        struct udp_pcb * pcb = udp_new();
        if (pcb != NULL)
        {
            fd_.udp = pcb;
            
            socket_created = true;
            
            err_t cc = udp_bind(fd_.udp, IP_ANY_TYPE, ipa.port);
            if (cc == ERR_OK)
            {
                udp_recv(fd_.udp, &udp_listener_receive_notify, this);
                
                // recreate the target based on (possibly)
                // system-assigned values

                // max length of the string representing the target
                // (arbitrary, but should allow for prefix,
                // address and port)
                const std::size_t max_target_length = 30;

                char * buf = static_cast<char *>(
                    alloc_->allocate(max_target_length));

                if (buf != NULL)
                {
                    port_type assigned_port = fd_.udp->local_port;
                    
                    // the address was provided by user
                    // and resolved to number in the parse phase
                    
                    // note: address is in host byte order
                    u32_t host = ntohl(ipa.host.addr);
                    
                    int a, b, c, d;
                    a = (host >> 24) & 0xff;
                    b = (host >> 16) & 0xff;
                    c = (host >> 8) & 0xff;
                    d = host & 0xff;
                    
                    ::snprintf(buf, max_target_length,
                        "udp://%d.%d.%d.%d:%d", a, b, c, d, assigned_port);
                    
                    target_ = buf;
                }
                else
                {
                    res = core::no_memory;
                }
            }
            else
            {
                handle_io_error_with_code("udp listener bind", cc,
                    io_error_callback_, io_error_callback_hint_);

                res = core::io_error;
            }
        }
        else
        {
            handle_io_error("create udp listener socket",
                io_error_callback_, io_error_callback_hint_);

            res = core::io_error;
        }

        if (res == core::ok)
        {
            // allocate udp frame buffer of appropriate size

            frame_buffer_size_ = group_->get_options().udp_frame_size;

            frame_buffer_ = static_cast<char *>(
                alloc_->allocate(frame_buffer_size_));
            if (frame_buffer_ == NULL)
            {
                res = core::no_memory;
            }
        }

        if (res != core::ok)
        {
            if (socket_created)
            {
                udp_remove(fd_.udp);
            }

            if (target_ != NULL)
            {
                alloc_->deallocate(target_);
            }

            if (frame_buffer_ != NULL)
            {
                alloc_->deallocate(frame_buffer_);
            }
        }
    }
    
    return res;
}

void listener::close_resource()
{
    if (protocol_ == proto_udp)
    {
        if (accepted_pbuf_ != NULL)
        {
            (void)pbuf_free(accepted_pbuf_);
        }
        
        udp_remove(fd_.udp);
    }
    else if (protocol_ == proto_tcp)
    {
        (void)tcp_close(fd_.tcp);
    }
}

// synchronized by caller
core::result listener::do_some_work()
{
    // accept new connection from the listening socket
    // assume that the listening socket is already ready for "reading"
    // (otherwise this operation can block)

    core::result res;

    switch (protocol_)
    {
    case proto_tcp:
        res = accept_tcp();
        break;
    case proto_udp:
        res = accept_udp();
        break;

    default:
        // the file protocol is impossible here
        fatal_failure(__FILE__, __LINE__);

        // unreachable, to please the compiler
        res = core::unexpected_value;
    }

    return res;
}

core::result listener::accept_tcp()
{
    core::result res;
    
    if (accepted_pcb_)
    {
        // in host by order:
        u32_t remote_address = ntohl(accepted_pcb_->remote_ip.addr);
        
        u16_t remote_port = accepted_pcb_->remote_port;
        
        // target created for the channel will have the form:
        // "tcp://xxx.xxx.xxx.xxx:yyyyyyy"
        const std::size_t target_size = 30;
        char * target = static_cast<char *>(
            alloc_->allocate(target_size));
        if (target != NULL)
        {
            // note: address is in host byte order
            int a, b, c, d;
            a = (remote_address >> 24) & 0xff;
            b = (remote_address >> 16) & 0xff;
            c = (remote_address >> 8) & 0xff;
            d = remote_address & 0xff;
            
            ::snprintf(target, target_size, "tcp://%d.%d.%d.%d:%d",
                a, b, c, d, remote_port);
            
            const std::size_t preferred_frame_size =
                group_->get_options().tcp_frame_size;
            
            core::channel_descriptor new_descriptor;
            channel * new_channel;
            
            // the accepted socket is passed to the channel group
            // and the ownership is transferred there
            
            io_descriptor_type accepted_fd_;
            accepted_fd_.tcp = accepted_pcb_;
            
            res = group_->add_existing(target, accepted_fd_, proto_tcp,
                preferred_frame_size, new_descriptor, new_channel);
            
            if (res == core::ok)
            {
                new_channel->post_init_accepted_channel();
                
                // at this point the new channel is already seen
                // in the proper state
                
                // this channel was freshly created
                // -> call the connection hook
                
                if (connection_hook_ != NULL)
                {
                    mtx_->unlock();
                    
                    std::size_t index;
                    std::size_t sequence_number;
                    new_descriptor.get_details(index, sequence_number);
                    
                    try
                    {
                        connection_hook_(connection_hook_hint_,
                            target, index, sequence_number);
                    }
                    catch (...)
                    {
                        // ignore exceptions from user code
                    }
                    
                    mtx_->lock();
                }
            }
        }
        else
        {
            res = core::no_memory;
        }
    }
    else
    {
        // defensively, allow hazards with selector:
        
        res = core::ok;
    }
    
    return res;
}

// synchronized by caller
core::result listener::accept_udp()
{
    // "Accepting" at the UDP listener is an artificial term
    // that really relates to receiving of regular datagram
    // and not to establishing any connection.
    // The datagram is already received by the notification function,
    // so here it is placed in the appropriate channel
    // according to the sender address information.
    
    core::result res = core::io_error;
    
    if (accepted_pbuf_ != NULL)
    {
        u16_t bytes_copied = pbuf_copy_partial(accepted_pbuf_, frame_buffer_, frame_buffer_size_, 0ul);
        if (bytes_copied != 0ul)
        {
            // target created for the channel will have the form:
            // "udp://xxx.xxx.xxx.xxx:yyyyyyy"
            const std::size_t target_size = 30;
            char target[target_size];
            
            // in host by order:
            u32_t address = ntohl(accepted_addr_.addr);
            
            int a, b, c, d;
            a = (address >> 24) & 0xff;
            b = (address >> 16) & 0xff;
            c = (address >> 8) & 0xff;
            d = address & 0xff;
            
            ::snprintf(target, target_size, "udp://%d.%d.%d.%d:%d",
                a, b, c, d, accepted_port_);

            core::channel_descriptor descriptor;

            // make sure the channel for the client exists
            // (create if it is a new channel)

            bool created_new_channel;
            res = group_->open(target, descriptor,
                created_new_channel, false);

            // at this point the new channel is already seen
            // in the proper state

            if (created_new_channel)
            {
                // this channel was freshly created
                // -> call the connection hook

                if (res == core::ok && connection_hook_ != NULL)
                {
                    mtx_->unlock();

                    std::size_t index;
                    std::size_t sequence_number;
                    descriptor.get_details(index, sequence_number);

                    try
                    {
                        connection_hook_(connection_hook_hint_,
                            target, index, sequence_number);
                    }
                    catch (...)
                    {
                        // ignore exceptions from user code
                    }

                    mtx_->lock();
                }
            }

            if (res == core::ok)
            {
                // inject the received datagram into the channel

                res = group_->process_complete_incoming_frame(
                    descriptor, frame_buffer_, bytes_copied);
            }
        }
        else
        {
            // silently drop packets that do not fit in the frame buffer
        }
    }
    else
    {
        // defensive against possible hazards with selector,
        // it is OK to attempt reading a packet that does not exist
        
        res = core::ok;
    }
    
    (void)pbuf_free(accepted_pbuf_);
    accepted_pbuf_ = NULL;

    return res;
}
