// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

#include "selector.h"
#include <yami4-core/allocator.h>
#include "../channel.h"
#include "../io_error_handler.h"
#include "../listener.h"

#include <cstring>
#include <errno.h>

using namespace yami;
using namespace details;

core::result selector::init(allocator & alloc)
{
    io_error_callback_ = NULL;

    alloc_ = &alloc;

    last_flags_ = 0ul;
    pending_listener_event_ = false;
    pending_channel_input_event_ = false;
    pending_channel_output_event_ = false;
    
    core::result res;
    
    events_ = xEventGroupCreate();
    if (events_ != NULL)
    {
        res = core::ok;
    }
    else
    {
        handle_io_error("selector flags create",
            io_error_callback_, io_error_callback_hint_);
        
        res = core::io_error;
    }

    return res;
}

void selector::install_io_error_logger(
    core::io_error_function io_error_callback,
    void * io_error_callback_hint)
{
    io_error_callback_ = io_error_callback;
    io_error_callback_hint_ = io_error_callback_hint;
}

void selector::clean()
{
    vEventGroupDelete(events_);
}

void selector::reset()
{
    // actually, events are "consumed" by the wait operation
}

core::result selector::add_channel(channel & ch,
    bool allow_outgoing_traffic, bool allow_incoming_traffic)
{
    io_direction direction = ch.get_io_direction();
    protocol proto = ch.get_protocol();
    
    if ((direction == output) || (direction == inout))
    {
        // output operation is always possible on UDP and TCP
        // and there are no separate notifications to signal this,
        // so the appropriate "event" needs to be synthesised
        
        (void)xEventGroupSetBits(events_, channel_output_event);
    }
    else
    {
        // otherwise (input-only channels),
        // the channel notifies the events object directly
        // and asynchronously to the selector actions
    }
    
    // synthesise the pending events, if there are any
    
    if (allow_incoming_traffic && pending_channel_input_event_)
    {
        pending_channel_input_event_ = false;
        
        (void)xEventGroupSetBits(events_, channel_input_event);
    }
    
    if (allow_outgoing_traffic && pending_channel_output_event_)
    {
        pending_channel_output_event_ = false;
        
        (void)xEventGroupSetBits(events_, channel_output_event);
    }
    
    allow_outgoing_channel_traffic_ = allow_outgoing_traffic;
    allow_incoming_channel_traffic_ = allow_incoming_traffic;

    return core::ok;
}

core::result selector::add_listener(listener & lst, bool allow_incoming_traffic)
{
    // listeners operate directly on the event group,
    // so there is no need to "add" them here
    
    // however, if there is a pending listener event and
    // the incoming traffic is allowed, then that pending event
    // should be immediately restored
    
    if (allow_incoming_traffic && pending_listener_event_)
    {
        pending_listener_event_ = false;
        
        (void)xEventGroupSetBits(events_, listener_event);
    }
    
    allow_incoming_listener_traffic_ = allow_incoming_traffic;
    
    return core::ok;
}

void selector::get_channel_usage(std::size_t & max_allowed, std::size_t & used)
{
    max_allowed = 0;
    used = 0;
}

core::result selector::wait(std::size_t timeout)
{
    core::result res;
    
    EventBits_t events_of_interest = selector_interrupt_event |
        (allow_incoming_listener_traffic_ ? listener_event : 0) |
        channel_input_event |
        channel_output_event;
    
    last_flags_ = xEventGroupWaitBits(events_, events_of_interest, pdTRUE, pdFALSE, timeout);
    if (last_flags_ != 0ul)
    {
        res = core::ok;
    }
    else
    {
        res = core::timed_out;
    }
    
    return res;
}

bool selector::is_channel_ready(
    const channel & ch, io_direction & direction) const
{
    // attempt to process incoming and outgoing packets for all channels
    
    // without regard to which one actually reported the event:
    
    bool ready_for_reading;
    bool ready_for_writing;
    bool registered_channel_input_event_ = (last_flags_ & channel_input_event) != 0ul;
    bool registered_channel_output_event_ = (last_flags_ & channel_output_event) != 0ul;
    
    if (allow_incoming_channel_traffic_)
    {
        ready_for_reading = registered_channel_input_event_;
    }
    else
    {
        // if the incoming traffic is not allowed
        // (for example because the message queues are full),
        // then the incoming packets should not be processed,
        // but their pending events should be remembered
        
        pending_channel_input_event_ = registered_channel_input_event_;
        
        ready_for_reading = false;
    }
    
    // in the case of LwIP it is possible to have a single output event
    // for the whole set of channels, so channels need to be re-checked
    // for their actual readiness for writing
    
    io_direction actual_direction = ch.get_io_direction();
    if ((actual_direction == output) || (actual_direction == inout))
    {
        if (allow_outgoing_channel_traffic_)
        {
            ready_for_writing = registered_channel_output_event_;
        }
        else
        {
            // if the outgoing traffic is not allowed
            // (for example due to flow control policies at higher level),
            // then the outgoing packets should not be processed,
            // but their pending events should be remembered
            
            pending_channel_output_event_ = registered_channel_output_event_;
            
            ready_for_writing = false;
        }
    }
    else
    {
        ready_for_writing = false;
    }
    
    if (ready_for_reading && ready_for_writing)
    {
        direction = inout;
    }
    else if (ready_for_reading)
    {
        direction = input;
    }
    else if (ready_for_writing)
    {
        direction = output;
    }
    
    return ready_for_reading || ready_for_writing;
}

bool selector::is_listener_ready(const listener & lst) const
{
    // attempt to process incoming packets from all listeners
    // without regard to which one actually reported the event:
    
    bool res;
    bool registered_listener_event_ = (last_flags_ & listener_event) != 0ul;
    
    if (allow_incoming_listener_traffic_)
    {
        res = registered_listener_event_;
    }
    else
    {
        // if the incoming traffic is not allowed
        // (for example because the message queues are full),
        // then the listener should not be processed,
        // but its pending event should be remembered
        
        pending_listener_event_ = registered_listener_event_;
        
        res = false;
    }
    
    return res;
}

core::result selector::wake_up()
{
    // wake the selector up from its current wait state

    (void)xEventGroupSetBits(events_, selector_interrupt_event);
    
    return core::ok;
}
