// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

#include "../listener.h"
#include <yami4-core/allocator.h>
#include "../channel_group.h"
#include <yami4-core/fatal_errors.h>
#include "../io_error_handler.h"
#include "../network_utils.h"
#include "../options.h"
#include "../serialization.h"
#include <stdio.h>
#include <cstdlib>
#include <cstring>

#include <netdb.h>
#include <netinet/in.h>
#include <sys/netmgr.h>
#include <sys/neutrino.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>

using namespace yami;
using namespace details;

namespace // unnamed
{

// artificial counter for disambiguating client Unix names
// so that Unix client connections have different channel names
long long unix_client_id = 0;

// QNX messages are prefixed with the channel id of the default
// listener on the sender side;
// the prefix is added when the message is sent out
// and stripped when the message is received
const std::size_t chid_prefix_size = 4u;

// QNX messaging receiver thread function
void * receiver_thread_function(void * hint)
{
    msg_receiver * msg_rcv = reinterpret_cast<msg_receiver *>(hint);
    
    msg_rcv->do_run();
    
    return NULL;
}

int deserialize_chid_prefix(const char * buf)
{
    int result;
    
    const char * buffers[1] = { buf };
    std::size_t sizes[1] = { chid_prefix_size };
    std::size_t current_buffer = 0u;
    const char * buffer_position = buf;
    
    const core::result dummy =
        get_integer(buffers, sizes, 1u,
        current_buffer, buffer_position,
        result);
    
    return result;
}

} // unnamed namespace

bool msg_receiver::start(channel_group * group,
    char * buf, std::size_t buf_size, pid_t pid, int chid)
{
    bool res;
    
    group_ = group;
    buf_ = buf;
    buf_size_ = buf_size;
    frame_buffer_busy_ = false;
    io_error_occurred_ = false;
    
    pid_ = pid;
    chid_ = chid;
    
    active_ = true;
    
    int cc = pthread_mutex_init(&mtx_, NULL);
    if (cc != EOK)
    {
        fatal_failure(__FILE__, __LINE__);
    }
    
    cc = pthread_cond_init(&cond_, NULL);
    if (cc != EOK)
    {
        fatal_failure(__FILE__, __LINE__);
    }
    
    cc = pthread_create(&receiver_thread_, NULL, receiver_thread_function, this);
    if (cc == EOK)
    {
        res = true;
    }
    else
    {
        cc = pthread_mutex_destroy(&mtx_);
        if (cc != EOK)
        {
            fatal_failure(__FILE__, __LINE__);
        }
        
        cc = pthread_cond_destroy(&mtx_);
        if (cc != EOK)
        {
            fatal_failure(__FILE__, __LINE__);
        }
        
        res = false;
    }
    
    return res;
}

void msg_receiver::stop()
{
    int cc = pthread_mutex_lock(&mtx_);
    if (cc != EOK)
    {
        fatal_failure(__FILE__, __LINE__);
    }
    
    active_ = false;
    
    cc = pthread_mutex_unlock(&mtx_);
    if (cc != EOK)
    {
        fatal_failure(__FILE__, __LINE__);
    }
    
    // the receiver thread might be waiting on the condvar
    // for the buffer to be consumed,
    // or on MsgReceive for the new message to arrive
    // - both states can be notified without blocking,
    // either by condvar signal or by a pulse
    
    cc = pthread_cond_signal(&cond_);
    if (cc != EOK)
    {
        fatal_failure(__FILE__, __LINE__);
    }
    
    int cid = ::ConnectAttach(ND_LOCAL_NODE, pid_, chid_, 0u, 0);
    if (cid >= 0)
    {
        cc = ::MsgSendPulse(cid, -1, _PULSE_CODE_MINAVAIL, 0);
        if (cc == -1)
        {
            fatal_failure(__FILE__, __LINE__);
        }
        
        cc = ::ConnectDetach(cid);
        if (cc == -1)
        {
            fatal_failure(__FILE__, __LINE__);
        }
    }
    else
    {
        fatal_failure(__FILE__, __LINE__);
    }
    
    // wait for the receiver thread to exit and clean up
    
    cc = pthread_join(receiver_thread_, NULL);
    if (cc != EOK)
    {
        fatal_failure(__FILE__, __LINE__);
    }
    
    cc = pthread_mutex_destroy(&mtx_);
    if (cc != EOK)
    {
        fatal_failure(__FILE__, __LINE__);
    }
    
    cc = pthread_cond_destroy(&cond_);
    if (cc != EOK)
    {
        fatal_failure(__FILE__, __LINE__);
    }
}

bool msg_receiver::is_frame_ready() const
{
    bool result;
    
    int cc = pthread_mutex_lock(&mtx_);
    if (cc != EOK)
    {
        fatal_failure(__FILE__, __LINE__);
    }
    
    result = frame_buffer_busy_ || io_error_occurred_;
    
    cc = pthread_mutex_unlock(&mtx_);
    if (cc != EOK)
    {
        fatal_failure(__FILE__, __LINE__);
    }
    
    return result;
}

void msg_receiver::received_details(bool & io_error_occurred,
    std::size_t & msg_size, std::uint32_t & sender_node, pid_t & sender_pid) const
{
    int cc = pthread_mutex_lock(&mtx_);
    if (cc != EOK)
    {
        fatal_failure(__FILE__, __LINE__);
    }
    
    io_error_occurred = io_error_occurred_;
    if (io_error_occurred_ == false)
    {
        msg_size = message_info_.msglen;
        sender_node = message_info_.nd;
        sender_pid = message_info_.pid;
    }
    
    cc = pthread_mutex_unlock(&mtx_);
    if (cc != EOK)
    {
        fatal_failure(__FILE__, __LINE__);
    }
}

void msg_receiver::release_buffer()
{
    int cc = pthread_mutex_lock(&mtx_);
    if (cc != EOK)
    {
        fatal_failure(__FILE__, __LINE__);
    }
    
    frame_buffer_busy_ = false;
    
    cc = pthread_mutex_unlock(&mtx_);
    if (cc != EOK)
    {
        fatal_failure(__FILE__, __LINE__);
    }
    
    cc = pthread_cond_signal(&cond_);
    if (cc != EOK)
    {
        fatal_failure(__FILE__, __LINE__);
    }
}

void msg_receiver::do_run()
{
    bool keep_going = true;
    
    while (keep_going)
    {
        // wait for buffer to be empty or for the stop request
        
        int cc = pthread_mutex_lock(&mtx_);
        if (cc != EOK)
        {
            fatal_failure(__FILE__, __LINE__);
        }
        
        while (active_ && frame_buffer_busy_)
        {
            cc = pthread_cond_wait(&cond_, &mtx_);
            if (cc != EOK)
            {
                fatal_failure(__FILE__, __LINE__);
            }
        }
        
        keep_going = active_;
        
        cc = pthread_mutex_unlock(&mtx_);
        if (cc != EOK)
        {
            fatal_failure(__FILE__, __LINE__);
        }
        
        if (keep_going)
        {
            // frame buffer must have been released,
            // so it is OK to read the next arriving message
            
            int rcvid = ::MsgReceive(chid_, buf_, buf_size_, &message_info_);
            if (rcvid > 0)
            {
                // regular message was received,
                // store it in the frame buffer until it is consumed
                // by the agent's worker thread;
                // interrupt the selector so that this listener can be
                // checked for readiness and processed
                
                if (message_info_.msglen > 4)
                {
                    // To be confirmed:
                    // When the client, after ConnectAttach, gets aborted
                    // without calling ConnectDetach,
                    // then MsgReceive above unblocks with rcvid > 0 and msglen == 4.
                    // This looks like a valid message, but it is not
                    // and should be a pulse instead.
                    //
                    // As a workaround, messages with msglen <= 4 are ignored here.
                    
                    cc = ::MsgReply(rcvid, 0, NULL, 0);
                    if (cc == -1)
                    {
                        fatal_failure(__FILE__, __LINE__);
                    }
                    
                    cc = pthread_mutex_lock(&mtx_);
                    if (cc != EOK)
                    {
                        fatal_failure(__FILE__, __LINE__);
                    }
                    
                    frame_buffer_busy_ = true;
                    
                    cc = pthread_mutex_unlock(&mtx_);
                    if (cc != EOK)
                    {
                        fatal_failure(__FILE__, __LINE__);
                    }
                    
                    group_->interrupt_work_waiter();
                }
                else
                {
                    // ignoring messages with msglen <= 4
                    
                    ::MsgReply(rcvid, 0, NULL, 0);
                }
            }
            else if (rcvid == 0)
            {
                // pulse was received instead of regular message,
                // this allows the thread to verify in the next round
                // whether it is still active
            }
            else
            {
                // error while receiving remote message,
                // report it, without further details,
                // to the main worker thread
                
                io_error_occurred_ = true;
                
                group_->interrupt_work_waiter();
            }
        }
    }
}

core::result listener::prepare_tcp(const char * address)
{
    protocol_ = proto_tcp;

    ip_address ipa;
    core::result res = parse_address(*alloc_, address, ipa,
        io_error_callback_, io_error_callback_hint_);

    if (res == core::ok)
    {
        const int SOCKET_ERROR = -1;

        io_descriptor_type fd = ::socket(AF_INET, SOCK_STREAM, 0);
        if (fd != SOCKET_ERROR)
        {
            res = core::ok;
            if (group_->get_options().tcp_reuseaddr)
            {
                res = set_reuseaddr(fd);
            }
        }
        else
        {
            handle_io_error("create tcp listener socket",
                io_error_callback_, io_error_callback_hint_);

            res = core::io_error;
        }

        if (res == core::ok)
        {
            sockaddr_in local;

            std::memset(&local, 0, sizeof(local));

            local.sin_family = AF_INET;
            local.sin_port = ipa.port;

            if (ipa.host != 0)
            {
                local.sin_addr.s_addr = ipa.host;
            }
            else
            {
                local.sin_addr.s_addr = htonl(INADDR_ANY);
            }

            int cc = ::bind(fd,
                reinterpret_cast<sockaddr *>(&local), sizeof(local));
            if (cc != SOCKET_ERROR)
            {
                const int backlog =
                    group_->get_options().tcp_listen_backlog;

                cc = ::listen(fd, backlog);
                if (cc != SOCKET_ERROR)
                {
                    fd_ = fd;
                }
                else
                {
                    handle_io_error("tcp listen",
                        io_error_callback_, io_error_callback_hint_);

                    res = core::io_error;
                }

                if (res == core::ok)
                {
                    // recreate the target based on (possibly)
                    // system-assigned values

                    // max length of the string representing the target
                    // (arbitrary, but should allow for prefix,
                    // address and port)
                    // note that address alone can be 256 chars
                    const std::size_t max_target_length = 270;

                    char * buf = static_cast<char *>(
                        alloc_->allocate(max_target_length));

                    if (buf != NULL)
                    {
                        socklen_t dummy = sizeof(local);
                        cc = ::getsockname(fd,
                            reinterpret_cast<sockaddr*>(&local), &dummy);
                        if (cc != SOCKET_ERROR)
                        {
                            const int assigned_port = ntohs(local.sin_port);

                            // assigned_addr is in network byte order
                            const int assigned_addr = local.sin_addr.s_addr;

                            if (assigned_addr == 0)
                            {
                                // if assigned address is local ANY,
                                // recreate it from gethostname

                                char this_host[max_target_length];
                                cc = ::gethostname(
                                    this_host, max_target_length);
                                if (cc == 0)
                                {
                                    ::snprintf(buf, max_target_length,
                                        "tcp://%s:%d",
                                        this_host, assigned_port);
                                }
                                else
                                {
                                    handle_io_error(
                                        "tcp listener gethostname",
                                        io_error_callback_,
                                        io_error_callback_hint_);

                                    res = core::io_error;
                                }
                            }
                            else
                            {
                                // the address was provided by user
                                // and resolved to number in the parse phase

                                const unsigned char * addr_bytes =
                                    reinterpret_cast<const unsigned char *>(
                                        &assigned_addr);

                                // note: address is in network byte order
                                ::snprintf(buf, max_target_length,
                                    "tcp://%d.%d.%d.%d:%d",
                                    static_cast<int>(addr_bytes[0]),
                                    static_cast<int>(addr_bytes[1]),
                                    static_cast<int>(addr_bytes[2]),
                                    static_cast<int>(addr_bytes[3]),
                                    assigned_port);
                            }

                            target_ = buf;
                        }
                        else
                        {
                            alloc_->deallocate(buf);

                            handle_io_error("tcp listener getsockname",
                                io_error_callback_, io_error_callback_hint_);

                            res = core::io_error;
                        }
                    }
                    else
                    {
                        res = core::no_memory;
                    }
                }
            }
            else
            {
                handle_io_error("tcp listener bind",
                    io_error_callback_, io_error_callback_hint_);

                res = core::io_error;
            }
        }

        if (res != core::ok)
        {
            if (fd != SOCKET_ERROR)
            {
                ::close(fd);
            }

            if (target_ != NULL)
            {
                alloc_->deallocate(target_);
            }
        }
    }

    return res;
}

core::result listener::prepare_udp(const char * address)
{
    protocol_ = proto_udp;

    ip_address ipa;
    core::result res = parse_address(*alloc_, address, ipa,
        io_error_callback_, io_error_callback_hint_);

    if (res == core::ok)
    {
        const int SOCKET_ERROR = -1;

        io_descriptor_type fd = ::socket(AF_INET, SOCK_DGRAM, 0);
        if (fd != SOCKET_ERROR)
        {
            sockaddr_in local;

            std::memset(&local, 0, sizeof(local));

            local.sin_family = AF_INET;
            local.sin_port = ipa.port;

            if (ipa.host != 0)
            {
                local.sin_addr.s_addr = ipa.host;
            }
            else
            {
                local.sin_addr.s_addr = htonl(INADDR_ANY);
            }

            int cc = ::bind(fd,
                reinterpret_cast<sockaddr *>(&local), sizeof(local));
            if (cc != SOCKET_ERROR)
            {
                fd_ = fd;

                // recreate the target based on (possibly)
                // system-assigned values

                // max length of the string representing the target
                // (arbitrary, but should allow for prefix,
                // address and port)
                // note that address alone can be 256 chars
                const std::size_t max_target_length = 270;

                char * buf = static_cast<char *>(
                    alloc_->allocate(max_target_length));

                if (buf != NULL)
                {
                    socklen_t dummy = sizeof(local);
                    cc = ::getsockname(fd,
                        reinterpret_cast<sockaddr*>(&local), &dummy);
                    if (cc != SOCKET_ERROR)
                    {
                        const int assigned_port = ntohs(local.sin_port);

                        // assigned_addr is in network byte order
                        const int assigned_addr = local.sin_addr.s_addr;

                        if (assigned_addr == 0)
                        {
                            // if assigned address is local ANY,
                            // recreate it from gethostname

                            char this_host[max_target_length];
                            cc = ::gethostname(
                                this_host, max_target_length);
                            if (cc == 0)
                            {
                                ::snprintf(buf, max_target_length,
                                    "udp://%s:%d",
                                    this_host, assigned_port);
                            }
                            else
                            {
                                handle_io_error(
                                    "udp listener gethostname",
                                    io_error_callback_,
                                    io_error_callback_hint_);

                                res = core::io_error;
                            }
                        }
                        else
                        {
                            // the address was provided by user
                            // and resolved to number in the parse phase

                            const unsigned char * addr_bytes =
                                reinterpret_cast<const unsigned char *>(
                                    &assigned_addr);

                            // note: address is in network byte order
                            ::snprintf(buf, max_target_length,
                                "udp://%d.%d.%d.%d:%d",
                                static_cast<int>(addr_bytes[0]),
                                static_cast<int>(addr_bytes[1]),
                                static_cast<int>(addr_bytes[2]),
                                static_cast<int>(addr_bytes[3]),
                                assigned_port);
                        }

                        target_ = buf;
                    }
                    else
                    {
                        alloc_->deallocate(buf);

                        handle_io_error("udp listener getsockname",
                            io_error_callback_, io_error_callback_hint_);

                        res = core::io_error;
                    }
                }
                else
                {
                    res = core::no_memory;
                }
            }
            else
            {
                handle_io_error("udp listener bind",
                    io_error_callback_, io_error_callback_hint_);

                res = core::io_error;
            }
        }
        else
        {
            handle_io_error("create udp listener socket",
                io_error_callback_, io_error_callback_hint_);

            res = core::io_error;
        }

        if (res == core::ok)
        {
            // allocate udp frame buffer of appropriate size

            frame_buffer_size_ = group_->get_options().udp_frame_size;

            frame_buffer_ = static_cast<char *>(
                alloc_->allocate(frame_buffer_size_));
            if (frame_buffer_ == NULL)
            {
                res = core::no_memory;
            }
        }

        if (res != core::ok)
        {
            if (fd != SOCKET_ERROR)
            {
                ::close(fd);
            }

            if (target_ != NULL)
            {
                alloc_->deallocate(target_);
            }

            if (frame_buffer_ != NULL)
            {
                alloc_->deallocate(frame_buffer_);
            }
        }
    }

    return res;
}

core::result listener::prepare_unix(const char * path)
{
    core::result res;

    protocol_ = proto_unix;

    sockaddr_un local;

    const std::size_t max_path_length = sizeof(local.sun_path) - 1;
    const std::size_t path_length = ::strlen(path);
    if (path_length >= max_path_length)
    {
        res = core::unexpected_value;
    }
    else
    {
        const int SOCKET_ERROR = -1;

        (void) ::unlink(path);

        io_descriptor_type fd = ::socket(AF_UNIX, SOCK_STREAM, 0);
        if (fd != SOCKET_ERROR)
        {
            std::memset(&local, 0, sizeof(local));

            local.sun_family = AF_UNIX;
            ::strcpy(local.sun_path, path);

            int cc = ::bind(fd,
                reinterpret_cast<sockaddr *>(&local), sizeof(local));
            if (cc != SOCKET_ERROR)
            {
                const int backlog =
                    group_->get_options().unix_listen_backlog;

                cc = ::listen(fd, backlog);
                if (cc != SOCKET_ERROR)
                {
                    fd_ = fd;

                    // the resolved target is exactly the same
                    // as the required one

                    const char unix_prefix[] = "unix://";
                    const std::size_t max_target_length =
                        path_length + sizeof(unix_prefix);

                    char * buf = static_cast<char *>(
                        alloc_->allocate(max_target_length));
                    if (buf != NULL)
                    {
                        ::snprintf(buf, max_target_length, "unix://%s", path);
                        target_ = buf;

                        res = core::ok;
                    }
                    else
                    {
                        res = core::no_memory;
                    }
                }
                else
                {
                    handle_io_error("unix listen",
                        io_error_callback_, io_error_callback_hint_);

                    res = core::io_error;
                }
            }
            else
            {
                handle_io_error("unix listener bind",
                    io_error_callback_, io_error_callback_hint_);

                res = core::io_error;
            }
        }
        else
        {
            handle_io_error("create unix listener socket",
                io_error_callback_, io_error_callback_hint_);

            res = core::io_error;
        }

        if (res != core::ok)
        {
            ::close(fd);

            if (target_ != NULL)
            {
                alloc_->deallocate(target_);
            }
        }
    }

    return res;
}

core::result listener::prepare_qnx()
{
    protocol_ = proto_qnx;
    
    core::result res;
    
    const pid_t pid = getpid();
    
    const unsigned int flags = 0u;
    const int chid = ::ChannelCreate(flags);
    
    if (chid >= 0)
    {
        // max length of the string representing the target
        // (arbitrary, but should allow for prefix,
        // node if, pid and channel id)
        const std::size_t max_target_length = 20;
        
        char * buf = static_cast<char *>(
            alloc_->allocate(max_target_length));
        if (buf != NULL)
        {
            ::snprintf(buf, max_target_length,
                "qnx://0:%d:%d", static_cast<int>(pid), chid);
            
            channel_id_ = chid;
            target_ = buf;
            
            res = core::ok;
        }
        else
        {
            res = core::no_memory;
        }
    }
    else
    {
        handle_io_error("create qnx channel",
            io_error_callback_, io_error_callback_hint_);
        
        res = core::io_error;
    }
    
    if (res == core::ok)
    {
        // allocate qnx frame buffer of appropriate size

        // the frame used for receiving messages is 4 bytes bigger to allow
        // the prefix containing the default listener channel id of the sender
        
        frame_buffer_size_ =
            group_->get_options().qnx_frame_size + chid_prefix_size;

        frame_buffer_ = static_cast<char *>(
            alloc_->allocate(frame_buffer_size_));
        if (frame_buffer_ != NULL)
        {
            bool cc = msg_receiver_.start(
                group_, frame_buffer_, frame_buffer_size_, pid, chid);
            if (cc)
            {
                group_->set_default_qnx_listening_channel_id(chid);
            }
            else
            {
                res = core::io_error;
            }
        }
        else
        {
            res = core::no_memory;
        }
    }
    
    if (res != core::ok)
    {
        // msg_receiver_ is started as the last step above,
        // so if there was any error, it was not (yet) started
        // and there is no need to stop it
        
        if (chid >= 0)
        {
            ::ChannelDestroy(chid);
        }
        
        if (target_ != NULL)
        {
            alloc_->deallocate(target_);
        }
        
        if (frame_buffer_ != NULL)
        {
            alloc_->deallocate(frame_buffer_);
        }
    }
    
    return res;
}

#ifdef YAMI4_WITH_OPEN_SSL
core::result listener::prepare_tcps(const char * address)
{
    protocol_ = proto_tcps;

    ip_address ipa;
    core::result res = parse_address(*alloc_, address, ipa,
        io_error_callback_, io_error_callback_hint_);

    if (res == core::ok)
    {
        const int SOCKET_ERROR = -1;

        io_descriptor_type fd = ::socket(AF_INET, SOCK_STREAM, 0);
        if (fd != SOCKET_ERROR)
        {
            res = core::ok;
            if (group_->get_options().tcp_reuseaddr)
            {
                res = set_reuseaddr(fd);
            }
        }
        else
        {
            handle_io_error("create tcps listener socket",
                io_error_callback_, io_error_callback_hint_);

            res = core::io_error;
        }

        if (res == core::ok)
        {
            sockaddr_in local;

            std::memset(&local, 0, sizeof(local));

            local.sin_family = AF_INET;
            local.sin_port = ipa.port;

            if (ipa.host != 0)
            {
                local.sin_addr.s_addr = ipa.host;
            }
            else
            {
                local.sin_addr.s_addr = htonl(INADDR_ANY);
            }

            int cc = ::bind(fd,
                reinterpret_cast<sockaddr *>(&local), sizeof(local));
            if (cc != SOCKET_ERROR)
            {
                const int backlog =
                    group_->get_options().tcp_listen_backlog;

                cc = ::listen(fd, backlog);
                if (cc != SOCKET_ERROR)
                {
                    fd_ = fd;
                }
                else
                {
                    handle_io_error("tcps listen",
                        io_error_callback_, io_error_callback_hint_);

                    res = core::io_error;
                }

                if (res == core::ok)
                {
                    // recreate the target based on (possibly)
                    // system-assigned values

                    // max length of the string representing the target
                    // (arbitrary, but should allow for prefix,
                    // address and port)
                    // note that address alone can be 256 chars
                    const std::size_t max_target_length = 270;

                    char * buf = static_cast<char *>(
                        alloc_->allocate(max_target_length));

                    if (buf != NULL)
                    {
                        socklen_t dummy = sizeof(local);
                        cc = ::getsockname(fd,
                            reinterpret_cast<sockaddr*>(&local), &dummy);
                        if (cc != SOCKET_ERROR)
                        {
                            const int assigned_port = ntohs(local.sin_port);

                            // assigned_addr is in network byte order
                            const int assigned_addr = local.sin_addr.s_addr;

                            if (assigned_addr == 0)
                            {
                                // if assigned address is local ANY,
                                // recreate it from gethostname

                                char this_host[max_target_length];
                                cc = ::gethostname(
                                    this_host, max_target_length);
                                if (cc == 0)
                                {
                                    ::snprintf(buf, max_target_length,
                                        "tcps://%s:%d",
                                        this_host, assigned_port);
                                }
                                else
                                {
                                    handle_io_error(
                                        "tcps listener gethostname",
                                        io_error_callback_,
                                        io_error_callback_hint_);

                                    res = core::io_error;
                                }
                            }
                            else
                            {
                                // the address was provided by user
                                // and resolved to number in the parse phase

                                const unsigned char * addr_bytes =
                                    reinterpret_cast<const unsigned char *>(
                                        &assigned_addr);

                                // note: address is in network byte order
                                ::snprintf(buf, max_target_length,
                                    "tcps://%d.%d.%d.%d:%d",
                                    static_cast<int>(addr_bytes[0]),
                                    static_cast<int>(addr_bytes[1]),
                                    static_cast<int>(addr_bytes[2]),
                                    static_cast<int>(addr_bytes[3]),
                                    assigned_port);
                            }

                            target_ = buf;
                        }
                        else
                        {
                            alloc_->deallocate(buf);

                            handle_io_error("tcps listener getsockname",
                                io_error_callback_, io_error_callback_hint_);

                            res = core::io_error;
                        }
                    }
                    else
                    {
                        res = core::no_memory;
                    }
                }
            }
            else
            {
                handle_io_error("tcps listener bind",
                    io_error_callback_, io_error_callback_hint_);

                res = core::io_error;
            }
        }

        if (res != core::ok)
        {
            if (fd != SOCKET_ERROR)
            {
                ::close(fd);
            }

            if (target_ != NULL)
            {
                alloc_->deallocate(target_);
            }
        }
    }

    return res;
}
#endif // YAMI4_WITH_OPEN_SSL

bool listener::is_qnx_receiver_ready() const
{
    return msg_receiver_.is_frame_ready();
}

void listener::close_resource()
{
    if (protocol_ != proto_qnx)
    {
        ::close(fd_);
    }
    else
    {
        msg_receiver_.stop();
        
        ::ChannelDestroy(channel_id_);
    }
}

// synchronized by caller
core::result listener::do_some_work()
{
    // accept new connection from the listening socket
    // assume that the listening socket is already ready for "reading"
    // (otherwise this operation can block)

    core::result res;

    switch (protocol_)
    {
    case proto_tcp:
        res = accept_tcp();
        break;
    case proto_udp:
        res = accept_udp();
        break;
    case proto_unix:
        res = accept_unix();
        break;
    case proto_qnx:
        res = accept_qnx();
        break;
#ifdef YAMI4_WITH_OPEN_SSL
    case proto_tcps:
        res = accept_tcps();
        break;
#endif // YAMI4_WITH_OPEN_SSL

    default:
        // the file protocol is impossible here
        fatal_failure(__FILE__, __LINE__);

        // unreachable, to please the compiler
        res = core::unexpected_value;
    }

    return res;
}

core::result listener::accept_tcp()
{
    core::result res;

    sockaddr_in client_addr;
    socklen_t client_addr_size = sizeof(client_addr);

    const int SOCKET_ERROR = -1;

    io_descriptor_type new_fd = ::accept(fd_,
        reinterpret_cast<sockaddr *>(&client_addr), &client_addr_size);
    if (new_fd != SOCKET_ERROR)
    {
        res = core::ok;

        if (group_->get_options().tcp_nonblocking)
        {
            res = set_nonblocking(new_fd);
        }

        if (res == core::ok)
        {
            if (group_->get_options().tcp_nodelay)
            {
                res = set_nodelay(new_fd);
            }
        }

        if (res == core::ok)
        {
            if (group_->get_options().tcp_keepalive)
            {
                res = set_keepalive(new_fd);
            }
        }

        if (res == core::ok)
        {
            // extract the client address and create new channel
            // with that address as the target

            // target created for the new channel will have the form:
            // "tcp://xxx.xxx.xxx.xxx:yyyyyyy"
            const std::size_t target_size = 30;
            char * target = static_cast<char *>(
                alloc_->allocate(target_size));
            if (target != NULL)
            {
                const unsigned char * tmp =
                    reinterpret_cast<unsigned char *>(&client_addr.sin_addr);

                ::snprintf(target, target_size, "tcp://%d.%d.%d.%d:%d",
                    static_cast<int>(tmp[0]),
                    static_cast<int>(tmp[1]),
                    static_cast<int>(tmp[2]),
                    static_cast<int>(tmp[3]),
                    ntohs(client_addr.sin_port));

                const std::size_t preferred_frame_size =
                    group_->get_options().tcp_frame_size;

                core::channel_descriptor new_descriptor;
                channel * dummy_new_channel;

                res = group_->add_existing(target, new_fd, proto_tcp,
                    preferred_frame_size, new_descriptor, dummy_new_channel);

                // at this point the new channel is already seen
                // in the proper state

                if (res == core::ok && connection_hook_ != NULL)
                {
                    mtx_->unlock();

                    std::size_t index;
                    std::size_t sequence_number;
                    new_descriptor.get_details(index, sequence_number);

                    try
                    {
                        connection_hook_(connection_hook_hint_,
                            target, index, sequence_number);
                    }
                    catch (...)
                    {
                        // ignore exceptions from user code
                    }

                    mtx_->lock();
                }

                if (res != core::ok)
                {
                    alloc_->deallocate(target);
                }
            }
            else
            {
                res = core::no_memory;
            }
        }

        if (res != core::ok)
        {
            ::close(new_fd);
        }
    }
    else
    {
        handle_io_error("tcp listener accept",
            io_error_callback_, io_error_callback_hint_);

        res = core::io_error;
    }

    return res;
}

// synchronized by caller
core::result listener::accept_udp()
{
    // "Accepting" at the UDP listener is an artificial term
    // that really relates to receiving of regular datagram
    // and not to establishing any connection.
    // The datagram is received and placed in the appropriate channel
    // according to the sender address information.

    core::result res;

    sockaddr_in client_addr;
    socklen_t client_addr_size = sizeof(client_addr);

    const int SOCKET_ERROR = -1;

    const ssize_t readn =
        ::recvfrom(fd_, frame_buffer_, frame_buffer_size_, 0,
            reinterpret_cast<sockaddr *>(&client_addr), &client_addr_size);
    if (readn != SOCKET_ERROR)
    {
        // extract the client address and make sure that a channel
        // with that address as the target exists

        // target created for the channel will have the form:
        // "udp://xxx.xxx.xxx.xxx:yyyyyyy"
        const std::size_t target_size = 30;
        char * target = static_cast<char *>(
            alloc_->allocate(target_size));
        if (target != NULL)
        {
            const unsigned char * tmp =
                reinterpret_cast<unsigned char *>(&client_addr.sin_addr);

            ::snprintf(target, target_size, "udp://%d.%d.%d.%d:%d",
                static_cast<int>(tmp[0]),
                static_cast<int>(tmp[1]),
                static_cast<int>(tmp[2]),
                static_cast<int>(tmp[3]),
                ntohs(client_addr.sin_port));

            core::channel_descriptor descriptor;

            // make sure the channel for the client exists
            // (create if it is a new channel)

            bool created_new_channel;
            res = group_->open(target, descriptor,
                created_new_channel, false);

            alloc_->deallocate(target);

            // at this point the new channel is already seen
            // in the proper state

            if (created_new_channel)
            {
                // this channel was freshly created
                // -> call the connection hook

                if (res == core::ok && connection_hook_ != NULL)
                {
                    mtx_->unlock();

                    std::size_t index;
                    std::size_t sequence_number;
                    descriptor.get_details(index, sequence_number);

                    try
                    {
                        connection_hook_(connection_hook_hint_,
                            target, index, sequence_number);
                    }
                    catch (...)
                    {
                        // ignore exceptions from user code
                    }

                    mtx_->lock();
                }
            }

            if (res == core::ok)
            {
                // inject the received datagram into the channel

                res = group_->process_complete_incoming_frame(
                    descriptor, frame_buffer_, readn);
            }
        }
        else
        {
            res = core::no_memory;
        }
    }
    else
    {
        handle_io_error("udp listener receive",
            io_error_callback_, io_error_callback_hint_);

        res = core::io_error;
    }

    return res;
}

core::result listener::accept_unix()
{
    core::result res;

    sockaddr_un client_addr;
    socklen_t client_addr_size = sizeof(client_addr);

    const int SOCKET_ERROR = -1;

    io_descriptor_type new_fd = ::accept(fd_,
        reinterpret_cast<sockaddr *>(&client_addr), &client_addr_size);
    if (new_fd != SOCKET_ERROR)
    {
        res = core::ok;

        if (group_->get_options().unix_nonblocking)
        {
            res = set_nonblocking(new_fd);
        }

        if (res == core::ok)
        {
            // extract the path and create new channel
            // with that address as the target

            // target created for the new channel will have the form:
            // "unix://path"
            // buf size for target should be at least Unix path + prefix + id
            const std::size_t target_size = sizeof(client_addr.sun_path) + 28;
            char * target = static_cast<char *>(
                alloc_->allocate(target_size));
            if (target != NULL)
            {
                int cc = ::getsockname(new_fd,
                    reinterpret_cast<sockaddr*>(&client_addr),
                    &client_addr_size);
                if (cc != SOCKET_ERROR)
                {
                    ::snprintf(target, target_size, "unix://%s$%lld",
                        client_addr.sun_path, ++unix_client_id);

                    const std::size_t preferred_frame_size =
                        group_->get_options().unix_frame_size;

                    core::channel_descriptor new_descriptor;
                    channel * dummy_new_channel;

                    res = group_->add_existing(target, new_fd, proto_unix,
                        preferred_frame_size, new_descriptor, dummy_new_channel);

                    // at this point the new channel is already seen
                    // in the proper state

                    if (res == core::ok && connection_hook_ != NULL)
                    {
                        mtx_->unlock();

                        std::size_t index;
                        std::size_t sequence_number;
                        new_descriptor.get_details(index, sequence_number);

                        try
                        {
                            connection_hook_(connection_hook_hint_,
                                target, index, sequence_number);
                        }
                        catch (...)
                        {
                            // ignore exceptions from user code
                        }

                        mtx_->lock();
                    }
                }
                else
                {
                    handle_io_error("unix listener getsockname",
                        io_error_callback_, io_error_callback_hint_);

                    res = core::io_error;
                }

                if (res != core::ok)
                {
                    alloc_->deallocate(target);
                }
            }
            else
            {
                res = core::no_memory;
            }
        }

        if (res != core::ok)
        {
            ::close(new_fd);
        }
    }
    else
    {
        handle_io_error("unix listener accept",
            io_error_callback_, io_error_callback_hint_);

        res = core::io_error;
    }

    return res;
}

core::result listener::accept_qnx()
{
    // Similarly to UDP, "accepting" at the QNX listener is an artificial term
    // that really relates to receiving of regular datagram
    // and not to establishing any connection.
    // The datagram is received and placed in the appropriate channel
    // according to the sender address information.
    //
    // Contrary to UDP, the QNX datagram is already captured
    // and stored in the buffer managed by the agent
    // (that was done by a receiver thread),
    // so it can be consumed without any further system calls.
    
    core::result res;
    
    bool io_error_occurred;
    std::size_t rcv_size;
    uint32_t sender_node;
    pid_t sender_pid;
    
    msg_receiver_.received_details(io_error_occurred,
        rcv_size, sender_node, sender_pid);
    if (io_error_occurred == false)
    {
        // extract the client address and make sure that a channel
        // with that address as the target exists
        // target created for the channel will have the form:
        // "qnx://node:pid:chid"
        //
        // the message-processing logic here is similar to that for UDP
        
        const std::size_t target_size = 30;
        char * target = static_cast<char *>(
            alloc_->allocate(target_size));
        if (target != NULL)
        {
            int sender_chid = deserialize_chid_prefix(frame_buffer_);
            
            ::snprintf(target, target_size, "qnx://%u:%d:%d",
                static_cast<unsigned int>(sender_node),
                static_cast<int>(sender_pid),
                sender_chid);
            
            core::channel_descriptor descriptor;
            
            // make sure the channel for the client exists
            // (create if it is a new channel)
            
            bool created_new_channel;
            res = group_->open(target, descriptor,
                created_new_channel, false);
            
            alloc_->deallocate(target);
            
            // at this point the new channel is already seen
            // in the proper state
            
            if (created_new_channel)
            {
                // this channel was freshly created
                // -> call the connection hook
                
                if (res == core::ok && connection_hook_ != NULL)
                {
                    mtx_->unlock();
                    
                    std::size_t index;
                    std::size_t sequence_number;
                    descriptor.get_details(index, sequence_number);
                    
                    try
                    {
                        connection_hook_(connection_hook_hint_,
                            target, index, sequence_number);
                    }
                    catch (...)
                    {
                        // ignore exceptions from user code
                    }
                    
                    mtx_->lock();
                }
            }
            
            if (res == core::ok)
            {
                // inject the received datagram into the channel
                
                // the actual datagram is located after the prefix
                // carrying the default channel id of the sender,
                // and is therefore shorter by the prefix size
                
                res = group_->process_complete_incoming_frame(descriptor,
                    &frame_buffer_[chid_prefix_size],
                    rcv_size - chid_prefix_size);
            }
        }
        else
        {
            res = core::no_memory;
        }
    }
    else
    {
        handle_io_error("qnx listener receive",
            io_error_callback_, io_error_callback_hint_);
        
        res = core::io_error;
    }
    
    msg_receiver_.release_buffer();
    
    return res;
}

#ifdef YAMI4_WITH_OPEN_SSL
core::result listener::accept_tcps()
{
    core::result res;

    sockaddr_in client_addr;
    socklen_t client_addr_size = sizeof(client_addr);

    const int SOCKET_ERROR = -1;

    const io_descriptor_type new_fd = ::accept(fd_,
        reinterpret_cast<sockaddr *>(&client_addr), &client_addr_size);
    if (new_fd != SOCKET_ERROR)
    {
        res = core::ok;

        if (group_->get_options().tcp_nonblocking)
        {
            res = set_nonblocking(new_fd);
        }

        if (res == core::ok)
        {
            if (group_->get_options().tcp_nodelay)
            {
                res = set_nodelay(new_fd);
            }
        }

        if (res == core::ok)
        {
            if (group_->get_options().tcp_keepalive)
            {
                res = set_keepalive(new_fd);
            }
        }

        if (res == core::ok)
        {
            // extract the client address and create new channel
            // with that address as the target

            // target created for the new channel will have the form:
            // "tcp://xxx.xxx.xxx.xxx:yyyyyyy"
            const std::size_t target_size = 30;
            char * target = static_cast<char *>(
                alloc_->allocate(target_size));
            if (target != NULL)
            {
                const unsigned char * tmp =
                    reinterpret_cast<unsigned char *>(&client_addr.sin_addr);

                ::snprintf(target, target_size, "tcps://%d.%d.%d.%d:%d",
                    static_cast<int>(tmp[0]),
                    static_cast<int>(tmp[1]),
                    static_cast<int>(tmp[2]),
                    static_cast<int>(tmp[3]),
                    ntohs(client_addr.sin_port));

                const std::size_t preferred_frame_size =
                    group_->get_options().tcp_frame_size;

                core::channel_descriptor new_descriptor;

                res = group_->add_existing_ssl(target, new_fd, proto_tcps,
                    preferred_frame_size, new_descriptor);

                // at this point the new channel is already seen
                // in the proper state

                if (res == core::ok && connection_hook_ != NULL)
                {
                    mtx_->unlock();

                    std::size_t index;
                    std::size_t sequence_number;
                    new_descriptor.get_details(index, sequence_number);

                    try
                    {
                        connection_hook_(connection_hook_hint_,
                            target, index, sequence_number);
                    }
                    catch (...)
                    {
                        // ignore exceptions from user code
                    }

                    mtx_->lock();
                }

                if (res != core::ok)
                {
                    alloc_->deallocate(target);
                }
            }
            else
            {
                res = core::no_memory;
            }
        }

        if (res != core::ok)
        {
            ::close(new_fd);
        }
    }
    else
    {
        handle_io_error("tcps listener accept",
            io_error_callback_, io_error_callback_hint_);

        res = core::io_error;
    }

    return res;
}
#endif // YAMI4_WITH_OPEN_SSL
