// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

//
// README:
//
// This file is NOT a stand-alone program,
// its main function is at the end of the file
// and is supposed to be called from the actual main() function
// after all hardware and interrupts are initialized.
//
// This test relies on the nx_driver_stm32f4xx network driver,
// which should be modified appropriately for the target environment.
//
// The following addresses and mask are used during tests:
//
#define MY_ADDRESS IP_ADDRESS(192, 168, 11, 112)
#define MY_MASK 0xffffff00ul
#define SERVER_TARGET "tcp://192.168.11.111:12345"

#include <yami4-core/agent.h>
#include <yami4-core/parameters.h>

#include <cassert>
#include <cstdio>
#include <cstring>

#include <nx_api.h>
#include <tx_api.h>

#include "stm32f4xx_hal.h"
#include "nx_driver_stm32f4xx.h"

#define BASE_PACKET_SIZE 1536
#define PACKET_SIZE (BASE_PACKET_SIZE + sizeof(NX_PACKET))
#define PACKET_POOL_SIZE (30 * PACKET_SIZE)
char pool_memory[PACKET_POOL_SIZE + 16];
NX_PACKET_POOL pool;

#define IP_STACK_SIZE 2048
char ip_stack[IP_STACK_SIZE];
NX_IP ip;

#define ARP_SPACE_SIZE 1024
char arp_memory[ARP_SPACE_SIZE];

#define CLIENT_STACK_SIZE 2048
char client_stack[CLIENT_STACK_SIZE];
TX_THREAD client_th;

using namespace yami::core;

bool reply_arrived;

// the dispatch function is used to process the message
// that comes back from the server
// - in this case it is always a reply to the client message
// in more complex system the message headers should be used
// to match related messages
void dispatch_function(
    void * /* hint */,
    const char * /* source */,
    const char * /* header_buffers */ [],
    std::size_t /* header_buffer_sizes */ [],
    std::size_t /* num_of_header_buffers */,
    const char * body_buffers[],
    std::size_t body_buffer_sizes[],
    std::size_t num_of_body_buffers)
{
    // ignore the header, but deserialize the message body

    parameters body;

    result res = body.deserialize(
        body_buffers, body_buffer_sizes, num_of_body_buffers);
    assert(res == ok);

    // extract the value from server

    const char * value;
    std::size_t value_length;

    res = body.get_string("answer", value, value_length);
    assert(res == ok);

    assert(std::strncmp(value, "pong", value_length) == 0);

    reply_arrived = true;
}

void ping_pong(agent & client_agent, channel_descriptor cd)
{
    // send a single message and wait for response

    parameters empty_header;
    parameters body;

    result res = body.set_string("question", "ping");
    assert(res == ok);

    res = client_agent.post(cd, empty_header, body);
    assert(res == ok);

    // process the event loop until the reply is received
    // note: normally the event loop would be processed
    // in a separate thread, but since this program
    // has nothing else to do while waiting for response,
    // the event processing can be performed in the main thread

    const std::size_t time_out = 1000; // 1 second

    reply_arrived = false;
    while (reply_arrived == false)
    {
        res = client_agent.do_some_work(time_out);
        assert(res == ok);
    }

    // at this point the reply has been already processed
    // and validated
}

void client_thread_function(ULONG input)
{
    agent client_agent;

    // create the client agent and open a channel to the given server

    result res = client_agent.init(&dispatch_function, NULL, NULL, NULL);
    assert(res == ok);

    client_agent.set_netx_context(&ip, &pool);

    channel_descriptor cd;
    bool created_new;
    res = client_agent.open(SERVER_TARGET, cd, created_new);
    assert(res == ok);

    const uint32_t start_time = HAL_GetTick();

    const int iterations = 1000;
    for (int i = 0; i != iterations; ++i)
    {
        ping_pong(client_agent, cd);
    }

    const uint32_t stop_time = HAL_GetTick();
    const uint32_t duration = stop_time - start_time;

    const double performance = static_cast<double>(iterations) /
        duration * 100;

    std::printf("performed %d iterations in %d ticks -> %f messages per second\n",
        iterations, (int)duration, performance);

    // the cleanup is automatic in the agent's destructor
}

void tx_application_define(void * ptr)
{
    (void)nx_system_initialize();
    
    (void)nx_packet_pool_create(&pool, const_cast<char *>("Packet pool"), BASE_PACKET_SIZE, &pool_memory[0], PACKET_POOL_SIZE);
    
    (void)nx_ip_create(&ip, const_cast<char *>("IP instance"), MY_ADDRESS, MY_MASK, &pool, &nx_driver_stm32f4xx, &ip_stack[0], IP_STACK_SIZE, 1);
    
    (void)nx_arp_enable(&ip, &arp_memory[0], ARP_SPACE_SIZE);
    
    (void)nx_udp_enable(&ip);
    (void)nx_tcp_enable(&ip);
    
    (void)tx_thread_create(&client_th, const_cast<char *>("Client thread"), &client_thread_function, 0u, &client_stack[0], CLIENT_STACK_SIZE, 3, 3, TX_NO_TIME_SLICE, TX_AUTO_START);
}

extern "C"
void run_client()
{
    tx_kernel_enter();
}
