// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

//
// README:
//
// This file is NOT a stand-alone program,
// its main function is at the end of the file
// and is supposed to be called from the actual main() function
// after all hardware and interrupts are initialized.
// In order to suppress terminal I/O reports, define globally
// (both for the YAMI4 library and for this test file):
// #define YAMI4_NO_TERMINAL_IO
//
// Tests 3 and 4 require manual assistance, as data is supposed to be
// exchanged between the board and some external computer.
// For this reason, test3() and test4() are commented out.
//
// This test relies on the nx_driver_stm32f4xx network driver,
// which should be modified appropriately for the target environment.
//
// The following addresses and mask are used during tests:
//
#define MY_ADDRESS IP_ADDRESS(192, 168, 11, 112)
#define MY_MASK 0xffffff00ul
#define TARGET "udp://192.168.11.111:12345"
#define MY_TARGET "udp://192.168.11.112:12345"

#include <yami4-core/allocator.h>
#include "../../channel.h"
#include "../../channel_group.h"
#include "../../incoming_frame.h"
#include "../../listener.h"
#include "../../options.h"
#include "../../outgoing_frame.h"
#include <cassert>
#include <cstring>
#include <string>
#include <fstream>
#include <vector>
#include <algorithm>
#include <utility>

// selected per platform
#include <mutex.h>

#include <nx_api.h>
#include <tx_api.h>

#include "nx_driver_stm32f4xx.h"

#ifndef YAMI4_NO_TERMINAL_IO
#include <cstdio>
#endif

#define BASE_PACKET_SIZE 1536
#define PACKET_SIZE (BASE_PACKET_SIZE + sizeof(NX_PACKET))
#define PACKET_POOL_SIZE (10 * PACKET_SIZE)
char pool_memory[PACKET_POOL_SIZE + 16];
NX_PACKET_POOL pool;

#define IP_STACK_SIZE 1024
char ip_stack[IP_STACK_SIZE];
NX_IP ip;

#define ARP_SPACE_SIZE 1024
char arp_memory[ARP_SPACE_SIZE];

#define TEST_STACK_SIZE 2048
char test_stack[TEST_STACK_SIZE];
TX_THREAD test_th;

// dummy channel test
void test1()
{
    yami::details::allocator a;
    yami::details::mutex mtx;
    yami::details::channel ch;

    mtx.init();

    yami::details::options default_options;
    default_options.init(NULL);
    yami::core::result res = ch.init(a, mtx, default_options, NULL,
        "abc", NULL, NULL, NULL, NULL, NULL, NULL);
    assert(res == yami::core::bad_protocol);
    
    mtx.clean();
}

// node insertion
// test modified, originally relied on the file:// protocol,
// changed to dummy udp:// target, no actual transmission
void test2()
{
    yami::details::allocator a;
    yami::details::mutex mtx;
    yami::details::channel ch;

    mtx.init();

    yami::details::options default_options;
    default_options.init(NULL);
    ch.set_netx_context(&ip, &pool, NULL);
    yami::core::result res = ch.init(a, mtx, default_options, NULL,
        TARGET, NULL, NULL, NULL, NULL, NULL, NULL);
    assert(res == yami::core::ok);
    
    assert(ch.get_first_outgoing_frame() == NULL);
    assert(ch.get_last_outgoing_frame() == NULL);

    const yami::details::outgoing_frame * f;

    bool first_frame;

    char * buffers[1];
    std::size_t buffer_sizes[1];
    buffers[0] = NULL;
    buffer_sizes[0] = 123;
    res = ch.post(0, buffers, buffer_sizes, 1, first_frame, NULL, NULL);
    assert(res == yami::core::ok);
    assert(first_frame);
    f = ch.get_first_outgoing_frame();
    assert(f->size == 123);
    assert(f->next == NULL);
    assert(ch.get_last_outgoing_frame() == f);

    buffers[0] = NULL;
    buffer_sizes[0] = 456;
    // added at the end:
    res = ch.post(0, buffers, buffer_sizes, 1, first_frame, NULL, NULL);
    assert(res == yami::core::ok);
    assert(first_frame == false);
    f = ch.get_first_outgoing_frame();
    assert(f->size == 123);
    f = f->next;
    assert(f->size == 456);
    assert(f->next == NULL);
    assert(ch.get_last_outgoing_frame() == f);

    buffers[0] = NULL;
    buffer_sizes[0] = 789;
    // added at the end
    res = ch.post(0, buffers, buffer_sizes, 1, first_frame, NULL, NULL);
    assert(res == yami::core::ok);
    assert(first_frame == false);
    f = ch.get_first_outgoing_frame();
    assert(f->size == 123);
    f = f->next;
    assert(f->size == 456);
    f = f->next;
    assert(f->size == 789);
    assert(f->next == NULL);
    assert(ch.get_last_outgoing_frame() == f);

    char * buffers_a1[3];
    std::size_t buffer_sizes_a1[3];
    buffers_a1[0] = NULL;
    buffers_a1[1] = NULL;
    buffers_a1[2] = NULL;
    buffer_sizes_a1[0] = 1;
    buffer_sizes_a1[1] = 2;
    buffer_sizes_a1[2] = 3;
    // added after first
    res = ch.post(3, buffers_a1, buffer_sizes_a1, 3, first_frame, NULL, NULL);
    assert(res == yami::core::ok);
    assert(first_frame == false);
    f = ch.get_first_outgoing_frame();
    assert(f->size == 123);
    f = f->next;
    assert(f->size == 1);
    f = f->next;
    assert(f->size == 2);
    f = f->next;
    assert(f->size == 3);
    f = f->next;
    assert(f->size == 456);
    f = f->next;
    assert(f->size == 789);
    assert(f->next == NULL);
    assert(ch.get_last_outgoing_frame() == f);

    char * buffers_a2[2];
    std::size_t buffer_sizes_a2[2];
    buffers_a2[0] = NULL;
    buffers_a2[1] = NULL;
    buffer_sizes_a2[0] = 4;
    buffer_sizes_a2[1] = 5;
    // added after previous frames
    res = ch.post(1, buffers_a2, buffer_sizes_a2, 2, first_frame, NULL, NULL);
    assert(res == yami::core::ok);
    assert(first_frame == false);
    f = ch.get_first_outgoing_frame();
    assert(f->size == 123);
    f = f->next;
    assert(f->size == 1);
    f = f->next;
    assert(f->size == 2);
    f = f->next;
    assert(f->size == 3);
    f = f->next;
    assert(f->size == 4);
    f = f->next;
    assert(f->size == 5);
    f = f->next;
    assert(f->size == 456);
    f = f->next;
    assert(f->size == 789);
    assert(f->next == NULL);
    assert(ch.get_last_outgoing_frame() == f);

    bool close_me = false;
    res = ch.post_close(2, close_me); // poison pill added after fa1 frames
    assert(res == yami::core::ok);
    assert(close_me == false); // there are output frames, no immediate close
    f = ch.get_first_outgoing_frame();
    assert(f->size == 123);
    f = f->next;
    assert(f->size == 1);
    f = f->next;
    assert(f->size == 2);
    f = f->next;
    assert(f->size == 3);
    f = f->next;
    assert(f->close_flag); // poison pill
    assert(f->next == NULL);
    assert(ch.get_last_outgoing_frame() == f);

    ch.clean();

    yami::details::channel ch2;

    ch2.set_netx_context(&ip, &pool, NULL);
    res = ch2.init(a, mtx, default_options, NULL,
        TARGET, NULL, NULL, NULL, NULL, NULL, NULL);
    assert(res == yami::core::ok);
    
    res = ch2.post_close(2, close_me); // no poison pill added
    assert(res == yami::core::ok);
    assert(close_me); // there are no output frames, immediate close
    assert(ch2.get_first_outgoing_frame() == NULL);
    assert(ch2.get_last_outgoing_frame() == NULL);

    ch2.clean();
}

// file output
// test modified, originally relied on the file:// protocol,
// changed to the udp:// TARGET, there is actual transmission
// that requires data to be captured at the receiver side
void test3()
{
    yami::details::allocator a;
    yami::details::mutex mtx;
    yami::details::channel ch;

    mtx.init();

    yami::details::options default_options;
    default_options.init(NULL);
    ch.set_netx_context(&ip, &pool, NULL);
    yami::core::result res = ch.init(a, mtx, default_options, NULL,
        TARGET, NULL, NULL, NULL, NULL, NULL, NULL);
    assert(res == yami::core::ok);

    bool first_frame;

    const char str0[] = {'H', 'e', 'l', 'l', 'o'};
    const char str1[] = {' '};
    const char str2[] = {'c', 'h', 'a', 'n', 'n', 'e', 'l', '\n'};
    
    char * buffers[3];
    std::size_t buffer_sizes[3];
    buffers[0] = static_cast<char *>(a.allocate(sizeof(str0)));
    std::memcpy(buffers[0], str0, sizeof(str0));
    buffers[1] = static_cast<char *>(a.allocate(sizeof(str1)));
    std::memcpy(buffers[1], str1, sizeof(str1));
    buffers[2] = static_cast<char *>(a.allocate(sizeof(str2)));
    std::memcpy(buffers[2], str2, sizeof(str2));
    buffer_sizes[0] = sizeof(str0);
    buffer_sizes[1] = sizeof(str1);
    buffer_sizes[2] = sizeof(str2);
    
    res = ch.post(0, buffers, buffer_sizes, 3, first_frame, NULL, NULL);
    assert(res == yami::core::ok);
    assert(first_frame);
    
    assert(ch.get_first_outgoing_frame()->size == sizeof(str0));
    bool close_me = false;
    res = ch.do_some_work(yami::details::output, close_me);
    assert(res == yami::core::ok);
    assert(close_me == false);

    assert(ch.get_first_outgoing_frame()->size == sizeof(str1));
    res = ch.do_some_work(yami::details::output, close_me);
    assert(res == yami::core::ok);
    assert(close_me == false);

    assert(ch.get_first_outgoing_frame()->size == sizeof(str2));
    res = ch.do_some_work(yami::details::output, close_me);
    assert(res == yami::core::ok);

    // this was the last frame, but there is no reason to close the channel
    assert(close_me == false);

    ch.clean();
}

namespace // unnamed
{
// used by test4 to test the message dispatch callback

bool dispatch_called = false;
std::vector<std::vector<char> > stored_header_buffers;
std::vector<std::vector<char> > stored_body_buffers;

void incoming_message_dispatch_function(
    void * /* hint */,
    const char * /* source */,
    const char * header_buffers[],
    std::size_t header_buffer_sizes[],
    std::size_t num_of_header_buffers,
    const char * body_buffers[],
    std::size_t body_buffer_sizes[],
    std::size_t num_of_body_buffers)
{
    stored_header_buffers.clear();
    stored_body_buffers.clear();

    for (std::size_t i = 0; i != num_of_header_buffers; ++i)
    {
        std::vector<char> hb;
        hb.assign(header_buffers[i],
            header_buffers[i] + header_buffer_sizes[i]);

        stored_header_buffers.push_back(hb);
    }

    for (std::size_t i = 0; i != num_of_body_buffers; ++i)
    {
        std::vector<char> bb;
        bb.assign(body_buffers[i],
            body_buffers[i] + body_buffer_sizes[i]);

        stored_body_buffers.push_back(bb);
    }

    dispatch_called = true;
}

} // unnamed namespace

// file input - single-frame readout
// test modified, originally relied on the file:// protocol,
// changed to the udp:// MY_TARGET, there is actual transmission
// that requires data to be captured at the receiver side
void test4()
{
    // this should be sent to the board from other host,
    // for example by means of (from POSIX-like shell):
    //
    // $ echo -e -n '\x01\x00\x00\x00\xff\xff\xff\xff\x04\x00\x00\x00\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' | nc -u 192.168.11.112 12345
    //
    //    char content[] =
    //    {
    //        1, 0, 0, 0,             // message id
    //        0xff, 0xff, 0xff, 0xff, // frame number: -1, the only frame
    //        4, 0, 0, 0,             // size of message header
    //        8, 0, 0, 0,             // size of payload
    //        0, 0, 0, 0,             // empty header
    //        0, 0, 0, 0              // empty body
    //    };

    yami::details::allocator a;
    yami::details::channel_group ch_group;
    
    ch_group.set_netx_context(&ip, &pool);
    yami::core::result res = ch_group.init(a, NULL, incoming_message_dispatch_function, NULL, NULL, NULL);
    assert(res == yami::core::ok);
    
    res = ch_group.add_listener(MY_TARGET, NULL, NULL, NULL);
    assert(res == yami::core::ok);
    
    // Note: the UDP data should be sent from external host to this test
    // between add_listener() above and do_some_work() below.
    // The following sleep is necessary to give the NetX thread a chance to process
    // the incoming packet and invoke the listener notification function,
    // otherwise the notification is not called when executed under the debugger.
#ifndef YAMI4_NO_TERMINAL_IO
    std::printf("    waiting 10s for incoming packet...\n");
#endif
    tx_thread_sleep(1000); // 10s
    
    assert(dispatch_called == false);
    
    // read the frame

    res = ch_group.do_some_work(0, false, false);
    assert(res == yami::core::ok);
    
    // message should not be received here,
    // because the incoming traffic was not yet allowed

    assert(dispatch_called == false);

    // allow incoming traffic
    
    res = ch_group.do_some_work(0, false, true);
    assert(res == yami::core::ok);
    
    // this time the message should be already received

    assert(dispatch_called);

    // message header is empty:
    assert(stored_header_buffers.size() == 1);
    std::vector<char> buf = stored_header_buffers[0];
    assert(buf.size() == 4);
    assert(buf[0] == 0);
    assert(buf[1] == 0);
    assert(buf[2] == 0);
    assert(buf[3] == 0);

    // message body is also empty:
    assert(stored_body_buffers.size() == 1);
    buf = stored_body_buffers[0];
    assert(buf.size() == 4);
    assert(buf[0] == 0);
    assert(buf[1] == 0);
    assert(buf[2] == 0);
    assert(buf[3] == 0);

    ch_group.clean(false);
}

// file input - multple-frame, single message readout
// header-body border in the middle of frame
//void test5()
// removed, originally relied on the file:// protocol

// file input - multple-frame, single message readout
// header-body border between frames
//void test6()
// removed, originally relied on the file:// protocol

// file input - multple-frame, single message readout
// header-body border in the middle of frame
//void test7()
// removed, originally relied on the file:// protocol

// all permutations of several frames
//void test8()
// removed, originally relied on the file:// protocol

void do_tests(ULONG input)
{
#ifndef YAMI4_NO_TERMINAL_IO
    std::printf("channel tests\n");
    std::printf("test1 ...\n");
#endif
    
    test1();
    
#ifndef YAMI4_NO_TERMINAL_IO
    std::printf("    OK\n");
    std::printf("test2... \n");
#endif
    
    test2();
    
#ifndef YAMI4_NO_TERMINAL_IO
    std::printf("    OK\n");
    std::printf("test3... \n");
#endif
    
    //test3();
    
#ifndef YAMI4_NO_TERMINAL_IO
    std::printf("    OK\n");
    std::printf("test4... \n");
#endif
    
    //test4();
    
//    test5();
//    test6();
//    test7();
//    test8();
    
#ifndef YAMI4_NO_TERMINAL_IO
    std::printf("    OK\n");
    std::printf("channel tests done\n");
#endif
}

void tx_application_define(void * ptr)
{
    (void)nx_system_initialize();
    
    (void)nx_packet_pool_create(&pool, const_cast<char *>("Packet pool"), BASE_PACKET_SIZE, &pool_memory[0], PACKET_POOL_SIZE);
    
    (void)nx_ip_create(&ip, const_cast<char *>("IP instance"), MY_ADDRESS, MY_MASK, &pool, &nx_driver_stm32f4xx, &ip_stack[0], IP_STACK_SIZE, 1);
    
    (void)nx_arp_enable(&ip, &arp_memory[0], ARP_SPACE_SIZE);
    
    (void)nx_udp_enable(&ip);
    
    (void)tx_thread_create(&test_th, const_cast<char *>("Test thread"), &do_tests, 0u, &test_stack[0], TEST_STACK_SIZE, 3, 3, TX_NO_TIME_SLICE, TX_AUTO_START);
}

extern "C"
void channel_test_main()
{
    // Note: these tests need to be run from threads,
    // so the proper thread environment is first created
    // and the actual tests are executed from do_tests().
    
    tx_kernel_enter();
}
