// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

#include "../channel.h"
#include <yami4-core/allocator.h>
#include "../io_error_handler.h"
#include "../network_utils.h"
#include "../options.h"
#include "mutex.h"
#include <cstring>

#include <tx_api.h>
#include <nx_api.h>

using namespace yami;
using namespace details;

namespace // unnamed
{

void udp_channel_input_notify(NX_UDP_SOCKET * udp_socket)
{
    channel * this_channel = static_cast<channel *>(udp_socket->nx_udp_socket_reserved_ptr);
    
    this_channel->register_input_event();
}

void tcp_channel_input_notify(NX_TCP_SOCKET * tcp_socket)
{
    channel * this_channel = static_cast<channel *>(tcp_socket->nx_tcp_socket_reserved_ptr);

    this_channel->register_input_event();
}

} // unnamed namespace

void channel::set_netx_context(NX_IP * ip, NX_PACKET_POOL * packet_pool, TX_EVENT_FLAGS_GROUP * events)
{
    ip_ = ip;
    packet_pool_ = packet_pool;
    events_ = events;
}

namespace yami
{
namespace details
{

void tcp_disconnect_notify(NX_TCP_SOCKET * tcp_socket)
{
    channel * this_channel = static_cast<channel *>(tcp_socket->nx_tcp_socket_reserved_ptr);
    
    this_channel->close_tcp_connection();
    
    // artificial input event to allow read_bytes to properly process the EOF condition
    this_channel->register_input_event();
}

} // namespace details
} // namespace yami

void channel::close_tcp_connection()
{
    tcp_channel_disconnected_ = true;
}

core::result channel::post_init_accepted_channel()
{
    tcp_channel_disconnected_ = false;
    channel_created_by_listener_ = true;
    
    effective_fd_->tcp.nx_tcp_socket_reserved_ptr = this;
    
    core::result res;
    UINT cc = nx_tcp_socket_receive_notify(&effective_fd_->tcp, &tcp_channel_input_notify);
    if (cc == NX_SUCCESS)
    {
        // synthesize the input event to avoid race with packets that might
        // have arrived before installing the notification callback
        // (it is OK to have surplus event)
        
        register_input_event();
        
        res = core::ok;
    }
    else
    {
        handle_io_error_with_code("tcp accepted channel set input notification", cc,
            io_error_callback_, io_error_callback_hint_);
        
        res = core::io_error;
    }
    
    return res;
}

void channel::register_input_event()
{
    // notify selector about this event, so it can wake up:
    // (both TCP and UDP channel use this callback)
    
    UINT cc = tx_event_flags_set(events_, channel_input_event, TX_OR);
    if (cc != TX_SUCCESS)
    {
        handle_io_error_with_code("channel input event notify", cc,
            io_error_callback_, io_error_callback_hint_);
    }
}

void channel::close_connection()
{
    if (protocol_ == proto_udp)
    {
        (void)nx_udp_socket_unbind(&fd_.udp);
        (void)nx_udp_socket_delete(&fd_.udp);
    }
    else if (protocol_ == proto_tcp)
    {
        if (channel_created_by_listener_)
        {
            // this is a server-side channel,
            // created by the listener,
            // but owned by the channel
            
            (void)nx_tcp_socket_disconnect(&effective_fd_->tcp, NX_WAIT_FOREVER);
            (void)nx_tcp_server_socket_unaccept(&effective_fd_->tcp);
            (void)nx_tcp_socket_delete(&effective_fd_->tcp);
            
            alloc_->deallocate(effective_fd_);
        }
        else
        {
            // this is a client-side channel that owns the I/O descriptor,
            // created by establishing connection to the remote target
            
            (void)nx_tcp_socket_disconnect(&fd_.tcp, NX_WAIT_FOREVER);
            (void)nx_tcp_client_socket_unbind(&fd_.tcp);
            (void)nx_tcp_socket_delete(&fd_.tcp);
        }
    }
}

core::result channel::connect_to_tcp(const char * tcp_address)
{
    bool socket_created = false;
    ip_address ipa;
    core::result res = parse_address(*alloc_, tcp_address, ipa,
        io_error_callback_, io_error_callback_hint_);
    if (res == core::ok)
    {
        UINT cc = nx_tcp_socket_create(
            ip_, &fd_.tcp,
            const_cast<char *>("YAMI4 TCP Client Socket"),
            NX_IP_NORMAL,
            NX_FRAGMENT_OKAY,
            NX_IP_TIME_TO_LIVE,
            200,
            NULL,
            &tcp_disconnect_notify
        );
        
        if (cc == NX_SUCCESS)
        {
            fd_.tcp.nx_tcp_socket_reserved_ptr = this;
            
            socket_created = true;
            
            cc = nx_tcp_client_socket_bind(&fd_.tcp, NX_ANY_PORT, NX_WAIT_FOREVER);
            if (cc == NX_SUCCESS)
            {
                cc = nx_tcp_client_socket_connect(&fd_.tcp, ipa.host, ipa.port, NX_WAIT_FOREVER);
                if (cc == NX_SUCCESS)
                {
                    cc = nx_tcp_socket_receive_notify(&fd_.tcp, &tcp_channel_input_notify);
                    if (cc != NX_SUCCESS)
                    {
                        handle_io_error_with_code("tcp channel set input notification", cc,
                            io_error_callback_, io_error_callback_hint_);
                        
                        res = core::io_error;
                    }
                }
                else
                {
                    handle_io_error_with_code("tcp socket connect", cc,
                        io_error_callback_, io_error_callback_hint_);
                    
                    res = core::io_error;
                }
            }
        }
        else
        {
            handle_io_error_with_code("tcp socket create", cc,
                io_error_callback_, io_error_callback_hint_);
            
            res = core::io_error;
        }
    }
    
    if (res == core::ok)
    {
        // stream-oriented transport, allocate a read buffer
        
        buffer_capacity_ = configuration_options_.tcp_frame_size;
        
        frame_buffer_ = static_cast<char *>(
            alloc_->allocate(buffer_capacity_));
        if (frame_buffer_ == NULL)
        {
            res = core::no_memory;
        }
    }
    
    if (res == core::ok)
    {
        preferred_frame_size_ =
            configuration_options_.tcp_frame_size;
        
        protocol_ = proto_tcp;
        direction_ = inout;
        
        tcp_channel_disconnected_ = false;
    }
    else
    {
        if (socket_created)
        { 
            (void)nx_tcp_socket_disconnect(&fd_.tcp, NX_NO_WAIT);
            (void)nx_tcp_client_socket_unbind(&fd_.tcp);
            (void)nx_tcp_socket_delete(&fd_.tcp);
        }
    }
    
    return res;
}

core::result channel::connect_to_udp(const char * udp_address)
{
    ip_address ipa;
    core::result res = parse_address(*alloc_, udp_address, ipa,
        io_error_callback_, io_error_callback_hint_);
    if (res == core::ok)
    {
        UINT cc = nx_udp_socket_create(
            ip_, &fd_.udp,
            const_cast<char *>("YAMI4 UDP Socket"),
            NX_IP_NORMAL,
            NX_FRAGMENT_OKAY,
            0x80,
            50
        );
        
        if (cc == NX_SUCCESS)
        {
            cc = nx_udp_socket_bind(&fd_.udp, ipa.port, NX_WAIT_FOREVER);
            if (cc == NX_SUCCESS)
            {
                res = core::ok;
                
                if (configuration_options_.udp_checksum_enable)
                {
                    cc = nx_udp_socket_checksum_enable(&fd_.udp);
                    if (cc != NX_SUCCESS)
                    {
                        handle_io_error_with_code("checksum enable on udp socket", cc,
                            io_error_callback_, io_error_callback_hint_);
                        
                        res = core::io_error;
                    }
                }
                else
                {
                    cc = nx_udp_socket_checksum_disable(&fd_.udp);
                    if (cc != NX_SUCCESS)
                    {
                        handle_io_error_with_code("checksum disable on udp socket", cc,
                            io_error_callback_, io_error_callback_hint_);
                        
                        res = core::io_error;
                    }
                }
                
                if (res == core::ok)
                {
                    fd_.udp.nx_udp_socket_reserved_ptr = this;
                    
                    cc = nx_udp_socket_receive_notify(&fd_.udp, &udp_channel_input_notify);
                    if (cc != NX_SUCCESS)
                    {
                        handle_io_error_with_code("udp channel set input notification", cc,
                            io_error_callback_, io_error_callback_hint_);
                        
                        res = core::io_error;
                    }
                }
                
                if (res == core::ok)
                {
                    protocol_ = proto_udp;
                    direction_ = inout;
                    
                    // remember the target address and use it
                    // for each frame when it is sent
                    
                    target_address_ = ipa.host;
                    target_port_ = ipa.port;
                    
                    preferred_frame_size_ =
                        configuration_options_.udp_frame_size;
                    
                    datagram_buffer_ = static_cast<char *>(
                        alloc_->allocate(preferred_frame_size_));
                    if (datagram_buffer_ == NULL)
                    {
                        res = core::no_memory;
                    }
                }
                
                if (res != core::ok)
                {
                    (void)nx_udp_socket_unbind(&fd_.udp);
                    (void)nx_udp_socket_delete(&fd_.udp);
                }
            }
            else
            {
                handle_io_error_with_code("bind udp socket", cc,
                    io_error_callback_, io_error_callback_hint_);
                
                (void)nx_udp_socket_delete(&fd_.udp);
                
                res = core::io_error;
            }
        }
        else
        {
            handle_io_error_with_code("create udp socket", cc,
                io_error_callback_, io_error_callback_hint_);
            
            res = core::io_error;
        }
    }
    
    return res;
}

core::result channel::read_bytes(char * buf, std::size_t size,
    std::size_t & readn)
{
    NX_PACKET * packet_ptr;
    UINT cc;
    core::result res;

    switch (protocol_)
    {
    case proto_udp:

        {
            // the UDP frame has to be read as a single operation
            
            cc = nx_udp_socket_receive(&fd_.udp, &packet_ptr, NX_NO_WAIT);
            if (cc == NX_SUCCESS)
            {
                ULONG data_length;
                cc = nx_packet_length_get(packet_ptr, &data_length);
                if (cc == NX_SUCCESS)
                {
                    if (data_length <= size)
                    {
                        ULONG bytes_copied;
                        cc = nx_packet_data_retrieve(packet_ptr, buf, &bytes_copied);
                        if (cc == NX_SUCCESS)
                        {
                            readn = static_cast<std::size_t>(bytes_copied);
                            
                            res = core::ok;
                        }
                    }
                    else
                    {
                        // silently drop messages that do not fit in the buffer,
                        // but keep processing further packets
                        
                        readn = 0ul;
                        
                        res = core::ok;
                    }
                }
                
                if (cc != NX_SUCCESS)
                {
                    handle_io_error_with_code("udp process incoming packet", cc,
                        io_error_callback_, io_error_callback_hint_);
                    
                    res = core::io_error;
                }
                
                (void)nx_packet_release(packet_ptr);
            }
            else if (cc == NX_NO_PACKET)
            {
                // defensive against possible hazards with selector,
                // it is OK to attempt reading a packet that does not exist
                
                readn = 0ul;
                
                res = core::bad_state;
            }
            else
            {
                handle_io_error_with_code("udp receive", cc,
                    io_error_callback_, io_error_callback_hint_);
                
                res = core::io_error;
            }
        }
        
        break;

    case proto_tcp:
        
        {
            // the frame buffer was allocated at the connection time
            // or when the channel was added to the group,
            // so it can be readily used here
            
            res = core::ok;
            
            if (buffer_consumed_ == buffer_available_)
            {
                // read new portion of data to the buffer
                
                if (tcp_channel_disconnected_)
                {
                    res = core::channel_closed;
                }
                else
                {
                    cc = nx_tcp_socket_receive(&effective_fd_->tcp, &packet_ptr, NX_NO_WAIT);
                    if (cc == NX_SUCCESS)
                    {
                        ULONG data_length;
                        cc = nx_packet_length_get(packet_ptr, &data_length);
                        if (cc == NX_SUCCESS)
                        {
                            if (data_length <= buffer_capacity_)
                            {
                                ULONG bytes_copied;
                                cc = nx_packet_data_retrieve(packet_ptr, frame_buffer_, &bytes_copied);
                                if (cc == NX_SUCCESS)
                                {
                                    buffer_available_ = static_cast<std::size_t>(bytes_copied);
                                    buffer_consumed_ = 0;
                                    
                                    res = core::ok;
                                }
                                else
                                {
                                    handle_io_error_with_code("tcp packet data retrieve", cc,
                                        io_error_callback_, io_error_callback_hint_);
                                    
                                    res = core::io_error;
                                }
                            }
                            else
                            {
                                // drop messages that do not fit in the buffer,
                                // and report it as a communication error
                                
                                readn = 0ul;
                                
                                res = core::io_error;
                            }
                        }
                        
                        if (cc != NX_SUCCESS || res != core::ok)
                        {
                            handle_io_error_with_code("tcp process incoming packet", cc,
                                io_error_callback_, io_error_callback_hint_);
                            
                            res = core::io_error;
                        }
                        
                        (void)nx_packet_release(packet_ptr);
                    }
                    else if (cc == NX_NO_PACKET)
                    {
                        // defensive against possible hazards with selector,
                        // it is OK to attempt reading a packet that does not exist
                        
                        readn = 0ul;
                        
                        res = core::bad_state;
                    }
                    else
                    {
                        handle_io_error_with_code("tcp receive", cc,
                            io_error_callback_, io_error_callback_hint_);
                        
                        res = core::io_error;
                    }
                }
            }
            
            if (res == core::ok)
            {
                // consume data from the buffer
                
                const std::size_t left_in_buffer =
                    buffer_available_ - buffer_consumed_;
                std::size_t to_read;
                if (size <= left_in_buffer)
                {
                    to_read = size;
                }
                else
                {
                    // defensive, it is possible to ask for
                    // more data that fits in tcp_frame_size,
                    // because the remote side might have a different
                    // preferred frame size and that bigger size
                    // was already accepted in the header
                    // -> in this case the data is read piece-wise
                    // from the smaller buffer
                    
                    to_read = left_in_buffer;
                }
                
                std::memcpy(buf, frame_buffer_ + buffer_consumed_, to_read);
                
                buffer_consumed_ += to_read;
                
                readn = to_read;
            }
        }
        
        break;
    }
    
    return res;
}

core::result channel::write_bytes(const char * buf, std::size_t size,
    std::size_t & written)
{
    NX_PACKET * packet_ptr = NULL;
    UINT cc;
    core::result res;

    switch (protocol_)
    {
    case proto_udp:
    
        {
            // the UDP frame is either successfully pushed as a whole
            // or the operation is considered as failing
            
            // allocate packet from pool:
            cc = nx_packet_allocate(packet_pool_, &packet_ptr, NX_IPv4_UDP_PACKET, NX_NO_WAIT);
            if (cc == NX_SUCCESS)
            {
                cc = nx_packet_data_append(packet_ptr,
                    const_cast<char *>(buf), size, packet_pool_, NX_WAIT_FOREVER);
                if (cc == NX_SUCCESS)
                {
                    cc = nx_udp_socket_send(&fd_.udp, packet_ptr, target_address_, target_port_);
                    if (cc == NX_SUCCESS)
                    {
                        written = size;
                        
                        res = core::ok;
                    }
                    else
                    {
                        (void)nx_packet_release(packet_ptr);
                        
                        handle_io_error_with_code("udp send", cc,
                            io_error_callback_, io_error_callback_hint_);
                        
                        res = core::io_error;
                    }
                }
                else
                {
                    (void)nx_packet_release(packet_ptr);
                    
                    handle_io_error_with_code("udp packet prepare", cc,
                        io_error_callback_, io_error_callback_hint_);
                    
                    res = core::io_error;
                }
            }
            else
            {
                handle_io_error_with_code("udp packet allocate", cc,
                    io_error_callback_, io_error_callback_hint_);
                
                res = core::io_error;
            }
        }
        
        break;
        
    case proto_tcp:
        
        {
            // this is a socket stream channel
            
            // allocate packet from pool:
            cc = nx_packet_allocate(packet_pool_, &packet_ptr, NX_IPv4_TCP_PACKET, NX_NO_WAIT);
            if (cc == NX_SUCCESS)
            {
                cc = nx_packet_data_append(packet_ptr,
                   const_cast<char *>(buf), size, packet_pool_, NX_WAIT_FOREVER);
                if (cc == NX_SUCCESS)
                {
                    cc = nx_tcp_socket_send(&effective_fd_->tcp, packet_ptr, NX_WAIT_FOREVER);
                    if (cc == NX_SUCCESS)
                    {
                        written = size;
                        
                        res = core::ok;
                    }
                    else
                    {
                        (void)nx_packet_release(packet_ptr);
                        
                        handle_io_error_with_code("tcp send", cc,
                            io_error_callback_, io_error_callback_hint_);
                        
                        res = core::io_error;
                    }
                }
                else
                {
                    (void)nx_packet_release(packet_ptr);
                    
                    handle_io_error_with_code("tcp packet prepare", cc,
                        io_error_callback_, io_error_callback_hint_);
                    
                    res = core::io_error;
                }
            }
            else
            {
                handle_io_error_with_code("tcp packet allocate", cc,
                    io_error_callback_, io_error_callback_hint_);
                
                res = core::io_error;
            }
        }
        
        break;
    }

    return res;
}
