// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

#include "../listener.h"
#include <yami4-core/allocator.h>
#include "../channel_group.h"
#include <yami4-core/fatal_errors.h>
#include "../io_error_handler.h"
#include "../network_utils.h"
#include "../options.h"
#include <stdio.h>
#include <cstdlib>
#include <cstring>

#include <nx_api.h>
#include <tx_api.h>

using namespace yami;
using namespace details;

namespace // unnamed
{

void udp_listener_receive_notify(NX_UDP_SOCKET * udp_socket)
{
    listener * this_listener = static_cast<listener *>(udp_socket->nx_udp_socket_reserved_ptr);
    
    this_listener->register_event();
}

void tcp_listener_connect_notify(NX_TCP_SOCKET * tcp_socket, UINT port)
{
    listener * this_listener = static_cast<listener *>(tcp_socket->nx_tcp_socket_reserved_ptr);
    
    this_listener->register_event();
}

} // unnamed namespace

namespace yami
{
namespace details
{
    
// actual implementation in tx_channel.cpp
void tcp_disconnect_notify(NX_TCP_SOCKET *);

} // namespace details
} // namespace yami

void listener::set_netx_context(NX_IP * ip, TX_EVENT_FLAGS_GROUP * events)
{
    ip_ = ip;
    events_ = events;
}

void listener::register_event()
{
    // notify selector about this event, so it can wake up:
    // (both UDP and TCP listeners use this callback)
    
    UINT cc = tx_event_flags_set(events_, listener_event, TX_OR);
    if (cc != TX_SUCCESS)
    {
        handle_io_error_with_code("listener event notify", cc,
            io_error_callback_, io_error_callback_hint_);
    }
}

core::result listener::allocate_tcp()
{
    core::result res;
    
    allocated_fd_ = static_cast<io_descriptor_type *>(
        alloc_->allocate(sizeof(io_descriptor_type)));
    if (allocated_fd_ != NULL)
    {
        UINT cc = nx_tcp_socket_create(
            ip_, &allocated_fd_->tcp,
            const_cast<char *>("YAMI4 TCP Listener Socket"),
            NX_IP_NORMAL,
            NX_FRAGMENT_OKAY,
            NX_IP_TIME_TO_LIVE,
            0x80,
            NULL,
            &tcp_disconnect_notify
        );
        
        bool socket_created = false;
        bool socket_listening = false;
        
        if (cc == NX_SUCCESS)
        {
            socket_created = true;
            
            // store relation to the listener object for notification purposes
            
            allocated_fd_->tcp.nx_tcp_socket_reserved_ptr = this;
            
            const int backlog = group_->get_options().tcp_listen_backlog;
            
            cc = nx_tcp_server_socket_listen(ip_, tcp_server_port_, &allocated_fd_->tcp,
                backlog, tcp_listener_connect_notify);
            if (cc == NX_SUCCESS)
            {
                socket_listening = true;
                
                effective_fd_ = allocated_fd_;
                
                res = core::ok;
            }
            else
            {
                handle_io_error_with_code("tcp socket listen", cc,
                    io_error_callback_, io_error_callback_hint_);
                
                res = core::io_error;
            }
        }
        else
        {
            handle_io_error_with_code("create tcp listener socket", cc,
                io_error_callback_, io_error_callback_hint_);

            res = core::io_error;
        }
        
        if (res != core::ok)
        {
            if (socket_listening)
            {
                (void)nx_tcp_server_socket_unlisten(ip_, tcp_server_port_);
            }
            
            if (socket_created)
            {
                (void)nx_tcp_socket_delete(&allocated_fd_->tcp);
            }
            
            alloc_->deallocate(allocated_fd_);
        }
    }
    else
    {
        res = core::no_memory;
    }
    
    return res;
}

void listener::deallocate_tcp()
{
    (void)nx_tcp_server_socket_unlisten(ip_, tcp_server_port_);
    
    if (allocated_fd_ != NULL)
    {
        (void)nx_tcp_socket_delete(&allocated_fd_->tcp);
        
        alloc_->deallocate(allocated_fd_);
        
        allocated_fd_ = NULL;
    }
}

core::result listener::prepare_tcp(const char * address)
{
    protocol_ = proto_tcp;
    
    ip_address ipa;
    core::result res = parse_address(*alloc_, address, ipa,
        io_error_callback_, io_error_callback_hint_);
    
    if (res == core::ok)
    {
        if (ipa.port != 0)
        {
            tcp_server_port_ = ipa.port;
        }
        else
        {
            UINT cc = nx_tcp_free_port_find(ip_, 1, &tcp_server_port_);
            if (cc != NX_SUCCESS)
            {
                handle_io_error_with_code("tcp find free port", cc,
                    io_error_callback_, io_error_callback_hint_);
                
                res = core::io_error;
            }
        }
        
        if (res == core::ok)
        {
            res = allocate_tcp();
            
            if (res == core::ok)
            {
                // recreate the target based on (possibly)
                // system-assigned values
                
                // max length of the string representing the target
                // (arbitrary, but should allow for prefix,
                // address and port)
                const std::size_t max_target_length = 30;
                
                char * buf = static_cast<char *>(
                    alloc_->allocate(max_target_length));
                
                if (buf != NULL)
                {
                    // the address was provided by user
                    // and resolved to number in the parse phase
                    
                    // note: address is in host byte order
                    int a, b, c, d;
                    a = (ipa.host >> 24) & 0xff;
                    b = (ipa.host >> 16) & 0xff;
                    c = (ipa.host >> 8) & 0xff;
                    d = ipa.host & 0xff;
                    
                    ::snprintf(buf, max_target_length,
                        "tcp://%d.%d.%d.%d:%d", a, b, c, d, tcp_server_port_);
                    
                    target_ = buf;
                }
                else
                {
                    deallocate_tcp();
                    
                    res = core::no_memory;
                }
            }
        }
    }
    
    return res;
}

core::result listener::prepare_udp(const char * address)
{
    protocol_ = proto_udp;
    
    ip_address ipa;
    core::result res = parse_address(*alloc_, address, ipa,
        io_error_callback_, io_error_callback_hint_);
    
    if (res == core::ok)
    {
        UINT cc = nx_udp_socket_create(
            ip_, &fd_.udp,
            const_cast<char *>("YAMI4 UDP Listener Socket"),
            NX_IP_NORMAL,
            NX_FRAGMENT_OKAY,
            0x80,
            50
        );
        
        bool socket_created = false;
        if (cc == NX_SUCCESS)
        {
            socket_created = true;
            
            cc = nx_udp_socket_bind(&fd_.udp, ipa.port, NX_WAIT_FOREVER);
            if (cc == NX_SUCCESS)
            {
                if (group_->get_options().udp_checksum_enable)
                {
                    cc = nx_udp_socket_checksum_enable(&fd_.udp);
                    if (cc != NX_SUCCESS)
                    {
                        handle_io_error_with_code("udp listener checksum enable", cc,
                            io_error_callback_, io_error_callback_hint_);
                        
                        res = core::io_error;
                    }
                }
                else
                {
                    cc = nx_udp_socket_checksum_disable(&fd_.udp);
                    if (cc != NX_SUCCESS)
                    {
                        handle_io_error_with_code("udp listener checksum disable", cc,
                            io_error_callback_, io_error_callback_hint_);
                        
                        res = core::io_error;
                    }
                }
                
                if (res == core::ok)
                {
                    effective_fd_ = &fd_;
                    
                    // store relation to the listener object for notification purposes
                    
                    fd_.udp.nx_udp_socket_reserved_ptr = this;
                    
                    cc = nx_udp_socket_receive_notify(&fd_.udp, &udp_listener_receive_notify);
                    if (cc != NX_SUCCESS)
                    {
                        handle_io_error_with_code("udp listener set notification", cc,
                            io_error_callback_, io_error_callback_hint_);
                        
                        res = core::io_error;
                    }
                }
                
                if (res == core::ok)
                {
                    // recreate the target based on (possibly)
                    // system-assigned values
                    
                    // max length of the string representing the target
                    // (arbitrary, but should allow for prefix,
                    // address and port)
                    const std::size_t max_target_length = 30;
                    
                    char * buf = static_cast<char *>(
                        alloc_->allocate(max_target_length));
                    
                    if (buf != NULL)
                    {
                        UINT assigned_port;
                        cc = nx_udp_socket_port_get(&fd_.udp, &assigned_port);
                        if (cc == NX_SUCCESS)
                        {
                            // the address was provided by user
                            // and resolved to number in the parse phase
                            
                            // note: address is in host byte order
                            int a, b, c, d;
                            a = (ipa.host >> 24) & 0xff;
                            b = (ipa.host >> 16) & 0xff;
                            c = (ipa.host >> 8) & 0xff;
                            d = ipa.host & 0xff;
                            
                            ::snprintf(buf, max_target_length,
                                "udp://%d.%d.%d.%d:%d", a, b, c, d, assigned_port);
                            
                            target_ = buf;
                        }
                        else
                        {
                            alloc_->deallocate(buf);
                            
                            handle_io_error_with_code("udp listener port get", cc,
                                io_error_callback_, io_error_callback_hint_);
                            
                            res = core::io_error;
                        }
                    }
                    else
                    {
                        res = core::no_memory;
                    }
                }
            }
            else
            {
                handle_io_error_with_code("udp listener bind", cc,
                    io_error_callback_, io_error_callback_hint_);
                
                res = core::io_error;
            }
        }
        else
        {
            handle_io_error_with_code("create udp listener socket", cc,
                io_error_callback_, io_error_callback_hint_);
            
            res = core::io_error;
        }
        
        if (res == core::ok)
        {
            // allocate udp frame buffer of appropriate size
            
            frame_buffer_size_ = group_->get_options().udp_frame_size;
            
            frame_buffer_ = static_cast<char *>(
                alloc_->allocate(frame_buffer_size_));
            if (frame_buffer_ == NULL)
            {
                res = core::no_memory;
            }
        }
        
        if (res != core::ok)
        {
            if (socket_created)
            {
                (void)nx_udp_socket_unbind(&fd_.udp);
                (void)nx_udp_socket_delete(&fd_.udp);
            }
            
            if (target_ != NULL)
            {
                alloc_->deallocate(target_);
            }
            
            if (frame_buffer_ != NULL)
            {
                alloc_->deallocate(frame_buffer_);
            }
        }
    }
    
    return res;
}

void listener::close_resource()
{
    if (protocol_ == proto_udp)
    {
        (void)nx_udp_socket_unbind(&fd_.udp);
        (void)nx_udp_socket_delete(&fd_.udp);
    }
    else if (protocol_ == proto_tcp)
    {
        deallocate_tcp();
    }
}

// synchronized by caller
core::result listener::do_some_work()
{
    // accept new connection from the listening socket
    // assume that the listening socket is already ready for "reading"
    // (otherwise this operation can block)

    core::result res;

    switch (protocol_)
    {
    case proto_tcp:
        res = accept_tcp();
        break;
    case proto_udp:
        res = accept_udp();
        break;
    }

    return res;
}

core::result listener::accept_tcp()
{
    core::result res;
    
    UINT cc = nx_tcp_server_socket_accept(&effective_fd_->tcp, NX_WAIT_FOREVER);
    if (cc == NX_SUCCESS)
    {
        // create a new socket that will replace the existing one in the listener role
        
        io_descriptor_type * accepted_fd_ = allocated_fd_;
        
        cc = nx_tcp_server_socket_unlisten(ip_, tcp_server_port_);
        if (cc == NX_SUCCESS)
        {
            res = allocate_tcp();
            
            if (res == core::ok)
            {
                ULONG remote_address;
                ULONG remote_port;
                
                cc = nx_tcp_socket_peer_info_get(&accepted_fd_->tcp, &remote_address, &remote_port);
                if (cc == NX_SUCCESS)
                {
                    // target created for the channel will have the form:
                    // "tcp://xxx.xxx.xxx.xxx:yyyyyyy"
                    const std::size_t target_size = 30;
                    char * target = static_cast<char *>(
                        alloc_->allocate(target_size));
                    if (target != NULL)
                    {
                        // note: address is in host byte order
                        int a, b, c, d;
                        a = (remote_address >> 24) & 0xff;
                        b = (remote_address >> 16) & 0xff;
                        c = (remote_address >> 8) & 0xff;
                        d = remote_address & 0xff;
                        
                        ::snprintf(target, target_size, "tcp://%d.%d.%d.%d:%d",
                            a, b, c, d, remote_port);
                        
                        const std::size_t preferred_frame_size =
                            group_->get_options().tcp_frame_size;
                        
                        core::channel_descriptor new_descriptor;
                        channel * new_channel;
                        
                        // the accepted socket is passed to the channel group
                        // and the ownership is transferred there
                        
                        res = group_->add_existing(target, *accepted_fd_, proto_tcp,
                            preferred_frame_size, new_descriptor, new_channel);
                        
                        if (res == core::ok)
                        {
                            // the listening port is passed to the channel
                            // so that the socket can be relistened
                            // when the channel is closed
                            
                            res = new_channel->post_init_accepted_channel();
                            
                            if (res == core::ok)
                            {
                                // at this point the new channel is already seen
                                // in the proper state
                                
                                // this channel was freshly created
                                // -> call the connection hook
                                
                                if (connection_hook_ != NULL)
                                {
                                    mtx_->unlock();
                                    
                                    std::size_t index;
                                    std::size_t sequence_number;
                                    new_descriptor.get_details(index, sequence_number);
                                    
                                    try
                                    {
                                        connection_hook_(connection_hook_hint_,
                                            target, index, sequence_number);
                                    }
                                    catch (...)
                                    {
                                        // ignore exceptions from user code
                                    }
                                    
                                    mtx_->lock();
                                }
                            }
                            else
                            {
                                // unlikely, but close the channel cleanly
                                // in case of initialization errors
                                
                                (void)group_->close(new_descriptor, true, 0);
                            }
                        }
                    }
                    else
                    {
                        res = core::no_memory;
                    }
                }
                else
                {
                    handle_io_error_with_code("tcp peer info get", cc,
                        io_error_callback_, io_error_callback_hint_);
                    
                    res = core::io_error;
                }
                
                if (res != core::ok)
                {
                    // the accepted socket was not adopted by the channel group,
                    // but the new listener socket was already created
                    // - destroy the accepted one
                    
                    (void)nx_tcp_socket_disconnect(&accepted_fd_->tcp, NX_WAIT_FOREVER);
                    (void)nx_tcp_server_socket_unaccept(&accepted_fd_->tcp);
                    (void)nx_tcp_socket_delete(&accepted_fd_->tcp);
                    
                    alloc_->deallocate(accepted_fd_);
                }
            }

            if (res != core::ok)
            {
                deallocate_tcp();
            }
        }
        else
        {
            handle_io_error_with_code("tcp unlisten", cc,
                io_error_callback_, io_error_callback_hint_);
            
            res = core::io_error;
        }
    }
    else
    {
        handle_io_error_with_code("tcp accept", cc,
            io_error_callback_, io_error_callback_hint_);
        
        (void)nx_tcp_server_socket_unaccept(&effective_fd_->tcp);
        
        res = core::io_error;
    }
    
    return res;
}

// synchronized by caller
core::result listener::accept_udp()
{
    // "Accepting" at the UDP listener is an artificial term
    // that really relates to receiving of regular datagram
    // and not to establishing any connection.
    // The datagram is received and placed in the appropriate channel
    // according to the sender address information.
    
    core::result res = core::io_error;
    
    NX_PACKET * packet_ptr;
    
    // keep processing packets if they are available
    while (true)
    {
        UINT cc = nx_udp_socket_receive(&effective_fd_->udp, &packet_ptr, NX_NO_WAIT);
        if (cc == NX_SUCCESS)
        {
            ULONG remote_address;
            UINT remote_port;
            
            cc = nx_udp_packet_info_extract(packet_ptr, &remote_address, NULL, &remote_port, NULL);
            if (cc == NX_SUCCESS)
            {
                ULONG data_length;
                cc = nx_packet_length_get(packet_ptr, &data_length);
                if (cc == NX_SUCCESS)
                {
                    if (data_length <= frame_buffer_size_)
                    {
                        ULONG bytes_copied;
                        cc = nx_packet_data_retrieve(packet_ptr, frame_buffer_, &bytes_copied);
                        if (cc == NX_SUCCESS)
                        {
                            // target created for the channel will have the form:
                            // "udp://xxx.xxx.xxx.xxx:yyyyyyy"
                            const std::size_t target_size = 30;
                            char target[target_size];
                            
                            // note: address is in host byte order
                            int a, b, c, d;
                            a = (remote_address >> 24) & 0xff;
                            b = (remote_address >> 16) & 0xff;
                            c = (remote_address >> 8) & 0xff;
                            d = remote_address & 0xff;
                            
                            ::snprintf(target, target_size, "udp://%d.%d.%d.%d:%d",
                                a, b, c, d, remote_port);

                            core::channel_descriptor descriptor;

                            // make sure the channel for the client exists
                            // (create if it is a new channel)

                            bool created_new_channel;
                            res = group_->open(target, descriptor,
                                created_new_channel, false);

                            // at this point the new channel is already seen
                            // in the proper state

                            if (created_new_channel)
                            {
                                // this channel was freshly created
                                // -> call the connection hook

                                if (res == core::ok && connection_hook_ != NULL)
                                {
                                    mtx_->unlock();

                                    std::size_t index;
                                    std::size_t sequence_number;
                                    descriptor.get_details(index, sequence_number);

                                    try
                                    {
                                        connection_hook_(connection_hook_hint_,
                                            target, index, sequence_number);
                                    }
                                    catch (...)
                                    {
                                        // ignore exceptions from user code
                                    }

                                    mtx_->lock();
                                }
                            }

                            if (res == core::ok)
                            {
                                // inject the received datagram into the channel

                                res = group_->process_complete_incoming_frame(
                                    descriptor, frame_buffer_, bytes_copied);
                            }
                        }
                    }
                    else
                    {
                        // silently drop packets that do not fit in the frame buffer
                    }
                }
            }
            
            if (cc != NX_SUCCESS)
            {
                handle_io_error_with_code("udp listener process packet", cc,
                    io_error_callback_, io_error_callback_hint_);
            }
            
            (void)nx_packet_release(packet_ptr);
        }
        else if (cc == NX_NO_PACKET)
        {
            // defensive against possible hazards with selector,
            // it is OK to attempt reading a packet that does not exist
            
            res = core::ok;
            
            break;
        }
        else
        {
            handle_io_error_with_code("udp listener receive", cc,
                io_error_callback_, io_error_callback_hint_);
            
            res = core::io_error;
            
            break;
        }
    }
    
    return res;
}
