// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

//
// README:
//
// This file is NOT a stand-alone program,
// its main function is at the end of the file
// and is supposed to be a task function, started
// after all hardware and interrupts are initialized.
//
// The following addresses is used for this example:
//
#define MY_TARGET "udp://192.168.11.112:12345"

#include <yami4-core/agent.h>
#include <yami4-core/parameters.h>

#include <cassert>
#include <cstdio>

#include <ti/sysbios/BIOS.h>

#ifdef YAMI4_WITH_TI_SIMPLELINK
#include <ti/sysbios/knl/Task.h>
#include <ti/ndk/inc/socketndk.h>
#endif // YAMI4_WITH_TI_SIMPLELINK

using namespace yami::core;

void dispatch_function(
    void * hint,
    const char * source,
    const char * /* header_buffers */ [],
    std::size_t /* header_buffer_sizes */ [],
    std::size_t /* num_of_header_buffers */,
    const char * body_buffers[],
    std::size_t body_buffer_sizes[],
    std::size_t num_of_body_buffers)
{
    agent * server_agent = static_cast<agent *>(hint);

    // ignore the header, but deserialize the message body

    parameters body;

    result res = body.deserialize(
        body_buffers,
        body_buffer_sizes,
        num_of_body_buffers);
    assert(res == ok);

    // extract the parameters for calculations

    int a;
    int b;

    res = body.get_integer("a", a);
    assert(res == ok);

    res = body.get_integer("b", b);
    assert(res == ok);

    // prepare the answer with results of four calculations
    // the message header is not used here

    parameters empty_header;
    parameters reply_body;

    res = reply_body.set_integer("sum", a + b);
    assert(res == ok);

    res = reply_body.set_integer("difference", a - b);
    assert(res == ok);

    res = reply_body.set_integer("product", a * b);
    assert(res == ok);

    // if the ratio cannot be computed,
    // it is not included in the response
    // the client will interpret that fact properly
    if (b != 0)
    {
        res = reply_body.set_integer("ratio", a / b);
        assert(res == ok);
    }

    res = server_agent->post(
        source, empty_header, reply_body);
    assert(res == ok);

    std::printf("got message with parameters %d and %d, "
        "response has been sent back\n", a, b);
}

extern "C"
Void run_server(UArg arg0, UArg arg1)
{
#ifdef YAMI4_WITH_TI_SIMPLELINK
    (void)fdOpenSession(Task_self());
#endif // YAMI4_WITH_TI_SIMPLELINK

    agent server_agent;

    // create the server agent
    // and activate a listener on the given address

    result res = server_agent.init(
        &dispatch_function, &server_agent, NULL, NULL);
    assert(res == ok);

    const char * resolved_address;
    res = server_agent.add_listener(
        MY_TARGET, NULL, NULL, &resolved_address);
    assert(res == ok);

    std::printf("The server is listening on %s\n", resolved_address);

    // run the event loop
    // normally this loop would be executed by a separate
    // thread, but this server has nothing else to do,
    // so it is possible to process the I/O events here:

    std::printf("waiting for messages...\n");

    const std::size_t time_out = 1000; // 1 second

    while (true)
    {
        res = server_agent.do_some_work(time_out);
        assert(res == ok || res == timed_out);
    }

    // never reaches here
}
