// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

#include "selector.h"
#include <yami4-core/allocator.h>
#include "../channel.h"
#include "../io_error_handler.h"
#include "../listener.h"

#ifdef YAMI4_WITH_TI_SIMPLELINK
#include <cstring>
#include <arpa/inet.h>
#include <netinet/in.h>
#define SOCKET_ERROR -1

#include <ti/sysbios/knl/Task.h>
extern "C" void fdSelectAbort(void * hTask);
#endif // YAMI4_WITH_TI_SIMPLELINK

using namespace yami;
using namespace details;

core::result selector::init(allocator & alloc)
{
    (void)alloc;

    io_error_callback_ = NULL;

    core::result res;

#ifdef YAMI4_WITH_TI_SIMPLELINK
    // for some reason neither pipe nor NDK_pipe seem to be available in SimpleLink libraries,
    // (NDK_pipe is declared in socketndk.h, but no library contains its implementation)
    // so instead of a pipe, the fdSelectAbort(task) is used to interrupt the selector

    task_blocked_on_select_ = NULL;
    wait_aborted_ = false;

    // for some other reason, ::select cannot work with empty descriptor sets,
    // so dummy_fd_ is created only to serve as a placeholder file for the case
    // when it is necessary to block with timeout
    // even though there are no legitimate channels to wait for

    dummy_fd_ = ::socket(AF_INET, SOCK_STREAM, 0);

    res = core::ok;
#else
    // the pipe function expects two arguments of void* (which is also known as SOCKET)
    // as a descriptor type instead if an array of ints, so these have to be translated

    void * pipe_1;
    void * pipe_2;

    int cc = ::pipe(&pipe_1, &pipe_2);
    if (cc == 0)
    {
        interrupt_pipe_write_ = reinterpret_cast<int>(pipe_1);
        interrupt_pipe_read_ = reinterpret_cast<int>(pipe_2);

        res = core::ok;
    }
    else
    {
        res = core::io_error;
    }
#endif // YAMI4_WITH_TI_SIMPLELINK

    return res;
}

void selector::install_io_error_logger(
    core::io_error_function io_error_callback,
    void * io_error_callback_hint)
{
    io_error_callback_ = io_error_callback;
    io_error_callback_hint_ = io_error_callback_hint;
}

void selector::clean()
{
#if !defined(YAMI4_WITH_TI_SIMPLELINK)
    ::close(interrupt_pipe_write_);
    ::close(interrupt_pipe_read_);
#endif // !YAMI4_WITH_TI_SIMPLELINK
}

void selector::reset()
{
    FD_ZERO(&read_set_);
    //FD_ZERO(&write_set_);

#ifdef YAMI4_WITH_TI_SIMPLELINK
    num_of_channels_used_ = 0;
    num_of_descriptors_to_test_ = 0;
    some_channels_willing_to_write_ = false;
#else
    // always add the internal interrupt pipe

    FD_SET(interrupt_pipe_read_, &read_set_);
    
    num_of_channels_used_ = 1; // internal pipe
#endif // !YAMI4_WITH_TI_SIMPLELINK
}

core::result selector::add_channel(channel & ch,
    bool allow_outgoing_traffic, bool allow_incoming_traffic)
{
    io_descriptor_type fd = ch.get_io_descriptor();
    io_direction direction = ch.get_io_direction();

    if ((direction == input || direction == inout) && allow_incoming_traffic)
    {
        FD_SET(fd, &read_set_);

        if (fd > num_of_descriptors_to_test_ - 1)
        {
            num_of_descriptors_to_test_ = fd + 1;
        }
    }

    if ((direction == output || direction == inout) && allow_outgoing_traffic)
    {
        some_channels_willing_to_write_ = true;

        //FD_SET(fd, &write_set_);
    }
    
    ++num_of_channels_used_;

    return core::ok;
}

core::result selector::add_listener(listener & lst, bool allow_incoming_traffic)
{
    core::result res = core::ok;

    if (allow_incoming_traffic)
    {
        io_descriptor_type fd = lst.get_io_descriptor();

        FD_SET(fd, &read_set_);

        ++num_of_channels_used_;

        if (fd > num_of_descriptors_to_test_ - 1)
        {
            num_of_descriptors_to_test_ = fd + 1;
        }
    }

    return res;
}

void selector::get_channel_usage(std::size_t & max_allowed, std::size_t & used)
{
    max_allowed = FD_SETSIZE;
    used = num_of_channels_used_;
}

core::result selector::wait(std::size_t timeout)
{
    core::result res;

    struct timeval tv;

    tv.tv_sec = timeout / 1000;
    tv.tv_usec = 1000 * (timeout % 1000);

    int cc;

#ifdef YAMI4_WITH_TI_SIMPLELINK
    if ((task_blocked_on_select_ == NULL) && wait_aborted_)
    {
        // this can happen when the selector was interrupted (wake_up),
        // for example by means of adding channels or posting messages,
        // before the first wait on select
        // - in that case wake_up does not know which task should be aborted,
        // and wait_aborted_ is the only way to inform the selector
        // that the call to select should be avoided (it would block otherwise)

        // simulate immediate timeout

        cc = 0;
    }
    else
    {
        // this happens when the selector should genuinely wait on select,
        // until there is incoming data to read or timeout, or interrupt (wake_up)

        // note: there is a hazard risk here, if this path is taken for the first time
        // and the wake_up is called at *this* very moment by another task,
        // before task_blocked_on_select_ is set to the task that is about to call select
        // -> this task will block for at most the timeout period
        //    and then resume normal operation for subsequent cycles

        task_blocked_on_select_ = (void *)Task_self();

        if (num_of_channels_used_ != 0)
        {
            if (some_channels_willing_to_write_)
            {
                // simulate immediate readiness
                // (reading will be handled when there is nobody immediately willing to write)

                cc = 1;
            }
            else
            {
                cc = ::select(num_of_descriptors_to_test_, &read_set_, NULL /* &write_set_ */, NULL, &tv);
            }
        }
        else
        {
            // select cannot wait on empty sets, so dummy_fd_ is added
            // as a placeholder so that there is something to wait on

            fd_set dummy_set;
            FD_ZERO(&dummy_set);
            FD_SET(dummy_fd_, &dummy_set);
            cc = ::select(1, &dummy_set, NULL, NULL, &tv);
        }
    }
#else
    cc = ::select(0, &read_set_, NULL /* &write_set_ */, NULL, &tv);
#endif // YAMI4_WITH_TI_SIMPLELINK

    if (cc > 0)
    {
#ifdef YAMI4_WITH_TI_SIMPLELINK
        if (some_channels_willing_to_write_)
        {
            // at least one channel has data ready to be pushed out,
            // and it was rational to avoid blocking on select at all
            // -> read_set_ needs to be cleared, to prevent unnecessary (blocking) reads

            FD_ZERO(&read_set_);

            res = core::ok;
        }
        else
        {
            // the wait was interrupted because there is
            // a regular read event to process

            res = core::ok;
        }
#else
        // the wait might have been interrupted with the internal pipe
        // if that is the case, consume the dummy info

        if (FD_ISSET(interrupt_pipe_read_, &read_set_) != 0)
        {
            bool keep_trying = true;
            while (keep_trying)
            {
                char dummy;
                ssize_t rn =
                    ::recv(interrupt_pipe_read_, &dummy, sizeof(dummy), 0);

                if (rn == -1)
                {
                    // if the attempt to read is abandoned due to signal,
                    // it should be repeated
                    // otherwise it was an error

                    if (errno != EINTR)
                    {
                        handle_io_error("selector pipe read",
                            io_error_callback_, io_error_callback_hint_);

                        res = core::io_error;
                        keep_trying = false;
                    }
                }
                else
                {
                    res = core::ok;
                    keep_trying = false;
                }
            }
        }
        else
        {
            // the wait was interrupted because there is
            // a regular event to process

            res = core::ok;
        }
#endif // YAMI4_WITH_TI_SIMPLELINK
    }
    else if (cc == 0)
    {
#ifdef YAMI4_WITH_TI_SIMPLELINK
        if (wait_aborted_)
        {
            wait_aborted_ = false;

            // the selector pretends to be woken up due to interrupt,
            // but it should not carry on read_set flags,
            // because the sockets are not necessarily ready for reading
            // (they are always "ready" for writing, though)

            FD_ZERO(&read_set_);

            res = core::ok;
        }
        else
        {
            res = core::timed_out;
        }
#else
        res = core::timed_out;
#endif // YAMI4_WITH_TI_SIMPLELINK
    }
    else
    {
        if (cc == EINTR)
        {
            // the wait operation will be repeated in the outer loop
            
            res = core::ok;
        }
        else
        {
            handle_io_error("selector wait",
                io_error_callback_, io_error_callback_hint_);

            res = core::io_error;
        }
    }

    return res;
}

bool selector::is_channel_ready(
    const channel & ch, io_direction & direction) const
{
    io_descriptor_type fd = ch.get_io_descriptor();
    io_direction dir = ch.get_io_direction();

    bool ready_for_reading = false;
    bool ready_for_writing = false;

    if (dir == input || dir == inout)
    {
        if (FD_ISSET(fd, &read_set_) != 0)
        {
            ready_for_reading = true;
        }
    }

    if (dir == output || dir == inout)
    {
        // NDK does not recognize "ready for write",
        // assume sockets are always ready

        /*
        if (FD_ISSET(fd, &write_set_) != 0)
        {
            ready_for_writing = true;
        }
        */

        ready_for_writing = true;
    }

    if (ready_for_reading && ready_for_writing)
    {
        direction = inout;
    }
    else if (ready_for_reading)
    {
        direction = input;
    }
    else if (ready_for_writing)
    {
        direction = output;
    }

    return ready_for_reading || ready_for_writing;
}

bool selector::is_listener_ready(const listener & lst) const
{
    io_descriptor_type fd = lst.get_io_descriptor();
    
    return FD_ISSET(fd, &read_set_) != 0;
}

core::result selector::wake_up()
{
    // wake the selector up from its current wait state

    core::result res = core::ok;

#ifdef YAMI4_WITH_TI_SIMPLELINK
    wait_aborted_ = true;
    if (task_blocked_on_select_ != NULL)
    {
        fdSelectAbort(task_blocked_on_select_);
    }
#else
    bool keep_trying = true;
    while (keep_trying)
    {
        char dummy = '\0';
        ssize_t written =
            ::send(interrupt_pipe_write_, &dummy, sizeof(dummy), 0);

        if (written == -1)
        {
            // if the attempt to write is abandoned due to signal,
            // it should be repeated
            // otherwise it was an error

            if (errno != EINTR)
            {
                handle_io_error("selector pipe write",
                    io_error_callback_, io_error_callback_hint_);

                res = core::io_error;
                keep_trying = false;
            }
        }
        else
        {
            res = core::ok;
            keep_trying = false;
        }
    }
#endif // YAMI4_WITH_TI_SIMPLELINK

    return res;
}
