// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

#include "../channel.h"
#include <yami4-core/allocator.h>
#include "../io_error_handler.h"
#include "../network_utils.h"
#include "../options.h"
#include "mutex.h"
#include <cstring>

#include <errno.h>
#include <sys/socket.h>

#ifdef YAMI4_WITH_TI_SIMPLELINK
#include <netinet/in.h>
#endif // YAMI4_WITH_TI_SIMPLELINK

using namespace yami;
using namespace details;

void channel::close_connection()
{
    ::close(fd_);
}

core::result channel::connect_to_tcp(const char * tcp_address)
{
    ip_address ipa;
    core::result res = parse_address(*alloc_, tcp_address, ipa,
        io_error_callback_, io_error_callback_hint_);
    if (res == core::ok)
    {
        fd_ = ::socket(AF_INET, SOCK_STREAM, 0);
        if (fd_ < 0)
        {
            handle_io_error("create tcp socket",
                io_error_callback_, io_error_callback_hint_);

            res = core::io_error;
        }
    }

    if (res == core::ok)
    {
        protocol_ = proto_tcp;

        struct sockaddr_in servaddr;

        std::memset(&servaddr, 0, sizeof(servaddr));
        servaddr.sin_family = AF_INET;
        servaddr.sin_port = ipa.port;
        std::memcpy(&(servaddr.sin_addr), &ipa.host, sizeof(int));

        // non-blocking socket operations are not supported
        // blocking connection, but with socket bound locally

        struct sockaddr_in thisaddr;
        std::memset(&thisaddr, 0, sizeof(thisaddr));
        thisaddr.sin_family = AF_INET;
        thisaddr.sin_port = 0;
        std::memset(&thisaddr.sin_addr, 0, sizeof(int));

        int cc = bind(fd_,
            reinterpret_cast<sockaddr *>(&thisaddr), sizeof(thisaddr));
        if (cc == 0)
        {
            const int cc = ::connect(fd_,
                (struct sockaddr*)&servaddr, sizeof(servaddr));
            if (cc < 0)
            {
                handle_io_error("blocking tcp connect",
                    io_error_callback_, io_error_callback_hint_);

                res = core::io_error;
            }
        }
        else
        {
            handle_io_error("bind tcp socket",
                io_error_callback_, io_error_callback_hint_);

            res = core::io_error;
        }
    }

    if (res == core::ok)
    {
        if (configuration_options_.tcp_nodelay)
        {
            res = set_nodelay(fd_);
        }
    }

    if (res == core::ok)
    {
        if (configuration_options_.tcp_keepalive)
        {
            res = set_keepalive(fd_);
        }
    }

    if (res == core::ok)
    {
        // stream-oriented transport, allocate a read buffer
        
        buffer_capacity_ = configuration_options_.tcp_frame_size;
        
        frame_buffer_ = static_cast<char *>(
            alloc_->allocate(buffer_capacity_));
        if (frame_buffer_ == NULL)
        {
            res = core::no_memory;
        }
    }
    
    if (res == core::ok)
    {
        preferred_frame_size_ =
            configuration_options_.tcp_frame_size;

        direction_ = inout;
    }
    else
    {
        if (fd_ != empty_io_descriptor)
        { 
            ::close(fd_);
        }
    }

    return res;
}

core::result channel::connect_to_udp(const char * udp_address)
{
    ip_address ipa;
    core::result res = parse_address(*alloc_, udp_address, ipa,
        io_error_callback_, io_error_callback_hint_);
    if (res == core::ok)
    {
        fd_ = ::socket(AF_INET, SOCK_DGRAM, 0);
        if (fd_ < 0)
        {
            handle_io_error("create udp socket",
                io_error_callback_, io_error_callback_hint_);

            res = core::io_error;
        }
        else
        {
            protocol_ = proto_udp;
            direction_ = inout;

            struct sockaddr_in thisaddr;
            std::memset(&thisaddr, 0, sizeof(thisaddr));
            thisaddr.sin_family = AF_INET;
            thisaddr.sin_port = 0;
            std::memset(&thisaddr.sin_addr, 0, sizeof(int));

            int cc = bind(fd_,
                reinterpret_cast<sockaddr *>(&thisaddr), sizeof(thisaddr));
            if (cc == 0)
            {
                // remember the target address and use it
                // for each frame when it is sent

                target_address_size_ = sizeof(sockaddr_in);
                struct sockaddr_in * servaddr =
                    static_cast<sockaddr_in *>(
                        alloc_->allocate(target_address_size_));

                if (servaddr != NULL)
                {
                    std::memset(servaddr, 0, target_address_size_);
                    servaddr->sin_family = AF_INET;
                    servaddr->sin_port = ipa.port;
                    std::memcpy(&(servaddr->sin_addr), &ipa.host, sizeof(int));

                    preferred_frame_size_ =
                        configuration_options_.udp_frame_size;

                    datagram_buffer_ = static_cast<char *>(
                        alloc_->allocate(preferred_frame_size_));
                    if (datagram_buffer_ != NULL)
                    {
                        target_address_ = servaddr;
                    }
                    else
                    {
                        alloc_->deallocate(servaddr);
                        res = core::no_memory;
                    }
                }
                else
                {
                    res = core::no_memory;
                }
            }
            else
            {
                handle_io_error("bind udp socket",
                    io_error_callback_, io_error_callback_hint_);

                res = core::io_error;
            }
        }
    }

    return res;
}

core::result channel::read_bytes(char * buf, std::size_t size,
    std::size_t & readn)
{
    core::result res;

    switch (protocol_)
    {
    case proto_udp:

        {
            // the UDP frame has to be read as a single operation

            struct sockaddr_in remote_address;
            socklen_t address_size = sizeof(remote_address);

            ssize_t rn = ::recvfrom(fd_, buf, size, 0,
                reinterpret_cast<sockaddr *>(&remote_address), &address_size);

            if (rn >= 0)
            {
                readn = static_cast<std::size_t>(rn);
                res = core::ok;
            }
            else
            {
                handle_io_error("udp receive",
                    io_error_callback_, io_error_callback_hint_);

                res = core::io_error;
            }
        }

        break;

    case proto_tcp:

        {
            // the frame buffer was allocated at the connection time
            // or when the channel was added to the group,
            // so it can be readily used here
            
            res = core::ok;
            
            if (buffer_consumed_ == buffer_available_)
            {
                // read new portion of data to the buffer
                
                ssize_t rn = ::recv(fd_, frame_buffer_, buffer_capacity_, 0);
                if (rn == -1)
                {
                    if (errno == EINTR)
                    {
                        readn = 0;
                        res = core::ok;
                    }
                    else if (errno == ECONNRESET)
                    {
                        res = core::channel_closed;
                    }
#ifdef YAMI4_WITH_TI_SIMPLELINK
                    else if (errno == -104) // SLNETERR_BSD_ECONNRESET
                    {
                        res = core::channel_closed;
                    }
#endif // YAMI4_WITH_TI_SIMPLELINK
                    else
                    {
                        handle_io_error("stream read",
                            io_error_callback_, io_error_callback_hint_);
                        
                        res = core::io_error;
                    }
                }
                else
                {
                    if (rn != 0)
                    {
                        buffer_available_ = static_cast<std::size_t>(rn);
                        buffer_consumed_ = 0;
                    }
                    else
                    {
                        res = core::channel_closed;
                    }
                }
            }
            
            if (res == core::ok)
            {
                // consume data from the buffer
                
                const std::size_t left_in_buffer =
                    buffer_available_ - buffer_consumed_;
                std::size_t to_read;
                if (size <= left_in_buffer)
                {
                    to_read = size;
                }
                else
                {
                    // defensive, it is possible to ask for
                    // more data that fits in tcp_frame_size,
                    // because the remote side might have a different
                    // preferred frame size and that bigger size
                    // was already accepted in the header
                    // -> in this case the data is read piece-wise
                    // from the smaller buffer
                    
                    to_read = left_in_buffer;
                }
                
                std::memcpy(buf, frame_buffer_ + buffer_consumed_, to_read);
                
                buffer_consumed_ += to_read;
                
                readn = to_read;
            }
        }

        break;
    }

    return res;
}

core::result channel::write_bytes(const char * buf, std::size_t size,
    std::size_t & written)
{
    core::result res;

    switch (protocol_)
    {
    case proto_udp:
    
        {
            // the UDP frame is either successfully pushed as a whole
            // or the operation is considered as failing

            socklen_t address_size =
                static_cast<socklen_t>(target_address_size_);
            ssize_t wn = ::sendto(fd_, buf, size, 0,
                reinterpret_cast<sockaddr *>(target_address_), address_size);
            if (static_cast<std::size_t>(wn) == size)
            {
                res = core::ok;
            }
            else
            {
                handle_io_error("udp send",
                    io_error_callback_, io_error_callback_hint_);

                res = core::io_error;
            }
        }
        
        break;
        
    case proto_tcp:
    
        {
            // this is a socket stream channel

            ssize_t wn = ::send(fd_, buf, size, 0);
            if (wn == -1)
            {
                if (errno == EINTR ||
                    errno == EAGAIN ||    // Cygwin workaround,
                    errno == EWOULDBLOCK) // these were not seen on other platforms
                {
                    written = 0;
                    res = core::ok;
                }
                else
                {
                    handle_io_error("stream write",
                        io_error_callback_, io_error_callback_hint_);

                    res = core::io_error;
                }
            }
            else
            {
                written = static_cast<std::size_t>(wn);
                res = core::ok;
            }
        }
        
        break;
    }

    return res;
}
