// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

#include "../network_utils.h"
#include <yami4-core/core.h>
#include "../io_error_handler.h"

#include <cstdio>

#include <errno.h>
#include <sys/socket.h>

#ifdef YAMI4_WITH_TI_SIMPLELINK
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#endif // YAMI4_WITH_TI_SIMPLELINK

using namespace yami;
using namespace details;

core::result details::get_host_by_name(const char * hostname, int & host,
    core::io_error_function io_error_callback, void * io_error_callback_hint)
{
    int a, b, c, d;
    core::result res;

    int cc = std::sscanf(hostname, "%d.%d.%d.%d", &a, &b, &c, &d);
    if (cc == 4 &&
        a >= 0 && a <= 255 &&
        b >= 0 && b <= 255 &&
        c >= 0 && c <= 255 &&
        d >= 0 && d <= 255)
    {
        host = htonl(static_cast<int>(
            (static_cast<unsigned int>(a) << 24) +
            (static_cast<unsigned int>(b) << 16) +
            (static_cast<unsigned int>(c) << 8) +
            d));

        res = core::ok;
    }
    else
    {
        handle_io_error("parse IP address",
            io_error_callback, io_error_callback_hint);

        res = core::io_error;
    }

    return res;
}

int details::get_network_port(long port)
{
#ifdef YAMI4_WITH_TI_SIMPLELINK
    return htons(static_cast<uint16_t>(port));
#else
    return htons(static_cast<in_port_t>(port));
#endif // YAMI4_WITH_TI_SIMPLELINK
}

core::result details::set_reuseaddr(io_descriptor_type & fd)
{
    core::result res = core::ok;

#if !defined(YAMI4_WITH_TI_SIMPLELINK)
    const int value = 1;
    const int cc = ::setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
        &value, sizeof(int));
    if (cc != 0)
    {
        res = core::io_error;
    }
#endif // !YAMI4_WITH_TI_SIMPLELINK

    return res;
}

core::result details::set_nodelay(io_descriptor_type & fd)
{
    core::result res;

    const int value = 1;
    const int cc = ::setsockopt(fd, IPPROTO_TCP, TCP_NODELAY,
        &value, sizeof(int));
    if (cc == 0)
    {
        res = core::ok;
    }
    else
    {
        res = core::io_error;
    }

    return res;
}

core::result details::set_keepalive(io_descriptor_type & fd)
{
    core::result res;

    const int value = 1;
    const int cc = ::setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE,
        &value, sizeof(int));
    if (cc == 0)
    {
        res = core::ok;
    }
    else
    {
        res = core::io_error;
    }

    return res;
}
