// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

//
// README:
//
// This file is NOT a stand-alone program,
// its main function is at the end of the file
// and is supposed to be a task function, started
// after all hardware and interrupts are initialized.
//
// The following addresses and mask are used during tests:
//
#define SERVER_TARGET "udp://192.168.11.111:12345"

#include <yami4-cpp/yami.h>

#include <cstdio>

#include <FreeRTOS.h>
#include <task.h>

// worker thread function that is used to process I/O events
// for the client agent
void worker_thread_function(void * argument)
{
    yami::agent * client_agent = reinterpret_cast<yami::agent *>(argument);
    
    client_agent->run_worker();
}

extern "C"
void MX_LWIP_Init(void);

extern "C"
void run_client(void * argument)
{
    MX_LWIP_Init();
    
    try
    {
        // no dispatchers are needed in a pure-client program:
        yami::parameters options;
        options.set_integer(yami::option_names::dispatcher_threads, 0);
        
        yami::agent client_agent(options);

        // start the worker thread
        
        xTaskCreate(&worker_thread_function, "Worker thread", 1024, &client_agent, 3, NULL);

        // prepare lines of text for sending to the server
        // and post each one for transmission
        
        const std::size_t line_max_len = 20;
        char line[line_max_len];
        int counter = 0;
        
        while (true)
        {
            std::sprintf(line, "Hello %d", ++counter);
            
            yami::parameters params;

            // the "content" field name is arbitrary,
            // but needs to be recognized at the server side

            params.set_string("content", line);

            auto om = client_agent.send(SERVER_TARGET,
                    "printer", "print", params);

            std::printf("sent to server: %s\n", line);
            
            om->wait_for_transmission();
            
            vTaskDelay(1000);
        }
    }
    catch (const std::exception & e)
    {
        std::printf("error: %s\n", e.what());
    }
}
