// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

//
// README:
//
// This file is NOT a stand-alone program,
// its main function is at the end of the file
// and is supposed to be a task function, started
// after all hardware and interrupts are initialized.
//
// The following addresses and mask are used during tests:
//
#define MY_TARGET "udp://192.168.11.112:12345"

#include <yami4-cpp/yami.h>

#include <cstdio>

#include <FreeRTOS.h>
#include <task.h>

// worker thread function that is used to process I/O events
// for the server agent
void worker_thread_function(void * argument)
{
    yami::agent * server_agent = reinterpret_cast<yami::agent *>(argument);
    
    server_agent->run_worker();
}

// dispatcher thread function that is used to deliver incoming messages
// to the application code
// (this thread will ultimately invoke the call() function below)
void dispatcher_thread_function(void * argument)
{
    yami::agent * server_agent = reinterpret_cast<yami::agent *>(argument);
    
    server_agent->run_dispatcher(0); // first dispatcher index
}

void call(yami::incoming_message & im)
{
    const yami::parameters & params = im.get_parameters();

    // extract the content field
    // and print it on standard output

    std::printf("%s\n", params.get_string("content").c_str());
}

extern "C"
void MX_LWIP_Init(void);

extern "C"
void run_server(void * argument)
{
    MX_LWIP_Init();
    
    try
    {
        yami::agent server_agent;

        // start the worker and 1 dispatcher thread
        
        xTaskCreate(&worker_thread_function, "Worker thread", 1024, &server_agent, 3, NULL);
        
        xTaskCreate(&dispatcher_thread_function, "Dispatcher thread", 1024, &server_agent, 3, NULL);
        
        const std::string resolved_address =
            server_agent.add_listener(MY_TARGET);

        std::printf("The server is listening on %s\n", resolved_address.c_str());

        server_agent.register_object("printer", call);

        // block, allow worker and dispatcher to handle the incoming traffic
        vTaskDelay(portMAX_DELAY);
    }
    catch (const std::exception & e)
    {
        std::printf("error: %s\n", e.what());
    }
}
