// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

//
// README:
//
// This file is NOT a stand-alone program,
// its main function is at the end of the file
// and is supposed to be a task function, started
// after all hardware and interrupts are initialized.
//
// The following addresses and mask are used during tests:
//
#define PUBLISHER_TARGET "udp://192.168.11.111:12345"

#include <yami4-cpp/yami.h>

#include <cstdio>

#include <FreeRTOS.h>
#include <task.h>

// worker thread function that is used to process I/O events
// for the subscriber agent
void worker_thread_function(void * argument)
{
    yami::agent * subscriber_agent = reinterpret_cast<yami::agent *>(argument);
    
    subscriber_agent->run_worker();
}

// dispatcher thread function that is used to deliver incoming messages
// to the application code
// (this thread will ultimately invoke the update() function below)
void dispatcher_thread_function(void * argument)
{
    yami::agent * subscriber_agent = reinterpret_cast<yami::agent *>(argument);
    
    subscriber_agent->run_dispatcher(0); // first dispatcher index
}

void update(yami::incoming_message & message)
{
    const yami::parameters & content =
        message.get_parameters();

    const int value = content.get_integer("value");

    std::printf("received update: %d\n", value);
}

extern "C"
void MX_LWIP_Init(void);

extern "C"
void run_subscriber(void * argument)
{
    MX_LWIP_Init();
    
    try
    {
        yami::agent subscriber_agent;
        
        // start the worker and 1 dispatcher thread
        
        xTaskCreate(&worker_thread_function, "Worker thread", 1024, &subscriber_agent, 3, NULL);
        
        xTaskCreate(&dispatcher_thread_function, "Dispatcher thread", 1024, &subscriber_agent, 3, NULL);
        
        // prepare subscription update callback

        const std::string update_object_name =
            "update_handler";

        subscriber_agent.register_object(
            update_object_name, update);

        // subscribe to the producer

        yami::parameters params;
        params.set_string(
            "destination_object", update_object_name);

        subscriber_agent.send_one_way(PUBLISHER_TARGET,
            "random_number", "subscribe", params);

        std::printf("subscribed, waiting for updates\n");

        // block forever and receive updates in background
        vTaskDelay(portMAX_DELAY);
    }
    catch (const std::exception & e)
    {
        std::printf("error: %s\n", e.what());
    }
}
