// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

//
// README:
//
// This file is NOT a stand-alone program,
// its main function is at the end of the file
// and is supposed to be called from the actual main() function
// after all hardware and interrupts are initialized.
//
// This test relies on the nx_driver_stm32f4xx network driver,
// which should be modified appropriately for the target environment.
//
// The following addresses and mask are used during tests:
//
#define MY_ADDRESS IP_ADDRESS(192, 168, 11, 112)
#define MY_MASK 0xffffff00ul
#define SERVER_TARGET "tcp://192.168.11.111:12345"

#include <yami4-cpp/yami.h>

#include <cstdio>

#include <nx_api.h>
#include <tx_api.h>

#include "nx_driver_stm32f4xx.h"

#define BASE_PACKET_SIZE 1536
#define PACKET_SIZE (BASE_PACKET_SIZE + sizeof(NX_PACKET))
#define PACKET_POOL_SIZE (30 * PACKET_SIZE)
char pool_memory[PACKET_POOL_SIZE + 16];
NX_PACKET_POOL pool;

#define IP_STACK_SIZE 2048
char ip_stack[IP_STACK_SIZE];
NX_IP ip;

#define ARP_SPACE_SIZE 1024
char arp_memory[ARP_SPACE_SIZE];

#define CLIENT_STACK_SIZE 2048
char client_stack[CLIENT_STACK_SIZE];
TX_THREAD client_th;

#define WORKER_STACK_SIZE 2048
char worker_stack[WORKER_STACK_SIZE];
TX_THREAD worker_th;

// worker thread function that is used to process I/O events
// for the client agent
void worker_thread_function(ULONG input)
{
    yami::agent * client_agent = reinterpret_cast<yami::agent *>(input);
    
    client_agent->run_worker();
}

void client_thread_function(ULONG input)
{
    try
    {
        // no dispatchers are needed in a pure-client program:
        yami::parameters options;
        options.set_integer(yami::option_names::dispatcher_threads, 0);
        
        yami::agent client_agent(options);
        
        client_agent.set_netx_context(&ip, &pool);

        // start the worker thread
        
        (void)tx_thread_create(&worker_th, const_cast<char *>("Worker thread"),
            &worker_thread_function, reinterpret_cast<ULONG>(&client_agent),
            &worker_stack[0], WORKER_STACK_SIZE,
            3, 3, TX_NO_TIME_SLICE, TX_AUTO_START);

        // prepare lines of text for sending to the server
        // and post each one for transmission
        
        const std::size_t line_max_len = 20;
        char line[line_max_len];
        int counter = 0;
        
        while (true)
        {
            std::sprintf(line, "Hello %d", ++counter);
            
            yami::parameters params;

            // the "content" field name is arbitrary,
            // but needs to be recognized at the server side

            params.set_string("content", line);

            client_agent.send_one_way(SERVER_TARGET,
                "printer", "print", params);
            
            std::printf("sent to server: %s\n", line);
            
            // this not only makes a steady pace,
            // but also gives a chance to the worker thread to do its part
            
            tx_thread_sleep(100);
        }
    }
    catch (const std::exception & e)
    {
        std::printf("error: %s\n", e.what());
    }
}

void tx_application_define(void * ptr)
{
    (void)nx_system_initialize();
    
    (void)nx_packet_pool_create(&pool, const_cast<char *>("Packet pool"), BASE_PACKET_SIZE, &pool_memory[0], PACKET_POOL_SIZE);
    
    (void)nx_ip_create(&ip, const_cast<char *>("IP instance"), MY_ADDRESS, MY_MASK, &pool, &nx_driver_stm32f4xx, &ip_stack[0], IP_STACK_SIZE, 1);
    
    (void)nx_arp_enable(&ip, &arp_memory[0], ARP_SPACE_SIZE);
    
    (void)nx_udp_enable(&ip);
    (void)nx_tcp_enable(&ip);
    
    (void)tx_thread_create(&client_th, const_cast<char *>("Client thread"), &client_thread_function, 0u, &client_stack[0], CLIENT_STACK_SIZE, 3, 3, TX_NO_TIME_SLICE, TX_AUTO_START);
}

extern "C"
void run_client()
{
    tx_kernel_enter();
}
