// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

//
// README:
//
// This file is NOT a stand-alone program,
// its main function is at the end of the file
// and is supposed to be called from the actual main() function
// after all hardware and interrupts are initialized.
//
// This test relies on the nx_driver_stm32f4xx network driver,
// which should be modified appropriately for the target environment.
//
// The following addresses and mask are used during tests:
//
#define MY_ADDRESS IP_ADDRESS(192, 168, 11, 112)
#define MY_MASK 0xffffff00ul
#define MY_TARGET "tcp://192.168.11.112:12345"

#include <yami4-cpp/yami.h>

#include <cstdio>

#include <nx_api.h>
#include <tx_api.h>

#include "nx_driver_stm32f4xx.h"

#define BASE_PACKET_SIZE 1536
#define PACKET_SIZE (BASE_PACKET_SIZE + sizeof(NX_PACKET))
#define PACKET_POOL_SIZE (30 * PACKET_SIZE)
char pool_memory[PACKET_POOL_SIZE + 16];
NX_PACKET_POOL pool;

#define IP_STACK_SIZE 2048
char ip_stack[IP_STACK_SIZE];
NX_IP ip;

#define ARP_SPACE_SIZE 1024
char arp_memory[ARP_SPACE_SIZE];

#define SERVER_STACK_SIZE 2048
char server_stack[SERVER_STACK_SIZE];
TX_THREAD server_th;

#define WORKER_STACK_SIZE 2048
char worker_stack[WORKER_STACK_SIZE];
TX_THREAD worker_th;

#define DISPATCHER_STACK_SIZE 2048
char dispatcher_stack[DISPATCHER_STACK_SIZE];
TX_THREAD dispatcher_th;

// worker thread function that is used to process I/O events
// for the server agent
void worker_thread_function(ULONG input)
{
    yami::agent * server_agent = reinterpret_cast<yami::agent *>(input);
    
    server_agent->run_worker();
}

// dispatcher thread function that is used to deliver incoming messages
// to the application code
// (this thread will ultimately invoke the call() function below)
void dispatcher_thread_function(ULONG input)
{
    yami::agent * server_agent = reinterpret_cast<yami::agent *>(input);
    
    server_agent->run_dispatcher(0); // first dispatcher index
}

void call(yami::incoming_message & im)
{
    const yami::parameters & params = im.get_parameters();

    // extract the content field
    // and print it on standard output

    std::printf("%s\n", params.get_string("content").c_str());
}

void server_thread_function(ULONG input)
{
    try
    {
        yami::agent server_agent;

        server_agent.set_netx_context(&ip, &pool);

        // start the worker and 1 dispatcher thread
        
        (void)tx_thread_create(&worker_th, const_cast<char *>("Worker thread"),
            &worker_thread_function, reinterpret_cast<ULONG>(&server_agent),
            &worker_stack[0], WORKER_STACK_SIZE,
            3, 3, TX_NO_TIME_SLICE, TX_AUTO_START);

        (void)tx_thread_create(&dispatcher_th, const_cast<char *>("Dispatcher thread"),
            &dispatcher_thread_function, reinterpret_cast<ULONG>(&server_agent),
            &dispatcher_stack[0], DISPATCHER_STACK_SIZE,
            3, 3, TX_NO_TIME_SLICE, TX_AUTO_START);

        const std::string resolved_address =
            server_agent.add_listener(MY_TARGET);

        std::printf("The server is listening on %s\n", resolved_address.c_str());

        server_agent.register_object("printer", call);

        // block, allow worker and dispatcher to handle the incoming traffic
        tx_thread_sleep(0xFFFFFFFFul);
    }
    catch (const std::exception & e)
    {
        std::printf("error: %s\n", e.what());
    }
}

void tx_application_define(void * ptr)
{
    (void)nx_system_initialize();
    
    (void)nx_packet_pool_create(&pool, const_cast<char *>("Packet pool"), BASE_PACKET_SIZE, &pool_memory[0], PACKET_POOL_SIZE);
    
    (void)nx_ip_create(&ip, const_cast<char *>("IP instance"), MY_ADDRESS, MY_MASK, &pool, &nx_driver_stm32f4xx, &ip_stack[0], IP_STACK_SIZE, 1);
    
    (void)nx_arp_enable(&ip, &arp_memory[0], ARP_SPACE_SIZE);
    
    (void)nx_udp_enable(&ip);
    (void)nx_tcp_enable(&ip);
    
    (void)tx_thread_create(&server_th, const_cast<char *>("Server thread"), &server_thread_function, 0u, &server_stack[0], SERVER_STACK_SIZE, 3, 3, TX_NO_TIME_SLICE, TX_AUTO_START);
}

extern "C"
void run_server()
{
    tx_kernel_enter();
}
