// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

//
// README:
//
// This file is NOT a stand-alone program,
// its main function is at the end of the file
// and is supposed to be called from the actual main() function
// after all hardware and interrupts are initialized.
//
// This test relies on the nx_driver_stm32f4xx network driver,
// which should be modified appropriately for the target environment.
//
// The following addresses and mask are used during tests:
//
#define MY_ADDRESS IP_ADDRESS(192, 168, 11, 112)
#define MY_MASK 0xffffff00ul
#define PUBLISHER_TARGET "tcp://192.168.11.111:12345"

#include <yami4-cpp/yami.h>

#include <cstdio>

#include <nx_api.h>
#include <tx_api.h>

#include "nx_driver_stm32f4xx.h"

#define BASE_PACKET_SIZE 1536
#define PACKET_SIZE (BASE_PACKET_SIZE + sizeof(NX_PACKET))
#define PACKET_POOL_SIZE (30 * PACKET_SIZE)
char pool_memory[PACKET_POOL_SIZE + 16];
NX_PACKET_POOL pool;

#define IP_STACK_SIZE 2048
char ip_stack[IP_STACK_SIZE];
NX_IP ip;

#define ARP_SPACE_SIZE 1024
char arp_memory[ARP_SPACE_SIZE];

#define SUBSCRIBER_STACK_SIZE 2048
char subscriber_stack[SUBSCRIBER_STACK_SIZE];
TX_THREAD subscriber_th;

#define WORKER_STACK_SIZE 2048
char worker_stack[WORKER_STACK_SIZE];
TX_THREAD worker_th;

#define DISPATCHER_STACK_SIZE 2048
char dispatcher_stack[DISPATCHER_STACK_SIZE];
TX_THREAD dispatcher_th;

// worker thread function that is used to process I/O events
// for the subscriber agent
void worker_thread_function(ULONG input)
{
    yami::agent * subscriber_agent = reinterpret_cast<yami::agent *>(input);
    
    subscriber_agent->run_worker();
}

// dispatcher thread function that is used to deliver incoming messages
// to the application code
// (this thread will ultimately invoke the update() function below)
void dispatcher_thread_function(ULONG input)
{
    yami::agent * subscriber_agent = reinterpret_cast<yami::agent *>(input);
    
    subscriber_agent->run_dispatcher(0); // first dispatcher index
}

void update(yami::incoming_message & message)
{
    const yami::parameters & content =
        message.get_parameters();

    const int value = content.get_integer("value");

    std::printf("received update: %d\n", value);
}

void subscriber_thread_function(ULONG input)
{
    try
    {
        yami::agent subscriber_agent;
        
        subscriber_agent.set_netx_context(&ip, &pool);

        // start the worker and 1 dispatcher thread
        
        (void)tx_thread_create(&worker_th, const_cast<char *>("Worker thread"),
            &worker_thread_function, reinterpret_cast<ULONG>(&subscriber_agent),
            &worker_stack[0], WORKER_STACK_SIZE,
            3, 3, TX_NO_TIME_SLICE, TX_AUTO_START);

        (void)tx_thread_create(&dispatcher_th, const_cast<char *>("Dispatcher thread"),
            &dispatcher_thread_function, reinterpret_cast<ULONG>(&subscriber_agent),
            &dispatcher_stack[0], DISPATCHER_STACK_SIZE,
            3, 3, TX_NO_TIME_SLICE, TX_AUTO_START);

        // prepare subscription update callback

        const std::string update_object_name =
            "update_handler";

        subscriber_agent.register_object(
            update_object_name, update);

        // subscribe to the producer

        yami::parameters params;
        params.set_string(
            "destination_object", update_object_name);

        subscriber_agent.send_one_way(PUBLISHER_TARGET,
            "random_number", "subscribe", params);

        std::printf("subscribed, waiting for updates\n");

        // block forever and receive updates in background
        tx_thread_sleep(0xFFFFFFFFul);
    }
    catch (const std::exception & e)
    {
        std::printf("error: %s\n", e.what());
    }
}

void tx_application_define(void * ptr)
{
    (void)nx_system_initialize();
    
    (void)nx_packet_pool_create(&pool, const_cast<char *>("Packet pool"), BASE_PACKET_SIZE, &pool_memory[0], PACKET_POOL_SIZE);
    
    (void)nx_ip_create(&ip, const_cast<char *>("IP instance"), MY_ADDRESS, MY_MASK, &pool, &nx_driver_stm32f4xx, &ip_stack[0], IP_STACK_SIZE, 1);
    
    (void)nx_arp_enable(&ip, &arp_memory[0], ARP_SPACE_SIZE);
    
    (void)nx_udp_enable(&ip);
    (void)nx_tcp_enable(&ip);
    
    (void)tx_thread_create(&subscriber_th, const_cast<char *>("Subscriber thread"), &subscriber_thread_function, 0u, &subscriber_stack[0], SUBSCRIBER_STACK_SIZE, 3, 3, TX_NO_TIME_SLICE, TX_AUTO_START);
}

extern "C"
void run_subscriber()
{
    tx_kernel_enter();
}
