// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

//
// README:
//
// This file is NOT a stand-alone program,
// its main function is at the end of the file
// and is supposed to be a task function, started
// after all hardware and interrupts are initialized.
//
// The following address is used for this example:
//
#define SERVER_TARGET "udp://192.168.11.111:12345"

#include <yami4-cpp/yami.h>

#include <cstdio>

#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Task.h>

#ifdef YAMI4_WITH_TI_SIMPLELINK
#include <ti/ndk/inc/socketndk.h>
#endif // YAMI4_WITH_TI_SIMPLELINK

#define WORKERSTACKSIZE 2048
Task_Struct workerStruct;
Char workerStack[WORKERSTACKSIZE];

// worker thread function that is used to process I/O events
// for the client agent
Void worker_thread_function(UArg arg0, UArg arg1)
{
#ifdef YAMI4_WITH_TI_SIMPLELINK
    (void)fdOpenSession(Task_self());
#endif // YAMI4_WITH_TI_SIMPLELINK

    yami::agent * client_agent = reinterpret_cast<yami::agent *>(arg0);
    
    client_agent->run_worker();
}

extern "C"
Void run_client(UArg arg0, UArg arg1)
{
#ifdef YAMI4_WITH_TI_SIMPLELINK
    (void)fdOpenSession(Task_self());
#endif // YAMI4_WITH_TI_SIMPLELINK

    int a = 100;
    int b = 4;

    try
    {
        // no dispatchers are needed in a pure-client program:
        yami::parameters options;
        options.set_integer(yami::option_names::dispatcher_threads, 0);
        
        yami::agent client_agent(options);

        // start the worker thread
        
        Task_Params workerParams;

        Task_Params_init(&workerParams);
        workerParams.stackSize = WORKERSTACKSIZE;
        workerParams.stack = &workerStack;
        workerParams.arg0 = reinterpret_cast<UArg>(&client_agent);
        Task_construct(&workerStruct, &worker_thread_function, &workerParams, NULL);
        
        yami::parameters params;
        params.set_integer("a", a);
        params.set_integer("b", b);

        auto om = client_agent.send(SERVER_TARGET,
                "calculator", "calculate", params);

        om->wait_for_completion();
        const yami::message_state state = om->get_state();
        if (state == yami::replied)
        {
            const yami::parameters & reply =
                om->get_reply();

            int sum = reply.get_integer("sum");
            int difference = reply.get_integer("difference");
            int product = reply.get_integer("product");

            int ratio;
            yami::parameter_entry ratio_entry;
            const bool ratio_defined =
                reply.find("ratio", ratio_entry);
            if (ratio_defined)
            {
                ratio = ratio_entry.get_integer();
            }

            std::printf("sum        = %d\n", sum);
            std::printf("difference = %d\n", difference);
            std::printf("product    = %d\n", product);

            std::printf("ratio      = ");
            if (ratio_defined)
            {
                std::printf("%d\n", ratio);
            }
            else
            {
                std::printf("<undefined>\n");
            }
        }
        else if (state == yami::rejected)
        {
            std::printf("The message has been rejected: %s\n",
                om->get_exception_msg().c_str());
        }
        else
        {
            std::printf("The message has been abandoned.\n");
        }
    }
    catch (const std::exception & e)
    {
        std::printf("error: %s\n", e.what());
    }
}
