// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

//
// README:
//
// This file is NOT a stand-alone program,
// its main function is at the end of the file
// and is supposed to be a task function, started
// after all hardware and interrupts are initialized.
//
// The following address is used for this example:
//
#define MY_TARGET "tcp://192.168.11.112:12345"

#include <yami4-cpp/yami.h>

#include <cstdio>

#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Task.h>

#ifdef YAMI4_WITH_TI_SIMPLELINK
#include <ti/ndk/inc/socketndk.h>
#endif // YAMI4_WITH_TI_SIMPLELINK

#define WORKERSTACKSIZE 2048
Task_Struct workerStruct;
Char workerStack[WORKERSTACKSIZE];

// worker thread function that is used to process I/O events
// for the publisher agent
Void worker_thread_function(UArg arg0, UArg arg1)
{
#ifdef YAMI4_WITH_TI_SIMPLELINK
    (void)fdOpenSession(Task_self());
#endif // YAMI4_WITH_TI_SIMPLELINK

    yami::agent * publisher_agent = reinterpret_cast<yami::agent *>(arg0);
    
    publisher_agent->run_worker();
}

#define DISPATCHERSTACKSIZE 4096
Task_Struct dispatcherStruct;
Char dispatcherStack[DISPATCHERSTACKSIZE];

// dispatcher thread function that is used to deliver incoming messages
// to the application code
// (this thread will ultimately process the subscription request for random_value)
Void dispatcher_thread_function(UArg arg0, UArg arg1)
{
    yami::agent * publisher_agent = reinterpret_cast<yami::agent *>(arg0);
    
    publisher_agent->run_dispatcher(0); // first dispatcher index
}

extern "C"
Void run_publisher(UArg arg0, UArg arg1)
{
#ifdef YAMI4_WITH_TI_SIMPLELINK
    (void)fdOpenSession(Task_self());
#endif // YAMI4_WITH_TI_SIMPLELINK

    try
    {
        yami::value_publisher random_value;

        yami::agent publisher_agent;
        
        // start the worker and 1 dispatcher thread
        
        Task_Params workerParams;

        Task_Params_init(&workerParams);
        workerParams.stackSize = WORKERSTACKSIZE;
        workerParams.stack = &workerStack;
        workerParams.arg0 = reinterpret_cast<UArg>(&publisher_agent);
        Task_construct(&workerStruct, &worker_thread_function, &workerParams, NULL);

        Task_Params dispatcherParams;

        Task_Params_init(&dispatcherParams);
        dispatcherParams.stackSize = DISPATCHERSTACKSIZE;
        dispatcherParams.stack = &dispatcherStack;
        dispatcherParams.arg0 = reinterpret_cast<UArg>(&publisher_agent);
        Task_construct(&dispatcherStruct, &dispatcher_thread_function, &dispatcherParams, NULL);
        
        const std::string resolved_address =
            publisher_agent.add_listener(MY_TARGET);

        std::printf("The server is listening on %s\n", resolved_address.c_str());

        publisher_agent.register_value_publisher(
            "random_number", random_value);

        // publish random numbers forever
        yami::parameters content;
        while (true)
        {
            const int random = std::rand() % 100;
            content.set_integer("value", random);

            std::printf("publishing value %d\n", random);

            random_value.publish(content);

            // pause for 1s
            Task_sleep(1000);
        }
    }
    catch (const std::exception & e)
    {
        std::printf("error: %s\n", e.what());
    }
}
