// Copyright Maciej Sobczak 2008-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

//
// README:
//
// This file is NOT a stand-alone program,
// its main function is at the end of the file
// and is supposed to be a task function, started
// after all hardware and interrupts are initialized.
//
// The following address is used for this example:
//
#define PUBLISHER_TARGET "udp://192.168.11.111:12345"

#include <yami4-cpp/yami.h>

#include <cstdio>

#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Task.h>

#ifdef YAMI4_WITH_TI_SIMPLELINK
#include <ti/ndk/inc/socketndk.h>
#endif // YAMI4_WITH_TI_SIMPLELINK

#define WORKERSTACKSIZE 2048
Task_Struct workerStruct;
Char workerStack[WORKERSTACKSIZE];

// worker thread function that is used to process I/O events
// for the subscriber agent
Void worker_thread_function(UArg arg0, UArg arg1)
{
#ifdef YAMI4_WITH_TI_SIMPLELINK
    (void)fdOpenSession(Task_self());
#endif // YAMI4_WITH_TI_SIMPLELINK

    yami::agent * subscriber_agent = reinterpret_cast<yami::agent *>(arg0);
    
    subscriber_agent->run_worker();
}

#define DISPATCHERSTACKSIZE 2048
Task_Struct dispatcherStruct;
Char dispatcherStack[DISPATCHERSTACKSIZE];

// dispatcher thread function that is used to deliver incoming messages
// to the application code
// (this thread will ultimately invoke the update() function below)
Void dispatcher_thread_function(UArg arg0, UArg arg1)
{
    yami::agent * subscriber_agent = reinterpret_cast<yami::agent *>(arg0);
    
    subscriber_agent->run_dispatcher(0); // first dispatcher index
}

void update(yami::incoming_message & message)
{
    const yami::parameters & content =
        message.get_parameters();

    const int value = content.get_integer("value");

    std::printf("received update: %d\n", value);
}

extern "C"
Void run_subscriber(UArg arg0, UArg arg1)
{
#ifdef YAMI4_WITH_TI_SIMPLELINK
    (void)fdOpenSession(Task_self());
#endif // YAMI4_WITH_TI_SIMPLELINK

    try
    {
        yami::agent subscriber_agent;
        
        // start the worker and 1 dispatcher thread
        
        Task_Params workerParams;

        Task_Params_init(&workerParams);
        workerParams.stackSize = WORKERSTACKSIZE;
        workerParams.stack = &workerStack;
        workerParams.arg0 = reinterpret_cast<UArg>(&subscriber_agent);
        Task_construct(&workerStruct, &worker_thread_function, &workerParams, NULL);

        Task_Params dispatcherParams;

        Task_Params_init(&dispatcherParams);
        dispatcherParams.stackSize = DISPATCHERSTACKSIZE;
        dispatcherParams.stack = &dispatcherStack;
        dispatcherParams.arg0 = reinterpret_cast<UArg>(&subscriber_agent);
        Task_construct(&dispatcherStruct, &dispatcher_thread_function, &dispatcherParams, NULL);
        
        // prepare subscription update callback

        const std::string update_object_name =
            "update_handler";

        subscriber_agent.register_object(
            update_object_name, update);

        // subscribe to the producer

        yami::parameters params;
        params.set_string(
            "destination_object", update_object_name);

        subscriber_agent.send_one_way(PUBLISHER_TARGET,
            "random_number", "subscribe", params);

        std::printf("subscribed, waiting for updates\n");

        // block forever and receive updates in background
        while (true)
        {
            Task_sleep(10000);
        }
    }
    catch (const std::exception & e)
    {
        std::printf("error: %s\n", e.what());
    }
}
