// Copyright Maciej Sobczak 2019-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

#include <yami4-cpp/yami.h>
#include "utils.h"

#include <condition_variable>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <memory>
#include <mutex>
#include <string>
#include <unordered_map>

namespace // unnamed namespace
{

const std::string & err_no_cmd = "No such command.";
const std::string & err_no_object = "No such object.";
const std::string & err_no_args = "Not enough arguments.";
const std::string & err_invalid_args = "Invalid arguments.";
const std::string & err_no_agent = "Agent not initialized.";

std::string parameter_type_str(yami::parameter_type t)
{
    std::string result;
    
    switch (t)
    {
    case yami::boolean:                       result = "boolean";                 break;
    case yami::integer:                       result = "integer";                 break;
    case yami::long_long:                     result = "long_long";               break;
    case yami::double_float:                  result = "double_float";            break;
    case yami::string:                        result = "string";                  break;
    case yami::binary:                        result = "binary";                  break;
    case yami::boolean_array:                 result = "boolean_array";           break;
    case yami::integer_array:                 result = "integer_array";           break;
    case yami::long_long_array:               result = "long_long_array";         break;
    case yami::double_float_array:            result = "double_float_array";      break;
    case yami::string_array:                  result = "string_array";            break;
    case yami::binary_array:                  result = "binary_array";            break;
    case yami::nested_parameters:             result = "nested_parameters";       break;
    case yami::nested_parameters_array:       result = "nested_parameters_array"; break;
    default:
        result = "unknown";
    }
    
    return result;
}

std::string message_state_str(yami::message_state state)
{
    std::string result;
    
    switch (state)
    {
    case yami::posted:      result = "posted";      break;
    case yami::transmitted: result = "transmitted"; break;
    case yami::abandoned:   result = "abandoned";   break;
    case yami::replied:     result = "replied";     break;
    case yami::rejected:    result = "rejected";    break;
    }

    return result;
}

std::unordered_map<std::string, std::shared_ptr<yami::parameters> > params;
std::unordered_map<std::string, std::shared_ptr<yami::outgoing_message> > out_msgs;
std::unordered_map<std::string, std::shared_ptr<yami::incoming_message> > in_msgs;

// these guard the in_msgs collection (incoming messages):
std::mutex mtx;
std::condition_variable cond;

// references to local params, incoming message params and replies to outgoing messages:
std::unordered_map<std::string, const yami::parameters *> const_params;

void generic_object(yami::incoming_message & msg)
{
    // note: intentional copy of obj_name
    std::string obj_name = msg.get_object_name();
    
    // the in_msgs collection stores at most 1 incoming message
    // for the given named object, so the dispatcher is blocked
    // until the previous message for the same object is consumed
    
    std::unique_lock<std::mutex> lck(mtx);
    
    while (in_msgs.find(obj_name) != in_msgs.end())
    {
        cond.wait(lck);
    }
    
    // message is moved to extend its lifetime
    in_msgs[obj_name] = std::shared_ptr<yami::incoming_message>(
        new yami::incoming_message(msg));
    
    cond.notify_one();
}

std::ostream * dump_stream = NULL;

void ok()
{
    std::cout << "OK\n";
    
    if (dump_stream != NULL)
    {
        *dump_stream << "(out) OK\n";
    }
}

void ok(const std::string & result)
{
    std::cout << "OK " << result << '\n';
    
    if (dump_stream != NULL)
    {
        *dump_stream << "(out) OK " << result << '\n';
    }
}

void ok(const char * result)
{
    std::cout << "OK " << result << '\n';
    
    if (dump_stream != NULL)
    {
        *dump_stream << "(out) OK " << result << '\n';
    }
}

void ok(bool result)
{
    if (result)
    {
        std::cout << "OK true\n";
    }
    else
    {
        std::cout << "OK false\n";
    }
    
    if (dump_stream != NULL)
    {
        if (result)
        {
            *dump_stream << "(out) OK true\n";
        }
        else
        {
            *dump_stream << "(out) OK false\n";
        }
    }
}

void error(const std::string & msg)
{
    std::cout << "ERROR " << msg << '\n';
    
    if (dump_stream != NULL)
    {
        *dump_stream << "(out) ERROR " << msg << '\n';
    }
}

void error(const std::exception & e)
{
    std::cout << "ERROR " << e.what() << '\n';
    
    if (dump_stream != NULL)
    {
        *dump_stream << "(out) ERROR " << e.what() << '\n';
    }
}

} // unnamed namespace

int main(int argc, char * argv[])
{
    std::ofstream dump_file;
    
    if (argc == 2)
    {
        if (std::string(argv[1]) == "--stderr")
        {
            dump_stream = &std::cerr;
        }
        else
        {
            dump_file.open(argv[1]);
            
            if (dump_file)
            {
                dump_stream = &dump_file;
            }
            else
            {
                std::cerr << "ERROR Cannot open dump file.\n";
                return EXIT_FAILURE;
            }
        }
    }
        
    std::unique_ptr<yami::agent> agent;
    
    std::string line;
    while (std::getline(std::cin, line))
    {
        if (dump_stream != NULL)
        {
            *dump_stream << "(in ) " << line << '\n';
        }
        
        std::vector<std::string> args = utils::tokenize(line);
        
        if (args.empty() == false)
        {
            if (args[0] == "params")
            {
                if (args.size() == 1)
                {
                    // command: params
                    // returns: <param-key>
                    
                    std::shared_ptr<yami::parameters> p(new yami::parameters());
                    
                    std::string key = utils::new_key();
                    
                    params[key] = p;
                    const_params[key] = p.get();
                    
                    ok(key);
                }
                else
                {
                    error(err_invalid_args);
                }
            }
            else if (args[0] == "clone")
            {
                // command: clone <param-key>
                // returns: <param-key>
                
                if (args.size() == 2)
                {
                    auto it = params.find(args[1]);
                    if (it != params.end())
                    {
                        try
                        {
                            yami::parameters & p = *(it->second);
                            
                            std::shared_ptr<yami::parameters> p2(new yami::parameters(p));
                            
                            std::string key = utils::new_key();
                            
                            params[key] = p2;
                            const_params[key] = p2.get();
                            
                            ok(key);
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "set_boolean")
            {
                // command: set_boolean <param-key> <field-name> <value>
                
                if (args.size() == 4)
                {
                    auto it = params.find(args[1]);
                    if (it != params.end())
                    {
                        if (args[3] == "true" || args[3] == "false")
                        {
                            yami::parameters & p = *(it->second);
                            
                            p.set_boolean(args[2], args[3] == "true");
                            
                            ok();
                        }
                        else
                        {
                            error(err_invalid_args);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "set_integer")
            {
                // command: set_integer <param-key> <field-name> <value>
                
                if (args.size() == 4)
                {
                    auto it = params.find(args[1]);
                    if (it != params.end())
                    {
                        try
                        {
                            yami::parameters & p = *(it->second);
                            
                            p.set_integer(args[2], std::stoi(args[3]));
                            
                            ok();
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "set_long_long")
            {
                // command: set_long_long <param-key> <field-name> <value>
                
                if (args.size() == 4)
                {
                    auto it = params.find(args[1]);
                    if (it != params.end())
                    {
                        try
                        {
                            yami::parameters & p = *(it->second);
                            
                            p.set_long_long(args[2], std::stoll(args[3]));
                            
                            ok();
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "set_double" || args[0] == "set_double_float")
            {
                // command: set_double <param-key> <field-name> <value>
                
                if (args.size() == 4)
                {
                    auto it = params.find(args[1]);
                    if (it != params.end())
                    {
                        try
                        {
                            yami::parameters & p = *(it->second);
                            
                            p.set_double_float(args[2], std::stod(args[3]));
                            
                            ok();
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "set_string")
            {
                // command: set_string <param-key> <field-name> <value>
                
                if (args.size() == 4)
                {
                    auto it = params.find(args[1]);
                    if (it != params.end())
                    {
                        try
                        {
                            yami::parameters & p = *(it->second);
                            
                            p.set_string(args[2], args[3]);
                            
                            ok();
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "set_binary")
            {
                // command: set_binary <param-key> <field-name> <value>
                
                if (args.size() == 4)
                {
                    auto it = params.find(args[1]);
                    if (it != params.end())
                    {
                        try
                        {
                            yami::parameters & p = *(it->second);
                            
                            const std::vector<unsigned char> & bin = utils::to_binary(args[3]);
                            
                            p.set_binary(args[2], &bin[0], bin.size());
                            
                            ok();
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "set_boolean_array")
            {
                // command: set_boolean_array <param-key> <field-name> <value> ...
                
                if (args.size() >= 4)
                {
                    auto it = params.find(args[1]);
                    if (it != params.end())
                    {
                        const std::size_t size = args.size() - 4 + 1;
                        
                        // this is to avoid std::vector<bool>, due to its special functions
                        std::unique_ptr<bool[]> a(new bool[size]);
                        
                        bool data_valid = true;
                        for (std::size_t i = 3; data_valid && i != args.size(); ++i)
                        {
                            if (args[i] == "true" || args[i] == "false")
                            {
                                a[i - 3] = args[i] == "true";
                            }
                            else
                            {
                                data_valid = false;
                            }
                        }
                        
                        if (data_valid)
                        {
                            yami::parameters & p = *(it->second);
                            
                            p.set_boolean_array(args[2], a.get(), size);
                            
                            ok();
                        }
                        else
                        {
                            error(err_invalid_args);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "set_integer_array")
            {
                // command: set_integer_array <param-key> <field-name> <value> ...
                
                if (args.size() >= 4)
                {
                    auto it = params.find(args[1]);
                    if (it != params.end())
                    {
                        std::vector<int> a;
                        
                        for (std::size_t i = 3; i != args.size(); ++i)
                        {
                            a.push_back(std::stoi(args[i]));
                        }
                        
                        yami::parameters & p = *(it->second);
                        
                        p.set_integer_array(args[2], &a[0], a.size());
                        
                        ok();
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "set_long_long_array")
            {
                // command: set_long_long_array <param-key> <field-name> <value> ...
                
                if (args.size() >= 4)
                {
                    auto it = params.find(args[1]);
                    if (it != params.end())
                    {
                        std::vector<long long> a;
                        
                        for (std::size_t i = 3; i != args.size(); ++i)
                        {
                            a.push_back(std::stoll(args[i]));
                        }
                        
                        yami::parameters & p = *(it->second);
                        
                        p.set_long_long_array(args[2], &a[0], a.size());
                        
                        ok();
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "set_double_float_array" || args[0] == "set_double_array")
            {
                // command: set_double_float_array <param-key> <field-name> <value> ...
                // command: set_double_array <param-key> <field-name> <value> ...
                
                if (args.size() >= 4)
                {
                    auto it = params.find(args[1]);
                    if (it != params.end())
                    {
                        std::vector<double> a;
                        
                        for (std::size_t i = 3; i != args.size(); ++i)
                        {
                            a.push_back(static_cast<double>(std::stold(args[i])));
                        }
                        
                        yami::parameters & p = *(it->second);
                        
                        p.set_double_float_array(args[2], &a[0], a.size());
                        
                        ok();
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "set_string_array")
            {
                // command: set_string_array <param-key> <field-name> <value> ...
                
                if (args.size() >= 4)
                {
                    auto it = params.find(args[1]);
                    if (it != params.end())
                    {
                        const std::size_t size = args.size() - 4 + 1;
                        
                        yami::parameters & p = *(it->second);
                        
                        p.create_string_array(args[2], size);
                        
                        for (std::size_t i = 3; i != args.size(); ++i)
                        {
                            p.set_string_in_array(args[2], i - 3, args[i]);
                        }
                        
                        ok();
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "set_binary_array")
            {
                // command: set_binary_array <param-key> <field-name> <value> ...
                
                if (args.size() >= 4)
                {
                    auto it = params.find(args[1]);
                    if (it != params.end())
                    {
                        const std::size_t size = args.size() - 4 + 1;
                        
                        yami::parameters & p = *(it->second);
                        
                        p.create_binary_array(args[2], size);
                        
                        for (std::size_t i = 3; i != args.size(); ++i)
                        {
                            std::vector<unsigned char> bin = utils::to_binary(args[i]);
                            
                            p.set_binary_in_array(args[2], i - 3, &bin[0], bin.size());
                        }
                        
                        ok();
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "get_boolean")
            {
                // command: get_boolean <param-key> <field-name>
                // returns: <value>
                
                if (args.size() == 3)
                {
                    auto it = const_params.find(args[1]);
                    if (it != const_params.end())
                    {
                        try
                        {
                            const yami::parameters & p = *(it->second);
                            
                            bool result = p.get_boolean(args[2]);
                            
                            ok(result ? "true" : "false");
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "get_integer")
            {
                // command: get_integer <param-key> <field-name>
                // returns: <value>
                
                if (args.size() == 3)
                {
                    auto it = const_params.find(args[1]);
                    if (it != const_params.end())
                    {
                        try
                        {
                            const yami::parameters & p = *(it->second);
                            
                            int result = p.get_integer(args[2]);
                            
                            ok(std::to_string(result));
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "get_long_long")
            {
                // command: get_long_long <param-key> <field-name>
                // returns: <value>
                
                if (args.size() == 3)
                {
                    auto it = const_params.find(args[1]);
                    if (it != const_params.end())
                    {
                        try
                        {
                            const yami::parameters & p = *(it->second);
                            
                            long long result = p.get_long_long(args[2]);
                            
                            ok(std::to_string(result));
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "get_double" || args[0] == "get_double_float")
            {
                // command: get_double <param-key> <field-name>
                // returns: <value>
                
                if (args.size() == 3)
                {
                    auto it = const_params.find(args[1]);
                    if (it != const_params.end())
                    {
                        try
                        {
                            const yami::parameters & p = *(it->second);
                            
                            double result = p.get_double_float(args[2]);
                            
                            ok(std::to_string(result));
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "get_string")
            {
                // command: get_string <param-key> <field-name>
                // returns: <value>
                
                if (args.size() == 3)
                {
                    auto it = const_params.find(args[1]);
                    if (it != const_params.end())
                    {
                        try
                        {
                            const yami::parameters & p = *(it->second);
                            
                            const std::string & result = p.get_string(args[2]);
                            
                            ok(utils::format(result));
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "get_binary")
            {
                // command: get_binary <param-key> <field-name>
                // returns: <value>
                
                if (args.size() == 3)
                {
                    auto it = const_params.find(args[1]);
                    if (it != const_params.end())
                    {
                        try
                        {
                            const yami::parameters & p = *(it->second);
                            
                            std::size_t size;
                            const void * raw_data = p.get_binary(args[2], size);
                            
                            const std::string & result = utils::from_binary(raw_data, size);
                            
                            ok(utils::format(result));
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "get_boolean_array")
            {
                // command: get_boolean_array <param-key> <field-name>
                // returns: <value> ...
                
                if (args.size() == 3)
                {
                    auto it = const_params.find(args[1]);
                    if (it != const_params.end())
                    {
                        try
                        {
                            const yami::parameters & p = *(it->second);
                            
                            std::size_t size;
                            const bool * raw_data = p.get_boolean_array(args[2], size);
                            
                            std::vector<std::string> result;
                            
                            for (std::size_t i = 0; i != size; ++i)
                            {
                                result.push_back(raw_data[i] ? "true" : "false");
                            }
                            
                            ok(utils::catenate(result));
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "get_integer_array")
            {
                // command: get_integer_array <param-key> <field-name>
                // returns: <value> ...
                
                if (args.size() == 3)
                {
                    auto it = const_params.find(args[1]);
                    if (it != const_params.end())
                    {
                        try
                        {
                            const yami::parameters & p = *(it->second);
                            
                            std::size_t size;
                            const int * raw_data = p.get_integer_array(args[2], size);
                            
                            std::vector<std::string> result;
                            
                            for (std::size_t i = 0; i != size; ++i)
                            {
                                result.push_back(std::to_string(raw_data[i]));
                            }
                            
                            ok(utils::catenate(result));
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
            }
            else if (args[0] == "get_long_long_array")
            {
                // command: get_long_long_array <param-key> <field-name>
                // returns: <value> ...
                
                if (args.size() == 3)
                {
                    auto it = const_params.find(args[1]);
                    if (it != const_params.end())
                    {
                        try
                        {
                            const yami::parameters & p = *(it->second);
                            
                            std::size_t size;
                            const long long * raw_data = p.get_long_long_array(args[2], size);
                            
                            std::vector<std::string> result;
                            
                            for (std::size_t i = 0; i != size; ++i)
                            {
                                result.push_back(std::to_string(raw_data[i]));
                            }
                            
                            ok(utils::catenate(result));
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "get_double_float_array" || args[0] == "get_double_array")
            {
                // command: get_double_float_array <param-key> <field-name>
                // command: get_double_array <param-key> <field-name>
                // returns: <value> ...
                
                if (args.size() == 3)
                {
                    auto it = const_params.find(args[1]);
                    if (it != const_params.end())
                    {
                        try
                        {
                            const yami::parameters & p = *(it->second);
                            
                            std::size_t size;
                            const double * raw_data = p.get_double_float_array(args[2], size);
                            
                            std::vector<std::string> result;
                            
                            for (std::size_t i = 0; i != size; ++i)
                            {
                                result.push_back(std::to_string(raw_data[i]));
                            }
                            
                            ok(utils::catenate(result));
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "get_string_array")
            {
                // command: get_string_array <param-key> <field-name>
                // returns: <value> ...
                
                if (args.size() == 3)
                {
                    auto it = const_params.find(args[1]);
                    if (it != const_params.end())
                    {
                        try
                        {
                            const yami::parameters & p = *(it->second);
                            
                            std::size_t size = p.get_string_array_length(args[2]);
                            
                            std::vector<std::string> result;
                            
                            for (std::size_t i = 0; i != size; ++i)
                            {
                                result.push_back(p.get_string_in_array(args[2], i));
                            }
                            
                            ok(utils::catenate(result));
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "get_binary_array")
            {
                // command: get_binary_array <param-key> <field-name>
                // returns: <value> ...
                
                if (args.size() == 3)
                {
                    auto it = const_params.find(args[1]);
                    if (it != const_params.end())
                    {
                        try
                        {
                            const yami::parameters & p = *(it->second);
                            
                            std::size_t size = p.get_binary_array_length(args[2]);
                            
                            std::vector<std::string> result;
                            
                            for (std::size_t i = 0; i != size; ++i)
                            {
                                std::size_t data_length;
                                const void * raw_data = p.get_binary_in_array(args[2], i, data_length);
                                
                                result.push_back(utils::from_binary(raw_data, data_length));
                            }
                            
                            ok(utils::catenate(result));
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "size")
            {
                // command: size <param-key>
                // returns: <value>
                
                if (args.size() == 2)
                {
                    auto it = const_params.find(args[1]);
                    if (it != const_params.end())
                    {
                        const yami::parameters & p = *(it->second);
                        
                        std::size_t result = p.size();
                        
                        ok(std::to_string(result));
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "fields")
            {
                // command: fields <param-key>
                // returns: <name> <type> ...
                
                if (args.size() == 2)
                {
                    auto cp_it = const_params.find(args[1]);
                    if (cp_it != const_params.end())
                    {
                        const yami::parameters & p = *(cp_it->second);
                        
                        std::vector<std::string> result;
                        
                        auto it = p.begin();
                        auto end = p.end();
                        for ( ; it != end; ++it)
                        {
                            const yami::parameter_entry & e = *it;
                            
                            result.push_back(e.name());
                            result.push_back(parameter_type_str(e.type()));
                        }
                        
                        ok(utils::catenate(result));
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "field_names")
            {
                // command: field_names <param-key>
                // returns: <name> ...
                
                if (args.size() == 2)
                {
                    auto cp_it = const_params.find(args[1]);
                    if (cp_it != const_params.end())
                    {
                        const yami::parameters & p = *(cp_it->second);
                        
                        std::vector<std::string> result;
                        
                        auto it = p.begin();
                        auto end = p.end();
                        for ( ; it != end; ++it)
                        {
                            const yami::parameter_entry & e = *it;
                            
                            result.push_back(e.name());
                        }
                        
                        ok(utils::catenate(result));
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "type")
            {
                // command: type <param-key> <field-name>
                // returns: <type>
                
                if (args.size() == 3)
                {
                    auto it = const_params.find(args[1]);
                    if (it != const_params.end())
                    {
                        try
                        {
                            const yami::parameters & p = *(it->second);
                            
                            yami::parameter_type t = p.type(args[2]);
                            
                            ok(parameter_type_str(t));
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "remove")
            {
                // command: remove <param-key> <field-name>
                
                if (args.size() == 3)
                {
                    auto it = params.find(args[1]);
                    if (it != params.end())
                    {
                        try
                        {
                            yami::parameters & p = *(it->second);
                            
                            p.remove(args[2]);
                            
                            ok();
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "init_agent")
            {
                if (args.size() == 1)
                {
                    // command: init_agent
                    
                    try
                    {
                        agent.reset(new yami::agent());
                        
                        ok();
                    }
                    catch (const std::exception & e)
                    {
                        error(e);
                    }
                }
                else if (args.size() == 2)
                {
                    // command: init_agent <param-key>
                    
                    auto it = const_params.find(args[1]);
                    if (it != const_params.end())
                    {
                        try
                        {
                            const yami::parameters & p = *(it->second);
                            
                            agent.reset(new yami::agent(p));
                            
                            ok();
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_invalid_args);
                }
            }
            else if (args[0] == "add_listener")
            {
                if (agent != nullptr)
                {
                    // command: add_listener <target>
                    
                    if (args.size() == 2)
                    {
                        try
                        {
                            const std::string & result =
                                agent->add_listener(args[1]);
                            
                            ok(result);
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_args);
                    }
                }
                else
                {
                    error(err_no_agent);
                }
            }
            else if (args[0] == "register_object")
            {
                if (agent != nullptr)
                {
                    // command: register_object <obj-name>
                    
                    if (args.size() == 2)
                    {
                        try
                        {
                            agent->register_object(args[1], generic_object);
                            
                            ok();
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_args);
                    }
                }
                else
                {
                    error(err_no_agent);
                }
            }
            else if (args[0] == "has_incoming")
            {
                // command: has_incoming <obj-name> ...
                // returns: true/false whether message arrived for one or more objects
                
                if (args.size() >= 2)
                {
                    std::unique_lock<std::mutex> lck(mtx);
                    
                    bool found_some_msg = false;
                    for (std::size_t i = 1; i != args.size(); ++i)
                    {
                        found_some_msg = found_some_msg ||
                            (in_msgs.find(args[i]) != in_msgs.end());
                    }
                    
                    ok(found_some_msg);
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "wait_for_incoming")
            {
                // command: wait_for_incoming <obj-name> ...
                // timing : blocks until message arrives for one or more objects
                
                if (args.size() >= 2)
                {
                    std::unique_lock<std::mutex> lck(mtx);
                    
                    while (true)
                    {
                        bool found_some_msg = false;
                        for (std::size_t i = 1; i != args.size(); ++i)
                        {
                            found_some_msg = found_some_msg ||
                                (in_msgs.find(args[i]) != in_msgs.end());
                        }
                        
                        if (found_some_msg)
                        {
                            break;
                        }
                        else
                        {
                            cond.wait(lck);
                        }
                    }
                    
                    ok();
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "get_source")
            {
                // command: get_source <obj-name>
                // returns: <target-name>
                
                if (args.size() == 2)
                {
                    std::unique_lock<std::mutex> lck(mtx);
                    
                    auto it = in_msgs.find(args[1]);
                    if (it != in_msgs.end())
                    {
                        yami::incoming_message & im = *(it->second);
                        
                        ok(im.get_source());
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "get_message_name")
            {
                // command: get_message_name <obj-name>
                // returns: <msg-name>
                
                if (args.size() == 2)
                {
                    std::unique_lock<std::mutex> lck(mtx);
                    
                    auto it = in_msgs.find(args[1]);
                    if (it != in_msgs.end())
                    {
                        yami::incoming_message & im = *(it->second);
                        
                        ok(utils::format(im.get_message_name()));
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "get_parameters")
            {
                // command: get_parameters <obj-name>
                // returns: <param-key>
                
                if (args.size() == 2)
                {
                    std::unique_lock<std::mutex> lck(mtx);
                    
                    auto it = in_msgs.find(args[1]);
                    if (it != in_msgs.end())
                    {
                        yami::incoming_message & im = *(it->second);
                        
                        const yami::parameters & p = im.get_parameters();
                        
                        const_params[args[1]] = &p;
                        
                        ok(args[1]);
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "reply")
            {
                if (args.size() >= 2)
                {
                    std::unique_lock<std::mutex> lck(mtx);
                    
                    auto it = in_msgs.find(args[1]);
                    if (it != in_msgs.end())
                    {
                        yami::incoming_message & im = *(it->second);
                        
                        if (args.size() == 2)
                        {
                            // command: reply <obj-name>
                            
                            im.reply();
                        }
                        else
                        {
                            // command: reply <obj-name> <param-key>
                            
                            const yami::parameters * p = const_params[args[2]];
                            
                            im.reply(*p);
                        }
                        
                        in_msgs.erase(args[1]);
                        const_params.erase(args[1]);
                        
                        cond.notify_one();
                        
                        ok();
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "reject")
            {
                if (args.size() >= 2)
                {
                    std::unique_lock<std::mutex> lck(mtx);
                    
                    auto it = in_msgs.find(args[1]);
                    if (it != in_msgs.end())
                    {
                        yami::incoming_message & im = *(it->second);
                        
                        if (args.size() == 2)
                        {
                            // command: reject <obj-name>
                            
                            im.reject();
                        }
                        else
                        {
                            // command: reject <obj-name> <error-msg>
                            
                            im.reject(args[2]);
                        }
                        
                        in_msgs.erase(args[1]);
                        const_params.erase(args[1]);
                        
                        cond.notify_one();
                        
                        ok();
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "consume")
            {
                if (args.size() == 2)
                {
                    // command: consume <obj-name>
                    
                    std::unique_lock<std::mutex> lck(mtx);
                    
                    auto it = in_msgs.find(args[1]);
                    if (it != in_msgs.end())
                    {
                        in_msgs.erase(args[1]);
                        const_params.erase(args[1]);
                        
                        cond.notify_one();
                        
                        ok();
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "send_one_way")
            {
                if (agent != nullptr)
                {
                    if (args.size() == 4)
                    {
                        // command: send_one_way <target> <object-name> <message-name>
                        
                        try
                        {
                            agent->send_one_way(args[1], args[2], args[3]);
                            
                            ok();
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else if (args.size() == 5)
                    {
                        // command: send_one_way <target> <object-name> <message-name> <param-key>
                        
                        auto it = const_params.find(args[4]);
                        if (it != const_params.end())
                        {
                            const yami::parameters & p = *(it->second);
                            
                            try
                            {
                                agent->send_one_way(args[1], args[2], args[3], p);
                                
                                ok();
                            }
                            catch (const std::exception & e)
                            {
                                error(e);
                            }
                        }
                        else
                        {
                            error(err_no_object);
                        }
                    }
                    else
                    {
                        error(err_no_args);
                    }
                }
                else
                {
                    error(err_no_agent);
                }
            }
            else if (args[0] == "send")
            {
                if (agent != nullptr)
                {
                    if (args.size() == 4)
                    {
                        // command: send <target> <object-name> <message-name>
                        // returns: <outmsg-key>
                        
                        try
                        {
                            std::shared_ptr<yami::outgoing_message> om =
                                agent->send(args[1], args[2], args[3]);
                            
                            std::string key = utils::new_key();
                            
                            out_msgs[key] = om;
                            
                            ok(key);
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else if (args.size() == 5)
                    {
                        // command: send <target> <object-name> <message-name> <param-key>
                        // returns: <outmsg-key>
                        
                        auto it = const_params.find(args[4]);
                        if (it != const_params.end())
                        {
                            const yami::parameters & p = *(it->second);
                            
                            try
                            {
                                std::shared_ptr<yami::outgoing_message> om =
                                    agent->send(args[1], args[2], args[3], p);
                                
                                std::string key = utils::new_key();
                                
                                out_msgs[key] = om;
                                
                                ok(key);
                            }
                            catch (const std::exception & e)
                            {
                                error(e);
                            }
                        }
                        else
                        {
                            error(err_no_object);
                        }
                    }
                    else
                    {
                        error(err_no_args);
                    }
                }
                else
                {
                    error(err_no_agent);
                }
            }
            else if (args[0] == "get_state")
            {
                // command: get_state <outmsg-key>
                // returns: <msg-state>
                
                if (args.size() == 2)
                {
                    auto it = out_msgs.find(args[1]);
                    
                    if (it != out_msgs.end())
                    {
                        yami::outgoing_message & om = *(it->second);
                        
                        try
                        {
                            yami::message_state state = om.get_state();
                            
                            ok(message_state_str(state));
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "wait_for_transmission")
            {
                if (args.size() >= 2)
                {
                    auto it = out_msgs.find(args[1]);
                    
                    if (it != out_msgs.end())
                    {
                        yami::outgoing_message & om = *(it->second);
                        
                        try
                        {
                            if (args.size() == 2)
                            {
                                // command: wait_for_transmission <outmsg-key>
                                
                                om.wait_for_transmission();
                                
                                ok();
                            }
                            else
                            {
                                // command: wait_for_transmission <outmsg-key> <timeout>
                                
                                bool result = om.wait_for_transmission(std::stoi(args[2]));
                                
                                if (result)
                                {
                                    ok();
                                }
                                else
                                {
                                    ok("timeout");
                                }
                            }
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "wait_for_completion")
            {
                if (args.size() >= 2)
                {
                    auto it = out_msgs.find(args[1]);
                    
                    if (it != out_msgs.end())
                    {
                        yami::outgoing_message & om = *(it->second);
                        
                        try
                        {
                            if (args.size() == 2)
                            {
                                // command: wait_for_completion <outmsg-key>
                                
                                om.wait_for_completion();
                                
                                ok();
                            }
                            else
                            {
                                // command: wait_for_completion <outmsg-key> <timeout>
                                
                                bool result = om.wait_for_completion(std::stoi(args[2]));
                                
                                if (result)
                                {
                                    ok();
                                }
                                else
                                {
                                    ok("timeout");
                                }
                            }
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "get_reply")
            {
                // command: get_reply <outmsg-key>
                // returns: <param-key>
                
                if (args.size() == 2)
                {
                    auto it = out_msgs.find(args[1]);
                    
                    if (it != out_msgs.end())
                    {
                        yami::outgoing_message & om = *(it->second);
                        
                        try
                        {
                            const yami::parameters & p = om.get_reply();
                            
                            const_params[args[1]] = &p;
                            
                            ok(args[1]);
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "get_exception_msg")
            {
                // command: get_exception_msg <outmsg-key>
                // returns: <exception-msg>
                
                if (args.size() == 2)
                {
                    auto it = out_msgs.find(args[1]);
                    
                    if (it != out_msgs.end())
                    {
                        yami::outgoing_message & om = *(it->second);
                        
                        try
                        {
                            ok(utils::format(om.get_exception_msg()));
                        }
                        catch (const std::exception & e)
                        {
                            error(e);
                        }
                    }
                    else
                    {
                        error(err_no_object);
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "delete")
            {
                // command: delete <obj-key>
                
                if (args.size() == 2)
                {
                    auto p_it = params.find(args[1]);
                    if (p_it != params.end())
                    {
                        params.erase(p_it);
                        const_params.erase(args[1]);
                        
                        ok();
                    }
                    else
                    {
                        auto om_it = out_msgs.find(args[1]);
                        if (om_it != out_msgs.end())
                        {
                            out_msgs.erase(om_it);
                            const_params.erase(args[1]);
                            
                            ok();
                        }
                        else
                        {
                            std::unique_lock<std::mutex> lck(mtx);
                            
                            auto im_it = in_msgs.find(args[1]);
                            if (im_it != in_msgs.end())
                            {
                                in_msgs.erase(im_it);
                                const_params.erase(args[1]);
                                
                                cond.notify_one();
                            }
                            else
                            {
                                error(err_no_object);
                            }
                        }
                    }
                }
                else
                {
                    error(err_no_args);
                }
            }
            else if (args[0] == "exit")
            {
                // command: exit
                
                break;
            }
            else
            {
                error(err_no_cmd);
            }
        } // if line not empty
    } // while getline
}
