// Copyright Maciej Sobczak 2019-2020.
// This file is part of YAMI4.
//
// YAMI4 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// YAMI4 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.

#include "utils.h"

#include <cctype>

namespace // unnamed
{

int key_index = 0;

int from_hex(char c)
{
    int digit;
    
    if (std::isdigit(c) != 0)
    {
        digit = c - '0';
    }
    else
    {
        if (std::islower(c) != 0)
        {
            digit = 10 + c - 'a';
        }
        else
        {
            digit = 10 + c - 'A';
        }
    }
    
    return digit;
}

char to_hex(int v)
{
    char digit;
    
    if (v <= 9)
    {
        digit = static_cast<char>('0' + v);
    }
    else
    {
        digit = static_cast<char>('a' + (v - 10));
    }
    
    return digit;
}

} // unnamed namespace

std::string utils::new_key()
{
    return std::to_string(++key_index);
}

std::vector<std::string> utils::tokenize(const std::string & str)
{
    std::vector<std::string> result;
    
    enum parser_state { not_in_string, in_string,
        in_escape, in_hex_1, in_hex_2, error };

    parser_state state = not_in_string;
    
    char hex_1;
    char hex_2;
    std::string token;
    for (std::size_t i = 0; i != str.size() && state != error; ++i)
    {
        char c = str[i];
        
        switch (state)
        {
        case not_in_string:
            if (std::isgraph(c) != 0)
            {
                if (c == '\"')
                {
                    state = in_string;
                }
                else
                {
                    token += c;
                }
            }
            else if (std::isblank(c) != 0)
            {
                if (token.empty() == false)
                {
                    result.push_back(token);
                    token.clear();
                }
            }
            
            break;
        
        case in_string:
            if (c == '\"')
            {
                result.push_back(token);
                token.clear();
                
                state = not_in_string;
            }
            else if (c == '\\')
            {
                state = in_escape;
            }
            else
            {
                token += c;
            }
            
            break;
        
        case in_escape:
            switch (c)
            {
            case '\'': token += '\''; state = in_string; break;
            case '\"': token += '\"'; state = in_string; break;
            case '\?': token += '\?'; state = in_string; break;
            case '\\': token += '\\'; state = in_string; break;
            case '0':  token += '\0'; state = in_string; break;
            case 'a':  token += '\a'; state = in_string; break;
            case 'b':  token += '\b'; state = in_string; break;
            case 'f':  token += '\f'; state = in_string; break;
            case 'n':  token += '\n'; state = in_string; break;
            case 'r':  token += '\r'; state = in_string; break;
            case 't':  token += '\t'; state = in_string; break;
            case 'v':  token += '\v'; state = in_string; break;
            case 'x':
                state = in_hex_1;
                break;
            
            default:
                state = error;
            }
            
            break;
        
        case in_hex_1:
            if (std::isxdigit(c) != 0)
            {
                hex_1 = c;
                
                state = in_hex_2;
            }
            else
            {
                state = error;
            }
            
            break;
        
        case in_hex_2:
            if (std::isxdigit(c) != 0)
            {
                hex_2 = c;
                
                token += static_cast<char>(16 * from_hex(hex_1) + from_hex(hex_2));
                
                state = in_string;
            }
            else
            {
                state = error;
            }
            
            break;
        
        case error:
            // do nothing till the end of the string
            break;
        }
    }
    
    if (state == not_in_string && token.empty() == false)
    {
        result.push_back(token);
    }
    
    if (state == error)
    {
        result.clear();
    }
    
    return result;
}

std::string utils::format(const std::string & str)
{
    std::string result;
    
    bool only_graphs = true;
    for (char c : str)
    {
        if (std::isprint(c) != 0)
        {
            result += c;
        }
        else
        {
            switch (c)
            {
            case '\'': result += "\\\'"; break;
            case '\"': result += "\\\""; break;
            case '\?': result += "\\\?"; break;
            case '\\': result += "\\\\"; break;
            case '\0': result += "\\0";  break;
            case '\a': result += "\\a";  break;
            case 'b':  result += "\\b";  break;
            case 'f':  result += "\\f";  break;
            case 'n':  result += "\\n";  break;
            case 'r':  result += "\\r";  break;
            case 't':  result += "\\t";  break;
            case 'v':  result += "\\v";  break;
            default:
                result += "\\x";
                result += to_hex(static_cast<int>(c) / 16);
                result += to_hex(static_cast<int>(c) % 16);
            }
        }
        
        if (std::isgraph(c) == 0)
        {
            only_graphs = false;
        }
    }
    
    if (only_graphs == false)
    {
        result = '\"' + result + '\"';
    }
    
    return result;
}

std::string utils::catenate(const std::vector<std::string> & tokens)
{
    std::string result;
    
    for (const std::string & t : tokens)
    {
        if (result.empty() == false)
        {
            result += ' ';
        }
        
        result += format(t);
    }
    
    return result;
}

std::vector<unsigned char> utils::to_binary(const std::string & str)
{
    const unsigned char * begin = reinterpret_cast<const unsigned char *>(str.data());
    const unsigned char * end = begin + str.size();
    
    std::vector<unsigned char> result(begin, end);
    
    return result;
}

std::string utils::from_binary(const void * data, std::size_t size)
{
    std::string result;
    
    result += '\"';
    
    for (std::size_t i = 0; i != size; ++i)
    {
        unsigned char b = *(static_cast<const unsigned char *>(data) + i);
        
        result += "\\x";
        result += to_hex(static_cast<int>(b) / 16);
        result += to_hex(static_cast<int>(b) % 16);
    }
    
    result += '\"';
    
    return result;
}
