(* ::Package:: *)

(*
   Copyright Maciej Sobczak 2019-2020.
   This file is part of YAMI4.
   
   YAMI4 is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   YAMI4 is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.
*)


BeginPackage["YAMI4`"];


openGateway::usage="openGateway[path] Finds and opens the gateway process.\n"<>
"Path is optional and if empty, gateway is searched on $Path."; 


YAMI4LongLong::usage="YAMI4LongLong[value] Wrapper head for long long numeric values.";


YAMI4Binary::usage="YAMI4Binary[{val1,val2,...}] Wrapper head for binary values (list of byte values).";


init::usage="init[gateway,options] Initializes the agent.\n"<>
"Configuration options, if present, can be provided like any other parameters object (as Association).";


addListener::usage="addListener[gateway,target] Adds a listener in the agent, for the specific target.\n"<>
"Returns the resolved target name.";


registerObject::usage="registerObject[gateway,objName] Creates a new server-side object with the given name.";


hasIncoming::usage="hasIncoming[gateway,objName] Returns True if the given server-side object has an incoming message ready for processing.";


waitForIncoming::usage="waitForIncoming[gateway,objName] Blocks execution until a new incoming message is available for the given object.\n"<>
"Note: the blocking happens in the gateway object, not in the Wolfram kernel or front-end.";


sourceName::usage="sourceName[gateway,objName] Returns remote endpoint of the client that sent request to the given object.";


messageName::usage="messageName[gateway,objName] Returns message name of the incoming request.";


requestParams::usage="requestParams[gateway,objName] Returns parameters (Association) of the object's incoming message.";


reply::usage="reply[gateway,objName,params] Sends reply to the client, with regard to the message received by the given object.\n"<>
"Reply parameters are optional.";


reject::usage="reject[gateway,objName,exceptionMsg] Sends rejection to the client, with regard to the message received by the given object.\n"<>
"Exception message is optional.";


consume::usage="consume[gateway,objName] Consumes (drops) the last message that was received for the given object.\n"<>
"The consumption can be necessary if neither reply nor reject is intended (for one-way messages),"<>
" so that subsequent messages can be received and processed.";


sendOneWay::usage="sendOneWay[gateway,target,objName,msgName,params] Send a one-way message (no reply expected) to the given remote target and object name.\n"<>
"Parameters are optional.";


send::usage="send[gateway,target,objName,msgName,params] Send a message to the given remote target and object name.\n"<>
"Returns a handle to the outgoing message object, that can be used for further interactions with the message.\n"<>
"Parameters are optional.";


state::usage="state[outMsg] Returns the current state of the given outgoing message.";


waitForTransmission::usage="waitForTransmission[outMsg] Blocks until the given message is fully transmitted.\n"<>
"Note: the blocking happens in the gateway object, not in the Wolfram kernel or front-end.";


waitForCompletion::usage="waitForCompletion[outMsg] Blocks until response or rejection is received for the given outgoing message.\n"<>
"Note: the blocking happens in the gateway object, not in the Wolfram kernel or front-end.";


response::usage="response[outMsg] Returns the reponse parameters (if available, for messages in replied state) for the given outgoing message.";


exceptionMsg::usage="exceptionMsg[outMsg] Returns the exception message (if available, for messages in rejected state) for the given outgoing message.";


close::usage="close[object] Closes the given object and cleans up its resources.\n"<>
"The object can be:\n"<>
" - a gateway process (if blocked, use killGateway),\n"<>
" - an outgoing message.";


killGateway::usage="killGateway[gateway] Unconditionally kills the gateway process.";


(* Public symbols denoting outgoing message states. *)
outgoingMsgStates={posted,transmitted,abandoned,replied,rejected};


Begin["Private`"];


(* For unified error messages. *)
yami4::err="`1`";


openGateway[path_String]:=YAMI4Gateway[StartProcess[path]]

openGateway[]:=Module[
	{
		gatewayPath
	},
	
	If[StringMatchQ[$System,"Microsoft"~~__],
		gatewayPath=FindFile["yami4gateway.exe"],
		gatewayPath=FindFile["yami4gateway"]
	];
	
	If[gatewayPath=!=$Failed,
		openGateway[gatewayPath],
		Message[yami4::err,"ERROR Gateway executable not found in standard paths."]
	]
]


close[g_YAMI4Gateway]:=With[
	{
		proc=g[[1]]
	},
	
	WriteLine[proc,"exit"];
	If[Read[proc]===EndOfFile,
		Null,
		Message[yami4::err,"ERROR Process did not close properly."]
	]
]


killGateway[g_YAMI4Gateway]:=With[
	{
		proc=g[[1]]
	},
	
	KillProcess[proc]
]


(* Helper, splits the gateway's response into String tokens. *)
parserStates={notInString,inString,inEscape};
split[str_String]:=Module[
	{
		result={},
		token={},
		state=notInString
	},
	
	Scan[
		Function[c, (* c is a consecutive character from str *)
			Switch[state,
				notInString,
					If[c!=" ",
						If[c=="\"",
							state=inString,
							AppendTo[token,c]
						],
						If[token!={},
							AppendTo[result,StringJoin[token]];
							token={}
						]
					],
				
				inString,
					If[c=="\"",
						AppendTo[result,StringJoin[token]];
						token={};
						state=notInString,
						If[c=="\\",
							state=inEscape,
							AppendTo[token,c]
						]
					],
				
				inEscape,
					Switch[c,
						"\'",
							AppendTo[token,"\'"]; state=inString,
						"\"",
							AppendTo[token,"\""]; state=inString,
						"\?",
							AppendTo[token,"\?"]; state=inString,
						"\\",
							AppendTo[token,"\\"]; state=inString,
						"0",
							AppendTo[token,"\000"]; state=inString,
						"a",
							AppendTo[token,FromCharacterCode[7]]; state=inString,
						"b",
							AppendTo[token,"\b"]; state=inString,
						"f",
							AppendTo[token,"\f"]; state=inString,
						"n",
							AppendTo[token,"\n"]; state=inString,
						"r",
							AppendTo[token,"\r"]; state=inString,
						"t",
							AppendTo[token,"\t"]; state=inString,
						"v",
							AppendTo[token,FromCharacterCode[11]]; state=inString,
						"x",
							AppendTo[token,"\\x"]; state=inString
					]
			]
		],
		Characters[str]
	];
	
	If[token!={},
		AppendTo[result,StringJoin[token]]
	];
	
	result
]

(* Helper, parses command's response. *)
parseResponse[r_String]:=With[
	{
		words=split[r]
	},
	
	If[words[[1]]=="OK",
		Switch[Length[words],
			1,
				Null,
			2,
				words[[2]],
			_,
				Rest[words]
		],
		Message[yami4::err,StringRiffle[Rest[words]]]
	]
]

(* Helper, sends command to the gateway and reads response. *)
runCommand[proc_,words_List]:=(
	WriteLine[proc,StringRiffle[words]];
	parseResponse[ReadLine[proc]]
)

runCommand[proc_,word_String]:=runCommand[proc,List[word]]


(* Helper, transforms a single integer (assumed in range 0..255) to 2-digit hex. *)
byteToHex[b_Integer]:="\\x"<>IntegerString[b,16,2]

(* Helper, transforms a binart (wrapped) object to string "\x03\xc7\x23..." *)
binToString[bin_YAMI4Binary]:=StringJoin[byteToHex/@bin[[1]]]

(* Helper, transforms a single hex value (2-digit, like "\x12") to integer *)
hexToByte[b_String]:=FromDigits[StringTake[b,-2],16]

(* Helper, transforms a string "\x03\xc7\x23..." to list of numeric values *)
stringToBin[str_String]:=hexToByte/@StringPartition[str,4]


(* Helper, transforms an Association object into parameters managed by gateway. *)
toParams[proc_,dict_Association]:=Module[
	{
		paramKey
	},
	
	paramKey=runCommand[proc,"params"];
	
	Scan[
		Function[kv, (* kv is key->value from dist *)
			With[
				{
					fieldName=ToString[kv[[1]]],
					fieldValue=kv[[2]]
				},
				
				Switch[fieldValue,
					True|False,
						runCommand[proc,
							{"set_boolean",paramKey,fieldName,If[fieldValue,"true","false"]}
						],
					
					_Integer,
						runCommand[proc,
							{"set_integer",paramKey,fieldName,ToString[fieldValue]}
						],
					
					_YAMI4LongLong,
						runCommand[proc,
							{"set_long_long",paramKey,fieldName,ToString[fieldValue[[1]]]}
						],
					
					_Real,
						runCommand[proc,
							{"set_double_float",paramKey,fieldName,ToString[fieldValue]}
						],
						
					_String,
						runCommand[proc,
							{"set_string",paramKey,fieldName,"\""<>ToString[fieldValue]<>"\""}
						],
						
					_YAMI4Binary,
						runCommand[proc,
							{"set_binary",paramKey,fieldName,"\""<>binToString[fieldValue]<>"\""}
						],
					
					_List,
						Switch[fieldValue[[1]],
							True|False,
								runCommand[proc,
									Flatten[{"set_boolean_array",paramKey,fieldName,
										Map[
											Function[v,
												If[v,"true","false"]
											],
											fieldValue
										]
									}]
								],
							
							_Integer,
								runCommand[proc,
									Flatten[{"set_integer_array",paramKey,fieldName,
										ToString/@fieldValue
									}]
								],
							
							_YAMI4LongLong,
								runCommand[proc,
									Flatten[{"set_long_long_array",paramKey,fieldName,
										Map[
											Function[ll, (* ll is expected to be YAMI4LongLong *)
												ToString[ll[[1]]]
											],
											fieldValue
										]
									}]
								],
							
							_Real,
								runCommand[proc,
									Flatten[{"set_double_float_array",paramKey,fieldName,
										ToString/@fieldValue
									}]
								],
							
							_String,
								runCommand[proc,
									Flatten[{"set_string_array",paramKey,fieldName,
										Map[
											Function[s, (* s is single string in array *)
												"\""<>ToString[s]<>"\""
											],
											fieldValue
										]
									}]
								],
							
							_YAMI4Binary,
								runCommand[proc,
									Flatten[{"set_binary_array",paramKey,fieldName,
										Map[
											Function[b, (* b is single binary in array *)
												"\""<>binToString[b]<>"\""
											],
											fieldValue
										]
									}]
								],
							
							_,
								Message[yami4::err,"ERROR Bad data type."]
						],
					
					_,
						Message[yami4::err,"ERROR Bad data type."]
				]
			]
		],
		Normal[dict]
	];
	
	paramKey
]


(* Helper, transforms a gateway-managed parameters object into Association *)
fromParams[proc_,paramKey_String]:=Module[
	{
		fields,
		tmp,
		result=<||>
	},
	
	fields=runCommand[proc,{"fields",paramKey}];
	Scan[
		Function[ft, (* ft is {field,type} *)
			With[
				{
					fieldName=ft[[1]],
					fieldType=ft[[2]]
				},
				
				Switch[fieldType,
					"boolean",
						tmp=runCommand[proc,
							{"get_boolean",paramKey,fieldName}
						];
						result[fieldName]=tmp=="true",
					
					"integer",
						tmp=runCommand[proc,
							{"get_integer",paramKey,fieldName}
						];
						result[fieldName]=FromDigits[tmp],
					
					"long_long",
						tmp=runCommand[proc,
							{"get_long_long",paramKey,fieldName}
						];
						result[fieldName]=YAMI4LongLong[tmp],
					
					"double_float",
						tmp=runCommand[proc,
							{"get_double_float",paramKey,fieldName}
						];
						result[fieldName]=ToExpression[tmp],
					
					"string",
						tmp=runCommand[proc,
							{"get_string",paramKey,fieldName}
						];
						result[fieldName]=StringTrim[tmp,"\""],
					
					"binary",
						tmp=runCommand[proc,
							{"get_binary",paramKey,fieldName}
						];
						result[fieldName]=YAMI4Binary[stringToBin[StringTrim[tmp,"\""]]],
					
					"boolean_array",
						tmp=runCommand[proc,
							{"get_boolean_array",paramKey,fieldName}
						];
						result[fieldName]=EqualTo["true"]/@tmp,
					
					"integer_array",
						tmp=runCommand[proc,
							{"get_integer_array",paramKey,fieldName}
						];
						result[fieldName]=FromDigits/@tmp,
					
					"long_long_array",
						tmp=runCommand[proc,
							{"get_long_long_array",paramKey,fieldName}
						];
						result[fieldName]=YAMI4LongLong/@(FromDigits/@tmp),
					
					"double_float_array",
						tmp=runCommand[proc,
							{"get_double_float_array",paramKey,fieldName}
						];
						result[fieldName]=ToExpression/@tmp,
					
					"string_array",
						tmp=runCommand[proc,
							{"get_string_array",paramKey,fieldName}
						];
						result[fieldName]=Map[
							Function[s, (* s is single string in array *)
								StringTrim[s,"\""]
							],
							tmp
						],
					
					"binary_array",
						tmp=runCommand[proc,
							{"get_binary_array",paramKey,fieldName}
						];
						result[fieldName]=Map[
							Function[b, (* b is single binary in array *)
								YAMI4Binary[stringToBin[StringTrim[b,"\""]]]
							],
							tmp
						],
					
					_,
						Message[yami4::err,"ERROR bad data type."]
				]
			]
		],
		Partition[fields,2]
	];
	
	result
]


init[g_YAMI4Gateway]:=With[
	{
		proc=g[[1]]
	},
	
	runCommand[proc,{"init_agent"}]
]


init[g_YAMI4Gateway,params_Association]:=With[
	{
		proc=g[[1]],
		paramKey,
		result
	},
	
	paramKey=toParams[proc,params];
	result=runCommand[proc,{"init_agent",paramKey}];
	runCommand[proc,{"delete",paramKey}];
	
	result
]


addListener[g_YAMI4Gateway,target_String]:=With[
	{
		proc=g[[1]]
	},
	
	runCommand[proc,{"add_listener",target}]
]


registerObject[g_YAMI4Gateway,objName_String]:=With[
	{
		proc=g[[1]]
	},
	
	runCommand[proc,{"register_object",objName}]
]


hasIncoming[g_YAMI4Gateway,objName_String]:=Module[
	{
		proc=g[[1]],
		result
	},
	
	result=runCommand[proc,{"has_incoming",objName}];
	
	result=="true"
]


waitForIncoming[g_YAMI4Gateway,objName_String]:=With[
	{
		proc=g[[1]]
	},
	
	runCommand[proc,{"wait_for_incoming",objName}]
]


sourceName[g_YAMI4Gateway,objName_String]:=With[
	{
		proc=g[[1]]
	},
	
	runCommand[proc,{"get_source",objName}]
]


messageName[g_YAMI4Gateway,objName_String]:=With[
	{
		proc=g[[1]]
	},
	
	runCommand[proc,{"get_message_name",objName}]
]


requestParams[g_YAMI4Gateway,objName_String]:=Module[
	{
		proc=g[[1]],
		paramKey
	},
	
	paramKey=runCommand[proc,{"get_parameters",objName}];
	fromParams[proc,paramKey]
]


reply[g_YAMI4Gateway,objName_String]:=Module[
	{
		proc=g[[1]]
	},
	
	runCommand[proc,{"reply", objName,paramKey}]
]

reply[g_YAMI4Gateway,objName_String,params_Association]:=Module[
	{
		proc=g[[1]],
		paramKey,
		result
	},
	
	paramKey=toParams[proc,params];
	result=runCommand[proc,{"reply", objName,paramKey}];
	runCommand[proc,{"delete",paramKey}];
	
	result
]


reject[g_YAMI4Gateway,objName_String]:=Module[
	{
		proc=g[[1]]
	},
	
	runCommand[proc,{"reject", objName}]
]

reject[g_YAMI4Gateway,objName_String,exceptionMsg_String]:=Module[
	{
		proc=g[[1]]
	},
	
	runCommand[proc,{"reject", objName,exceptionMsg}]
]


consume[g_YAMI4Gateway,objName_String]:=Module[
	{
		proc=g[[1]]
	},
	
	runCommand[proc,{"consume", objName}]
]


sendOneWay[g_YAMI4Gateway,target_String,objName_String,msgName_String]:=With[
	{
		proc=g[[1]]
	},
	
	runCommand[proc,{"send_one_way",target, objName,msgName}]
]

sendOneWay[g_YAMI4Gateway,target_String,objName_String,msgName_String,params_Association]:=Module[
	{
		proc=g[[1]],
		paramKey,
		result
	},
	
	paramKey=toParams[proc,params];
	result=runCommand[proc,{"send_one_way",target, objName,msgName,paramKey}];
	runCommand[proc,{"delete",paramKey}];
	
	result
]


send[g_YAMI4Gateway,target_String,objName_String,msgName_String]:=Module[
	{
		proc=g[[1]],
		msgKey
	},
	
	msgKey=runCommand[proc,{"send",target, objName,msgName}];
	
	YAMI4OutgoingMessage[proc,msgKey]
]

send[g_YAMI4Gateway,target_String,objName_String,msgName_String,params_Association]:=Module[
	{
		proc=g[[1]],
		paramKey,
		msgKey
	},
	
	paramKey=toParams[proc,params];
	msgKey=runCommand[proc,{"send",target, objName,msgName,paramKey}];
	runCommand[proc,{"delete",paramKey}];
	
	YAMI4OutgoingMessage[proc,msgKey]
]


state[msg_YAMI4OutgoingMessage]:=Module[
	{
		proc=msg[[1]],
		msgKey=msg[[2]],
		tmp
	},
	
	tmp=runCommand[proc,{"get_state",msgKey}];
	Switch[tmp,
		"posted",
			posted,
		"transmitted",
			transmitted,
		"abandoned",
			abandoned,
		"replied",
			replied,
		"rejected",
			rejected
	]
]


waitForTransmission[msg_YAMI4OutgoingMessage]:=Module[
	{
		proc=msg[[1]],
		msgKey=msg[[2]]
	},
	
	runCommand[proc,{"wait_for_transmission",msgKey}]
]


waitForCompletion[msg_YAMI4OutgoingMessage]:=Module[
	{
		proc=msg[[1]],
		msgKey=msg[[2]]
	},
	
	runCommand[proc,{"wait_for_completion",msgKey}]
]


response[msg_YAMI4OutgoingMessage]:=Module[
	{
		proc=msg[[1]],
		msgKey=msg[[2]],
		paramKey
	},
	
	paramKey=runCommand[proc,{"get_reply",msgKey}];
	fromParams[proc,paramKey]
]


exceptionMsg[msg_YAMI4OutgoingMessage]:=Module[
	{
		proc=msg[[1]],
		msgKey=msg[[2]]
	},
	
	runCommand[proc,{"get_exception_msg",msgKey}]
]


close[msg_YAMI4OutgoingMessage]:=With[
	{
		proc=msg[[1]],
		msgKey=msg[[2]]
	},
	
	runCommand[proc,{"delete",msgKey}]
]


End[];


EndPackage[];
